@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2014-2016 Raymond Hill
+    Copyright (C) 2014-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -19,37 +19,39 @@
     Home: https://github.com/gorhill/uBlock
 */
 
-/* global uDom */
-
-/******************************************************************************/
-
-(function() {
+/* global CodeMirror */
 
 'use strict';
 
 /******************************************************************************/
 
-var onAssetContentReceived = function(details) {
-    uDom('#content').text(details && (details.content || ''));
-};
+(function() {
 
-/******************************************************************************/
+    var q = window.location.search;
+    var matches = q.match(/^\?url=([^&]+)/);
+    if ( !matches || matches.length !== 2 ) { return; }
 
-var q = window.location.search;
-var matches = q.match(/^\?url=([^&]+)/);
-if ( !matches || matches.length !== 2 ) {
-    return;
-}
+    vAPI.messaging.send(
+        'default',
+        {
+            what : 'getAssetContent',
+            url: decodeURIComponent(matches[1])
+        },
+        function(details) {
+            cmEditor.setValue(details && details.content || '');
+        }   
+    );
 
-vAPI.messaging.send(
-    'default',
-    {
-        what : 'getAssetContent',
-        url: decodeURIComponent(matches[1])
-    },
-    onAssetContentReceived
-);
-
-/******************************************************************************/
+    var cmEditor = new CodeMirror(
+        document.getElementById('content'),
+        {
+            autofocus: true,
+            lineNumbers: true,
+            lineWrapping: true,
+            readOnly: true,
+            styleActiveLine: true
+        }
+    );
 
+    uBlockDashboard.patchCodeMirrorEditor(cmEditor);
 })();
