@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2014-2017 Raymond Hill
+    Copyright (C) 2014-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -20,10 +20,17 @@
 */
 
 
-/* global objectAssign */
-
 'use strict';
 
+/******************************************************************************/
+
+// Not all platforms may have properly declared vAPI.webextFlavor.
+
+if ( vAPI.webextFlavor === undefined ) {
+    vAPI.webextFlavor = { major: 0, soup: new Set([ 'ublock' ]) };
+}
+
+
 /******************************************************************************/
 
 var µBlock = (function() { // jshint ignore:line
@@ -35,14 +42,29 @@ var µBlock = (function() { // jshint ignore:line
         assetFetchTimeout: 30,
         autoUpdateAssetFetchPeriod: 120,
         autoUpdatePeriod: 7,
+        debugScriptlets: false,
         ignoreRedirectFilters: false,
         ignoreScriptInjectFilters: false,
-        manualUpdateAssetFetchPeriod: 2000,
+        manualUpdateAssetFetchPeriod: 500,
         popupFontSize: 'unset',
         suspendTabsUntilReady: false,
         userResourcesLocation: 'unset'
     };
 
+    var whitelistDefault = [
+        'about-scheme',
+        'chrome-extension-scheme',
+        'chrome-scheme',
+        'moz-extension-scheme',
+        'opera-scheme',
+        'vivaldi-scheme',
+        'wyciwyg-scheme',   // Firefox's "What-You-Cache-Is-What-You-Get"
+    ];
+    // https://github.com/gorhill/uBlock/issues/3693#issuecomment-379782428
+    if ( vAPI.webextFlavor.soup.has('webext') === false ) {
+        whitelistDefault.push('behind-the-scene');
+    }
+
     return {
         firstInstall: false,
 
@@ -51,7 +73,7 @@ var µBlock = (function() { // jshint ignore:line
 
         userSettings: {
             advancedUserEnabled: false,
-            alwaysDetachLogger: false,
+            alwaysDetachLogger: true,
             autoUpdate: true,
             cloudStorageEnabled: false,
             collapseBlocked: true,
@@ -73,8 +95,8 @@ var µBlock = (function() { // jshint ignore:line
 
         hiddenSettingsDefault: hiddenSettingsDefault,
         hiddenSettings: (function() {
-            var out = objectAssign({}, hiddenSettingsDefault),
-                json = vAPI.localStorage.getItem('hiddenSettings');
+            var out = Object.assign({}, hiddenSettingsDefault),
+                json = vAPI.localStorage.getItem('immediateHiddenSettings');
             if ( typeof json === 'string' ) {
                 try {
                     var o = JSON.parse(json);
@@ -89,27 +111,22 @@ var µBlock = (function() { // jshint ignore:line
                 catch(ex) {
                 }
             }
+            // Remove once 1.15.12+ is widespread.
+            vAPI.localStorage.removeItem('hiddenSettings');
             return out;
         })(),
 
         // Features detection.
         privacySettingsSupported: vAPI.browserSettings instanceof Object,
         cloudStorageSupported: vAPI.cloud instanceof Object,
+        canFilterResponseBody: vAPI.net.canFilterResponseBody === true,
+        canInjectScriptletsNow: vAPI.webextFlavor.soup.has('chromium'),
 
         // https://github.com/chrisaljoudi/uBlock/issues/180
         // Whitelist directives need to be loaded once the PSL is available
         netWhitelist: {},
         netWhitelistModifyTime: 0,
-        netWhitelistDefault: [
-            'about-scheme',
-            'behind-the-scene',
-            'chrome-extension-scheme',
-            'chrome-scheme',
-            'moz-extension-scheme',
-            'opera-scheme',
-            'vivaldi-scheme',
-            ''
-        ].join('\n'),
+        netWhitelistDefault: whitelistDefault.join('\n'),
 
         localSettings: {
             blockedRequestCount: 0,
@@ -118,10 +135,10 @@ var µBlock = (function() { // jshint ignore:line
         localSettingsLastModified: 0,
         localSettingsLastSaved: 0,
 
-        // read-only
+        // Read-only
         systemSettings: {
-            compiledMagic: 'pwvcdyqfkuek',
-            selfieMagic: 'pwvcdyqfkuek'
+            compiledMagic: 3,   // Increase when compiled format changes
+            selfieMagic: 3      // Increase when selfie format changes
         },
 
         restoreBackupSettings: {
@@ -144,7 +161,7 @@ var µBlock = (function() { // jshint ignore:line
 
         selfieAfter: 17 * oneMinute,
 
-        pageStores: {},
+        pageStores: new Map(),
         pageStoresToken: 0,
 
         storageQuota: vAPI.storage.QUOTA_BYTES,
@@ -153,18 +170,19 @@ var µBlock = (function() { // jshint ignore:line
         noopFunc: function(){},
 
         apiErrorCount: 0,
-        mouseX: -1,
-        mouseY: -1,
-        mouseURL: '',
+
+        mouseEventRegister: {
+            tabId: '',
+            x: -1,
+            y: -1,
+            url: ''
+        },
+
         epickerTarget: '',
         epickerZap: false,
         epickerEprom: null,
 
-        scriptlets: {
-        },
-
-        // so that I don't have to care for last comma
-        dummy: 0
+        scriptlets: {},
     };
 
 })();
