@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2014-2017 Raymond Hill
+    Copyright (C) 2014-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -30,10 +30,6 @@
 
 /******************************************************************************/
 
-var magicId = 'chmdgxwtetgu';
-
-/******************************************************************************/
-
 var Matrix = function() {
     this.reset();
 };
@@ -74,39 +70,8 @@ var nameToActionMap = {
 /******************************************************************************/
 
 // For performance purpose, as simple tests as possible
-var reHostnameVeryCoarse = /[g-z_-]/;
-var reIPv4VeryCoarse = /\.\d+$/;
-var reBadHostname = /[^0-9a-z_.\[\]:-]/;
-
-// http://tools.ietf.org/html/rfc5952
-// 4.3: "MUST be represented in lowercase"
-// Also: http://en.wikipedia.org/wiki/IPv6_address#Literal_IPv6_addresses_in_network_resource_identifiers
-
-var isIPAddress = function(hostname) {
-    if ( reHostnameVeryCoarse.test(hostname) ) {
-        return false;
-    }
-    if ( reIPv4VeryCoarse.test(hostname) ) {
-        return true;
-    }
-    return hostname.startsWith('[');
-};
-
-var toBroaderHostname = function(hostname) {
-    var pos = hostname.indexOf('.');
-    if ( pos !== -1 ) {
-        return hostname.slice(pos + 1);
-    }
-    return hostname !== '*' && hostname !== '' ? '*' : '';
-};
-
-var toBroaderIPAddress = function(ipaddress) {
-    return ipaddress !== '*' && ipaddress !== '' ? '*' : '';
-};
-
-var selectHostnameBroadener = function(hostname) {
-    return isIPAddress(hostname) ? toBroaderIPAddress : toBroaderHostname;
-};
+var reBadHostname = /[^0-9a-z_.\[\]:%-]/;
+var reNotASCII = /[^\x20-\x7F]/;
 
 /******************************************************************************/
 
@@ -115,78 +80,71 @@ Matrix.prototype.reset = function() {
     this.type = '';
     this.y = '';
     this.z = '';
-    this.rules = {};
+    this.rules = new Map();
+    this.changed = false;
+    this.decomposedSource = [];
+    this.decomposedDestination = [];
 };
 
 /******************************************************************************/
 
 Matrix.prototype.assign = function(other) {
-    var thisRules = this.rules;
-    var otherRules = other.rules;
-    var k;
-
     // Remove rules not in other
-    for ( k in thisRules ) {
-        if ( thisRules.hasOwnProperty(k) === false ) {
-            continue;
-        }
-        if ( otherRules.hasOwnProperty(k) === false ) {
-            delete thisRules[k];
+    for ( var k of this.rules.keys() ) {
+        if ( other.rules.has(k) === false ) {
+            this.rules.delete(k);
+            this.changed = true;
         }
     }
-
     // Add/change rules in other
-    for ( k in otherRules ) {
-        if ( otherRules.hasOwnProperty(k) === false ) {
-            continue;
+    for ( var entry of other.rules ) {
+        if ( this.rules.get(entry[0]) !== entry[1] ) {
+            this.rules.set(entry[0], entry[1]);
+            this.changed = true;
         }
-        thisRules[k] = otherRules[k];
     }
 };
 
 /******************************************************************************/
 
 Matrix.prototype.copyRules = function(other, srcHostname, desHostnames) {
-    var thisRules = this.rules;
-    var otherRules = other.rules;
-    var ruleKey, ruleValue;
 
     // Specific types
-    ruleValue = otherRules['* *'] || 0;
-    if ( ruleValue !== 0 ) {
-        thisRules['* *'] = ruleValue;
+    var bits = other.rules.get('* *');
+    if ( bits !== undefined ) {
+        this.rules.set('* *', bits);
     } else {
-        delete thisRules['* *'];
+        this.rules.delete('* *');
     }
-    ruleKey = srcHostname + ' *';
-    ruleValue = otherRules[ruleKey] || 0;
-    if ( ruleValue !== 0 ) {
-        thisRules[ruleKey] = ruleValue;
+    var key = srcHostname + ' *';
+    bits = other.rules.get(key);
+    if ( bits !== undefined ) {
+        this.rules.set(key, bits);
     } else {
-        delete thisRules[ruleKey];
+        this.rules.delete(key);
     }
 
     // Specific destinations
     for ( var desHostname in desHostnames ) {
-        if ( desHostnames.hasOwnProperty(desHostname) === false ) {
-            continue;
-        }
-        ruleKey = '* ' + desHostname;
-        ruleValue = otherRules[ruleKey] || 0;
-        if ( ruleValue !== 0 ) {
-            thisRules[ruleKey] = ruleValue;
+        if ( desHostnames.hasOwnProperty(desHostname) === false ) { continue; }
+        key = '* ' + desHostname;
+        bits = other.rules.get(key);
+        if ( bits !== undefined ) {
+            this.rules.set(key, bits);
         } else {
-            delete thisRules[ruleKey];
+            this.rules.delete(key);
         }
-        ruleKey = srcHostname + ' ' + desHostname ;
-        ruleValue = otherRules[ruleKey] || 0;
-        if ( ruleValue !== 0 ) {
-            thisRules[ruleKey] = ruleValue;
+        key = srcHostname + ' ' + desHostname ;
+        bits = other.rules.get(key);
+        if ( bits !== undefined ) {
+            this.rules.set(key, bits);
         } else {
-            delete thisRules[ruleKey];
+            this.rules.delete(key);
         }
     }
 
+    this.changed = true;
+
     return true;
 };
 
@@ -198,28 +156,25 @@ Matrix.prototype.copyRules = function(other, srcHostname, desHostnames) {
 // - from to *
 
 Matrix.prototype.hasSameRules = function(other, srcHostname, desHostnames) {
-    var thisRules = this.rules;
-    var otherRules = other.rules;
-    var ruleKey;
 
     // Specific types
-    ruleKey = '* *';
-    if ( (thisRules[ruleKey] || 0) !== (otherRules[ruleKey] || 0) ) {
+    var key = '* *';
+    if ( this.rules.get(key) !== other.rules.get(key) ) {
         return false;
     }
-    ruleKey = srcHostname + ' *';
-    if ( (thisRules[ruleKey] || 0) !== (otherRules[ruleKey] || 0) ) {
+    key = srcHostname + ' *';
+    if ( this.rules.get(key) !== other.rules.get(key) ) {
         return false;
     }
 
     // Specific destinations
     for ( var desHostname in desHostnames ) {
-        ruleKey = '* ' + desHostname;
-        if ( (thisRules[ruleKey] || 0) !== (otherRules[ruleKey] || 0) ) {
+        key = '* ' + desHostname;
+        if ( this.rules.get(key) !== other.rules.get(key) ) {
             return false;
         }
-        ruleKey = srcHostname + ' ' + desHostname ;
-        if ( (thisRules[ruleKey] || 0) !== (otherRules[ruleKey] || 0) ) {
+        key = srcHostname + ' ' + desHostname ;
+        if ( this.rules.get(key) !== other.rules.get(key) ) {
             return false;
         }
     }
@@ -232,19 +187,17 @@ Matrix.prototype.hasSameRules = function(other, srcHostname, desHostnames) {
 Matrix.prototype.setCell = function(srcHostname, desHostname, type, state) {
     var bitOffset = typeBitOffsets[type];
     var k = srcHostname + ' ' + desHostname;
-    var oldBitmap = this.rules[k];
-    if ( oldBitmap === undefined ) {
-        oldBitmap = 0;
-    }
+    var oldBitmap = this.rules.get(k) || 0;
     var newBitmap = oldBitmap & ~(3 << bitOffset) | (state << bitOffset);
     if ( newBitmap === oldBitmap ) {
         return false;
     }
     if ( newBitmap === 0 ) {
-        delete this.rules[k];
+        this.rules.delete(k);
     } else {
-        this.rules[k] = newBitmap;
+        this.rules.set(k, newBitmap);
     }
+    this.changed = true;
     return true;
 };
 
@@ -256,6 +209,7 @@ Matrix.prototype.unsetCell = function(srcHostname, desHostname, type) {
         return false;
     }
     this.setCell(srcHostname, desHostname, type, 0);
+    this.changed = true;
     return true;
 };
 
@@ -265,7 +219,7 @@ Matrix.prototype.unsetCell = function(srcHostname, desHostname, type) {
 
 Matrix.prototype.evaluateCell = function(srcHostname, desHostname, type) {
     var key = srcHostname + ' ' + desHostname;
-    var bitmap = this.rules[key];
+    var bitmap = this.rules.get(key);
     if ( bitmap === undefined ) {
         return 0;
     }
@@ -307,14 +261,13 @@ var domainFromHostname = µBlock.URI.domainFromHostname;
 
 /******************************************************************************/
 
-Matrix.prototype.evaluateCellZ = function(srcHostname, desHostname, type, broadener) {
+Matrix.prototype.evaluateCellZ = function(srcHostname, desHostname, type) {
+    µBlock.decomposeHostname(srcHostname, this.decomposedSource);
     this.type = type;
-    var bitOffset = typeBitOffsets[type];
-    var s = srcHostname;
-    var v;
-    for (;;) {
-        this.z = s;
-        v = this.rules[s + ' ' + desHostname];
+    let bitOffset = typeBitOffsets[type];
+    for ( let shn of this.decomposedSource ) {
+        this.z = shn;
+        let v = this.rules.get(shn + ' ' + desHostname);
         if ( v !== undefined ) {
             v = v >>> bitOffset & 3;
             if ( v !== 0 ) {
@@ -322,8 +275,6 @@ Matrix.prototype.evaluateCellZ = function(srcHostname, desHostname, type, broade
                 return v;
             }
         }
-        s = broadener(s);
-        if ( s === '' ) { break; }
     }
     // srcHostname is '*' at this point
     this.r = 0;
@@ -334,29 +285,24 @@ Matrix.prototype.evaluateCellZ = function(srcHostname, desHostname, type, broade
 
 Matrix.prototype.evaluateCellZY = function(srcHostname, desHostname, type) {
     // Pathological cases.
-    var d = desHostname;
-    if ( d === '' ) {
+    if ( desHostname === '' ) {
         this.r = 0;
         return 0;
     }
 
-    // Prepare broadening handlers -- depends on whether we are dealing with
-    // a hostname or IP address.
-    var broadenSource = selectHostnameBroadener(srcHostname),
-        broadenDestination = selectHostnameBroadener(desHostname);
-
     // Precedence: from most specific to least specific
 
     // Specific-destination, any party, any type
-    while ( d !== '*' ) {
-        this.y = d;
-        if ( this.evaluateCellZ(srcHostname, d, '*', broadenSource) !== 0 ) {
+    µBlock.decomposeHostname(desHostname, this.decomposedDestination);
+    for ( let dhn of this.decomposedDestination ) {
+        if ( dhn === '*' ) { break; }
+        this.y = dhn;
+        if ( this.evaluateCellZ(srcHostname, dhn, '*') !== 0 ) {
             return this.r;
         }
-        d = broadenDestination(d);
     }
 
-    var thirdParty = is3rdParty(srcHostname, desHostname);
+    let thirdParty = is3rdParty(srcHostname, desHostname);
 
     // Any destination
     this.y = '*';
@@ -365,34 +311,34 @@ Matrix.prototype.evaluateCellZY = function(srcHostname, desHostname, type) {
     if ( thirdParty ) {
         // 3rd-party, specific type
         if ( type === 'script' ) {
-            if ( this.evaluateCellZ(srcHostname, '*', '3p-script', broadenSource) !== 0 ) {
+            if ( this.evaluateCellZ(srcHostname, '*', '3p-script') !== 0 ) {
                 return this.r;
             }
         } else if ( type === 'sub_frame' ) {
-            if ( this.evaluateCellZ(srcHostname, '*', '3p-frame', broadenSource) !== 0 ) {
+            if ( this.evaluateCellZ(srcHostname, '*', '3p-frame') !== 0 ) {
                 return this.r;
             }
         }
         // 3rd-party, any type
-        if ( this.evaluateCellZ(srcHostname, '*', '3p', broadenSource) !== 0 ) {
+        if ( this.evaluateCellZ(srcHostname, '*', '3p') !== 0 ) {
             return this.r;
         }
     } else if ( type === 'script' ) {
         // 1st party, specific type
-        if ( this.evaluateCellZ(srcHostname, '*', '1p-script', broadenSource) !== 0 ) {
+        if ( this.evaluateCellZ(srcHostname, '*', '1p-script') !== 0 ) {
             return this.r;
         }
     }
 
     // Any destination, any party, specific type
     if ( supportedDynamicTypes.hasOwnProperty(type) ) {
-        if ( this.evaluateCellZ(srcHostname, '*', type, broadenSource) !== 0 ) {
+        if ( this.evaluateCellZ(srcHostname, '*', type) !== 0 ) {
             return this.r;
         }
     }
 
     // Any destination, any party, any type
-    if ( this.evaluateCellZ(srcHostname, '*', '*', broadenSource) !== 0 ) {
+    if ( this.evaluateCellZ(srcHostname, '*', '*') !== 0 ) {
         return this.r;
     }
 
@@ -478,22 +424,15 @@ Matrix.prototype.desHostnameFromRule = function(rule) {
 
 /******************************************************************************/
 
-Matrix.prototype.toString = function() {
+Matrix.prototype.toArray = function() {
     var out = [],
-        rule, type, val,
-        srcHostname, desHostname,
         toUnicode = punycode.toUnicode;
-    for ( rule in this.rules ) {
-        if ( this.rules.hasOwnProperty(rule) === false ) {
-            continue;
-        }
-        srcHostname = this.srcHostnameFromRule(rule);
-        desHostname = this.desHostnameFromRule(rule);
-        for ( type in typeBitOffsets ) {
-            if ( typeBitOffsets.hasOwnProperty(type) === false ) {
-                continue;
-            }
-            val = this.evaluateCell(srcHostname, desHostname, type);
+    for ( var key of this.rules.keys() ) {
+        var srcHostname = this.srcHostnameFromRule(key);
+        var desHostname = this.desHostnameFromRule(key);
+        for ( var type in typeBitOffsets ) {
+            if ( typeBitOffsets.hasOwnProperty(type) === false ) { continue; }
+            var val = this.evaluateCell(srcHostname, desHostname, type);
             if ( val === 0 ) { continue; }
             if ( srcHostname.indexOf('xn--') !== -1 ) {
                 srcHostname = toUnicode(srcHostname);
@@ -509,108 +448,92 @@ Matrix.prototype.toString = function() {
             );
         }
     }
-    return out.join('\n');
+    return out;
+};
+
+Matrix.prototype.toString = function() {
+    return this.toArray().join('\n');
 };
 
 /******************************************************************************/
 
 Matrix.prototype.fromString = function(text, append) {
-    var lineIter = new µBlock.LineIterator(text),
-        line, pos, fields,
-        srcHostname, desHostname, type, action,
-        reNotASCII = /[^\x20-\x7F]/,
-        toASCII = punycode.toASCII;
-
-    if ( append !== true ) {
-        this.reset();
-    }
-
+    var lineIter = new µBlock.LineIterator(text);
+    if ( append !== true ) { this.reset(); }
     while ( lineIter.eot() === false ) {
-        line = lineIter.next().trim();
-        pos = line.indexOf('# ');
-        if ( pos !== -1 ) {
-            line = line.slice(0, pos).trim();
-        }
-        if ( line === '' ) {
-            continue;
-        }
-
-        // URL net filtering rules
-        if ( line.indexOf('://') !== -1 ) {
-            continue;
-        }
-
-        // Valid rule syntax:
-
-        // srcHostname desHostname type state
-        //      type = a valid request type
-        //      state = [`block`, `allow`, `noop`]
-
-        // Lines with invalid syntax silently ignored
-
-        fields = line.split(/\s+/);
-        if ( fields.length !== 4 ) {
-            continue;
-        }
-
-        // Ignore special rules:
-        //   hostname-based switch rules
-        if ( fields[0].endsWith(':') ) {
-            continue;
-        }
-
-        // Performance: avoid punycoding if hostnames are made only of
-        // ASCII characters.
-        srcHostname = fields[0];
-        if ( reNotASCII.test(srcHostname) ) {
-            srcHostname = toASCII(srcHostname);
-        }
-        desHostname = fields[1];
-        if ( reNotASCII.test(desHostname) ) {
-            desHostname = toASCII(desHostname);
-        }
-
-        // https://github.com/chrisaljoudi/uBlock/issues/1082
-        // Discard rules with invalid hostnames
-        if ( (srcHostname !== '*' && reBadHostname.test(srcHostname)) ||
-             (desHostname !== '*' && reBadHostname.test(desHostname))
-        ) {
-            continue;
-        }
-
-        type = fields[2];
-        if ( typeBitOffsets.hasOwnProperty(type) === false ) {
-            continue;
-        }
-
-        // https://github.com/chrisaljoudi/uBlock/issues/840
-        // Discard invalid rules
-        if ( desHostname !== '*' && type !== '*' ) {
-            continue;
-        }
-
-        action = nameToActionMap[fields[3]];
-        if ( typeof action !== 'number' || action < 0 || action > 3 ) {
-            continue;
-        }
-
-        this.setCell(srcHostname, desHostname, type, action);
+        this.addFromRuleParts(lineIter.next().trim().split(/\s+/));
     }
 };
 
 /******************************************************************************/
 
+Matrix.prototype.validateRuleParts = function(parts) {
+    if ( parts.length < 4 ) { return; }
+
+    // Ignore hostname-based switch rules
+    if ( parts[0].endsWith(':') ) { return; }
+
+    // Ignore URL-based rules
+    if ( parts[1].indexOf('/') !== -1 ) { return; }
+
+    if ( typeBitOffsets.hasOwnProperty(parts[2]) === false ) { return; }
+
+    if ( nameToActionMap.hasOwnProperty(parts[3]) === false ) { return; }
+
+    // https://github.com/chrisaljoudi/uBlock/issues/840
+    //   Discard invalid rules
+    if ( parts[1] !== '*' && parts[2] !== '*' ) { return; }
+
+    // Performance: avoid punycoding if hostnames are made only of ASCII chars.
+    if ( reNotASCII.test(parts[0]) ) { parts[0] = punycode.toASCII(parts[0]); }
+    if ( reNotASCII.test(parts[1]) ) { parts[1] = punycode.toASCII(parts[1]); }
+
+    // https://github.com/chrisaljoudi/uBlock/issues/1082
+    //   Discard rules with invalid hostnames
+    if (
+        (parts[0] !== '*' && reBadHostname.test(parts[0])) ||
+        (parts[1] !== '*' && reBadHostname.test(parts[1]))
+    ) {
+        return;
+    }
+
+    return parts;
+};
+
+/******************************************************************************/
+
+Matrix.prototype.addFromRuleParts = function(parts) {
+    if ( this.validateRuleParts(parts) !== undefined ) {
+        this.setCell(parts[0], parts[1], parts[2], nameToActionMap[parts[3]]);
+        return true;
+    }
+    return false;
+};
+
+Matrix.prototype.removeFromRuleParts = function(parts) {
+    if ( this.validateRuleParts(parts) !== undefined ) {
+        this.setCell(parts[0], parts[1], parts[2], 0);
+        return true;
+    }
+    return false;
+};
+
+/******************************************************************************/
+
+var magicId = 1;
+
 Matrix.prototype.toSelfie = function() {
     return {
         magicId: magicId,
-        rules: this.rules
+        rules: Array.from(this.rules)
     };
 };
 
-/******************************************************************************/
-
 Matrix.prototype.fromSelfie = function(selfie) {
-    this.rules = selfie.rules;
+    if ( selfie.magicId !== magicId ) { return false; }
+    this.rules = new Map(selfie.rules);
+    this.changed = true;
+    return true;
 };
 
 /******************************************************************************/
