@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2015-2016 Raymond Hill
+    Copyright (C) 2015-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -32,7 +32,11 @@
 var showdomButton = uDom.nodeFromId('showdom');
 
 // Don't bother if the browser is not modern enough.
-if ( typeof Map === 'undefined' || Map.polyfill || typeof WeakMap === 'undefined' ) {
+if (
+    typeof Map === 'undefined' ||
+    Map.polyfill ||
+    typeof WeakMap === 'undefined'
+) {
     showdomButton.classList.add('disabled');
     return;
 }
@@ -40,28 +44,61 @@ if ( typeof Map === 'undefined' || Map.polyfill || typeof WeakMap === 'undefined
 /******************************************************************************/
 
 var logger = self.logger;
-var messaging = vAPI.messaging;
-
-var inspectedTabId = '';
+var inspectorConnectionId;
+var inspectedTabId = 0;
 var inspectedURL = '';
 var inspectedHostname = '';
-var pollTimer = null;
-var fingerprint = null;
 var inspector = uDom.nodeFromId('domInspector');
 var domTree = uDom.nodeFromId('domTree');
-var tabSelector = uDom.nodeFromId('pageSelector');
 var uidGenerator = 1;
 var filterToIdMap = new Map();
 
 /******************************************************************************/
 
+var messaging = vAPI.messaging;
+
+messaging.addChannelListener('loggerUI', function(msg) {
+    switch ( msg.what ) {
+    case 'connectionBroken':
+        if ( inspectorConnectionId === msg.id ) {
+            filterToIdMap.clear();
+            logger.removeAllChildren(domTree);
+            inspectorConnectionId = undefined;
+        }
+        injectInspector();
+        break;
+    case 'connectionMessage':
+        if ( msg.payload.what === 'domLayoutFull' ) {
+            inspectedURL = msg.payload.url;
+            inspectedHostname = msg.payload.hostname;
+            renderDOMFull(msg.payload);
+        } else if ( msg.payload.what === 'domLayoutIncremental' ) {
+            renderDOMIncremental(msg.payload);
+        }
+        break;
+    case 'connectionRequested':
+        if ( msg.from !== 'domInspector' ) { return false; }
+        if (
+            msg.tabId === undefined ||
+            msg.tabId !== inspectedTabId
+        ) {
+            return false;
+        }
+        filterToIdMap.clear();
+        logger.removeAllChildren(domTree);
+        inspectorConnectionId = msg.id;
+        return true;
+    }
+});
+
+/******************************************************************************/
+
 var nodeFromDomEntry = function(entry) {
     var node, value;
     var li = document.createElement('li');
     li.setAttribute('id', entry.nid);
     // expander/collapser
-    node = document.createElement('span');
-    li.appendChild(node);
+    li.appendChild(document.createElement('span'));
     // selector
     node = document.createElement('code');
     node.textContent = entry.sel;
@@ -190,16 +227,14 @@ var renderDOMIncremental = function(response) {
     //  1 = node added
     // -1 = node removed
     var journal = response.journal;
-    var nodes = response.nodes;
+    var nodes = new Map(response.nodes);
     var entry, previous, li, ul;
     for ( var i = 0, n = journal.length; i < n; i++ ) {
         entry = journal[i];
         // Remove node
         if ( entry.what === -1 ) {
             li = document.getElementById(entry.nid);
-            if ( li === null ) {
-                continue;
-            }
+            if ( li === null ) { continue; }
             patchIncremental(li, -1);
             li.parentNode.removeChild(li);
             continue;
@@ -218,7 +253,7 @@ var renderDOMIncremental = function(response) {
                 continue;
             }
             ul = previous.parentElement;
-            li = nodeFromDomEntry(nodes[entry.nid]);
+            li = nodeFromDomEntry(nodes.get(entry.nid));
             ul.insertBefore(li, previous.nextElementSibling);
             patchIncremental(li, 1);
             continue;
@@ -237,7 +272,7 @@ var renderDOMIncremental = function(response) {
                 li.appendChild(ul);
                 li.classList.add('branch');
             }
-            li = nodeFromDomEntry(nodes[entry.nid]);
+            li = nodeFromDomEntry(nodes.get(entry.nid));
             ul.appendChild(li);
             patchIncremental(li, 1);
             continue;
@@ -350,7 +385,7 @@ var startDialog = (function() {
         };
     })();
 
-    var onClick = function(ev) {
+    var onClicked = function(ev) {
         var target = ev.target;
 
         // click outside the dialog proper
@@ -367,83 +402,51 @@ var startDialog = (function() {
         }
     };
 
-    var onCooked = function(entries) {
-        if ( Array.isArray(entries) === false ) {
-            return;
+    var showCommitted = function() {
+        messaging.sendTo(inspectorConnectionId, {
+            what: 'showCommitted',
+            hide: hideSelectors.join(',\n'),
+            unhide: unhideSelectors.join(',\n')
+        });
+    };
+
+    var showInteractive = function() {
+        messaging.sendTo(inspectorConnectionId, {
+            what: 'showInteractive',
+            hide: hideSelectors.join(',\n'),
+            unhide: unhideSelectors.join(',\n')
+        });
+    };
+
+    var start = function() {
+        hideSelectors = [];
+        textarea.addEventListener('input', onInputChanged);
+        var node;
+        for ( node of domTree.querySelectorAll('code.off') ) {
+            if ( node.classList.contains('filter') === false ) {
+                hideSelectors.push(selectorFromNode(node));
+            }
         }
-        hideSelectors = entries;
-        var taValue = [], i, node;
+        var taValue = [];
         var d = new Date();
         taValue.push('! ' + d.toLocaleString() + ' ' + inspectedURL);
-        for ( i = 0; i < entries.length; i++ ) {
-            taValue.push(inspectedHostname + '##' + entries[i]);
+        for ( var selector of hideSelectors ) {
+            taValue.push(inspectedHostname + '##' + selector);
         }
         var ids = new Set(), id;
-        var nodes = domTree.querySelectorAll('code.filter.off');
-        for ( i = 0; i < nodes.length; i++ ) {
-            node = nodes[i];
+        for ( node of domTree.querySelectorAll('code.filter.off') ) {
             id = node.getAttribute('data-filter-id');
-            if ( ids.has(id) ) {
-                continue;
-            }
+            if ( ids.has(id) ) { continue; }
             ids.add(id);
             unhideSelectors.push(node.textContent);
             taValue.push(inspectedHostname + '#@#' + node.textContent);
         }
         textarea.value = taValue.join('\n');
         document.body.appendChild(dialog);
-        dialog.addEventListener('click', onClick, true);
+        dialog.addEventListener('click', onClicked, true);
         showCommitted();
     };
 
-    var showCommitted = function() {
-        messaging.sendTo(
-            'loggerUI',
-            {
-                what: 'showCommitted',
-                hide: hideSelectors.join(',\n'),
-                unhide: unhideSelectors.join(',\n')
-            },
-            inspectedTabId,
-            'domInspector'
-        );
-    };
-
-    var showInteractive = function() {
-        messaging.sendTo(
-            'loggerUI',
-            {
-                what: 'showInteractive',
-                hide: hideSelectors.join(',\n'),
-                unhide: unhideSelectors.join(',\n')
-            },
-            inspectedTabId,
-            'domInspector'
-        );
-    };
-
-    var start = function() {
-        textarea.addEventListener('input', onInputChanged);
-        var node, entries = [];
-        var nodes = domTree.querySelectorAll('code.off');
-        for ( var i = 0; i < nodes.length; i++ ) {
-            node = nodes[i];
-            if ( node.classList.contains('filter') === false ) {
-                entries.push({
-                    nid: nidFromNode(node),
-                    selector: selectorFromNode(node)
-                });
-            }
-        }
-        messaging.sendTo(
-            'loggerUI',
-            { what: 'cookFilters', entries: entries },
-            inspectedTabId,
-            'domInspector',
-            onCooked
-        );
-    };
-
     var stop = function() {
         if ( inputTimer !== null ) {
             clearTimeout(inputTimer);
@@ -453,7 +456,7 @@ var startDialog = (function() {
         hideSelectors = [];
         unhideSelectors = [];
         textarea.removeEventListener('input', onInputChanged);
-        dialog.removeEventListener('click', onClick, true);
+        dialog.removeEventListener('click', onClicked, true);
         document.body.removeChild(dialog);
     };
 
@@ -462,12 +465,10 @@ var startDialog = (function() {
 
 /******************************************************************************/
 
-var onClick = function(ev) {
+var onClicked = function(ev) {
     ev.stopPropagation();
 
-    if ( inspectedTabId === '' ) {
-        return;
-    }
+    if ( inspectedTabId === 0 ) { return; }
 
     var target = ev.target;
     var parent = target.parentElement;
@@ -479,30 +480,28 @@ var onClick = function(ev) {
         parent.classList.contains('branch') &&
         target === parent.firstElementChild
     ) {
-        target.parentElement.classList.toggle('show');
+        var state = parent.classList.toggle('show');
+        if ( !state ) {
+            for ( var node of parent.querySelectorAll('.branch') ) {
+                node.classList.remove('show');
+            }
+        }
         return;
     }
 
     // Not a node or filter 
-    if ( target.localName !== 'code' ) {
-        return;
-    }
+    if ( target.localName !== 'code' ) { return; }
 
     // Toggle cosmetic filter
     if ( target.classList.contains('filter') ) {
-        messaging.sendTo(
-            'loggerUI',
-            {
-                what: 'toggleFilter',
-                original: false,
-                target: target.classList.toggle('off'),
-                selector: selectorFromNode(target),
-                filter: selectorFromFilter(target),
-                nid: ''
-            },
-            inspectedTabId,
-            'domInspector'
-        );
+        messaging.sendTo(inspectorConnectionId, {
+            what: 'toggleFilter',
+            original: false,
+            target: target.classList.toggle('off'),
+            selector: selectorFromNode(target),
+            filter: selectorFromFilter(target),
+            nid: nidFromNode(target)
+        });
         uDom('[data-filter-id="' + target.getAttribute('data-filter-id') + '"]', inspector).toggleClass(
             'off',
             target.classList.contains('off')
@@ -510,18 +509,13 @@ var onClick = function(ev) {
     }
     // Toggle node
     else {
-        messaging.sendTo(
-            'loggerUI',
-            {
-                what: 'toggleNodes',
-                original: true,
-                target: target.classList.toggle('off') === false,
-                selector: selectorFromNode(target),
-                nid: nidFromNode(target)
-            },
-            inspectedTabId,
-            'domInspector'
-        );
+        messaging.sendTo(inspectorConnectionId, {
+            what: 'toggleNodes',
+            original: true,
+            target: target.classList.toggle('off') === false,
+            selector: selectorFromNode(target),
+            nid: nidFromNode(target)
+        });
     }
 
     var cantCreate = domTree.querySelector('.off') === null;
@@ -537,38 +531,25 @@ var onMouseOver = (function() {
 
     var timerHandler = function() {
         mouseoverTimer = null;
-        messaging.sendTo(
-            'loggerUI',
-            {
-                what: 'highlightOne',
-                selector: selectorFromNode(mouseoverTarget),
-                nid: nidFromNode(mouseoverTarget),
-                scrollTo: true
-            },
-            inspectedTabId,
-            'domInspector'
-        );
+        messaging.sendTo(inspectorConnectionId, {
+            what: 'highlightOne',
+            selector: selectorFromNode(mouseoverTarget),
+            nid: nidFromNode(mouseoverTarget),
+            scrollTo: true
+        });
     };
 
     return function(ev) {
-        if ( inspectedTabId === '' ) {
-            return;
-        }
+        if ( inspectedTabId === 0 ) { return; }
         // Convenience: skip real-time highlighting if shift key is pressed.
-        if ( ev.shiftKey ) {
-            return;
-        }
+        if ( ev.shiftKey ) { return; }
         // Find closest `li`
         var target = ev.target;
         while ( target !== null ) {
-            if ( target.localName === 'li' ) {
-                break;
-            }
+            if ( target.localName === 'li' ) { break; }
             target = target.parentElement;
         }
-        if ( target === mouseoverTarget ) {
-            return;
-        }
+        if ( target === mouseoverTarget ) { return; }
         mouseoverTarget = target;
         if ( mouseoverTimer === null ) {
             mouseoverTimer = vAPI.setTimeout(timerHandler, 50);
@@ -579,122 +560,33 @@ var onMouseOver = (function() {
 /******************************************************************************/
 
 var currentTabId = function() {
-    if ( showdomButton.classList.contains('active') === false ) {
-        return '';
-    }
-    var tabId = logger.tabIdFromClassName(tabSelector.value) || '';
-    return tabId !== 'bts' ? tabId : '';
+    if ( showdomButton.classList.contains('active') === false ) { return 0; }
+    return logger.tabIdFromPageSelector();
 };
 
 /******************************************************************************/
 
-var fetchDOMAsync = (function() {
-    var onFetched = function(response) {
-        if ( !response || currentTabId() !== inspectedTabId ) {
-            shutdownInspector();
-            injectInspectorAsync(250);
-            return;
-        }
-
-        switch ( response.status ) {
-        case 'full':
-            renderDOMFull(response);
-            fingerprint = response.fingerprint;
-            inspectedURL = response.url;
-            inspectedHostname = response.hostname;
-            break;
-
-        case 'incremental':
-            renderDOMIncremental(response);
-            break;
-
-        case 'nochange':
-        case 'busy':
-            break;
-
-        default:
-            break;
-        }
-
-        fetchDOMAsync();
-    };
-
-    var onTimeout = function() {
-        pollTimer = null;
-        messaging.sendTo(
-            'loggerUI',
-            {
-                what: 'domLayout',
-                fingerprint: fingerprint
-            },
-            inspectedTabId,
-            'domInspector',
-            onFetched
-        );
-    };
-
-    // Poll for DOM layout data every ~2 seconds at most
-    return function(delay) {
-        if ( pollTimer === null ) {
-            pollTimer = vAPI.setTimeout(onTimeout, delay || 2003);
-        }
-    };
-})();
-
-/******************************************************************************/
-
 var injectInspector = function() {
     var tabId = currentTabId();
-    // No valid tab, go back
-    if ( tabId === '' ) {
-        injectInspectorAsync();
-        return;
-    }
+    if ( tabId === 0 ) { return; }
     inspectedTabId = tabId;
-    fingerprint = null;
-    messaging.send(
-        'loggerUI',
-        {
-            what: 'scriptlet',
-            tabId: tabId,
-            scriptlet: 'dom-inspector'
-        }
-    );
-    fetchDOMAsync(250);
-};
-
-/******************************************************************************/
-
-var injectInspectorAsync = function(delay) {
-    if ( pollTimer !== null ) {
-        return;
-    }
-    if ( showdomButton.classList.contains('active') === false ) {
-        return;
-    }
-    pollTimer = vAPI.setTimeout(function() {
-        pollTimer = null;
-        injectInspector();
-    }, delay || 1001);
+    messaging.send('loggerUI', {
+        what: 'scriptlet',
+        tabId: tabId,
+        scriptlet: 'dom-inspector'
+    });
 };
 
 /******************************************************************************/
 
 var shutdownInspector = function() {
-    if ( inspectedTabId !== '' ) {
-        messaging.sendTo(
-            'loggerUI',
-            { what: 'shutdown' },
-            inspectedTabId,
-            'domInspector'
-        );
+    if ( inspectorConnectionId !== undefined ) {
+        messaging.disconnectFrom(inspectorConnectionId);
+        inspectorConnectionId = undefined;
     }
     logger.removeAllChildren(domTree);
-    if ( pollTimer !== null ) {
-        clearTimeout(pollTimer);
-        pollTimer = null;
-    }
-    inspectedTabId = '';
+    inspector.classList.add('vCompact');
+    inspectedTabId = 0;
 };
 
 /******************************************************************************/
@@ -702,34 +594,38 @@ var shutdownInspector = function() {
 var onTabIdChanged = function() {
     if ( inspectedTabId !== currentTabId() ) {
         shutdownInspector();
-        injectInspectorAsync(250);
+        injectInspector();
     }
 };
 
 /******************************************************************************/
 
+var toggleVCompactView = function() {
+    var state = !inspector.classList.toggle('vCompact');
+    var branches = document.querySelectorAll('#domInspector li.branch');
+    for ( var branch of branches ) {
+        branch.classList.toggle('show', state);
+    }
+};
+
+var toggleHCompactView = function() {
+    inspector.classList.toggle('hCompact');
+};
+
+/******************************************************************************/
+
 var toggleHighlightMode = function() {
-    messaging.sendTo(
-        'loggerUI',
-        {
-            what: 'highlightMode',
-            invert: uDom.nodeFromSelector('#domInspector .permatoolbar .highlightMode').classList.toggle('invert')
-        },
-        inspectedTabId,
-        'domInspector'
-    );
+    messaging.sendTo(inspectorConnectionId, {
+        what: 'highlightMode',
+        invert: uDom.nodeFromSelector('#domInspector .permatoolbar .highlightMode').classList.toggle('invert')
+    });
 };
 
 /******************************************************************************/
 
 var revert = function() {
     uDom('#domTree .off').removeClass('off');
-    messaging.sendTo(
-        'loggerUI',
-        { what: 'resetToggledNodes' },
-        inspectedTabId,
-        'domInspector'
-    );
+    messaging.sendTo(inspectorConnectionId, { what: 'resetToggledNodes' });
     inspector.querySelector('.permatoolbar .revert').classList.add('disabled');
     inspector.querySelector('.permatoolbar .commit').classList.add('disabled');
 };
@@ -738,9 +634,11 @@ var revert = function() {
 
 var toggleOn = function() {
     window.addEventListener('beforeunload', toggleOff);
-    tabSelector.addEventListener('change', onTabIdChanged);
-    domTree.addEventListener('click', onClick, true);
+    document.addEventListener('tabIdChanged', onTabIdChanged);
+    domTree.addEventListener('click', onClicked, true);
     domTree.addEventListener('mouseover', onMouseOver, true);
+    uDom.nodeFromSelector('#domInspector .vCompactToggler').addEventListener('click', toggleVCompactView);
+    uDom.nodeFromSelector('#domInspector .hCompactToggler').addEventListener('click', toggleHCompactView);
     uDom.nodeFromSelector('#domInspector .permatoolbar .highlightMode').addEventListener('click', toggleHighlightMode);
     uDom.nodeFromSelector('#domInspector .permatoolbar .revert').addEventListener('click', revert);
     uDom.nodeFromSelector('#domInspector .permatoolbar .commit').addEventListener('click', startDialog);
@@ -752,13 +650,15 @@ var toggleOn = function() {
 var toggleOff = function() {
     shutdownInspector();
     window.removeEventListener('beforeunload', toggleOff);
-    tabSelector.removeEventListener('change', onTabIdChanged);
-    domTree.removeEventListener('click', onClick, true);
+    document.removeEventListener('tabIdChanged', onTabIdChanged);
+    domTree.removeEventListener('click', onClicked, true);
     domTree.removeEventListener('mouseover', onMouseOver, true);
+    uDom.nodeFromSelector('#domInspector .vCompactToggler').removeEventListener('click', toggleVCompactView);
+    uDom.nodeFromSelector('#domInspector .hCompactToggler').removeEventListener('click', toggleHCompactView);
     uDom.nodeFromSelector('#domInspector .permatoolbar .highlightMode').removeEventListener('click', toggleHighlightMode);
     uDom.nodeFromSelector('#domInspector .permatoolbar .revert').removeEventListener('click', revert);
     uDom.nodeFromSelector('#domInspector .permatoolbar .commit').removeEventListener('click', startDialog);
-    inspectedTabId = '';
+    inspectedTabId = 0;
 };
 
 /******************************************************************************/
@@ -778,5 +678,3 @@ showdomButton.addEventListener('click', toggle);
 /******************************************************************************/
 
 })();
-
-
