@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2015-2017 Raymond Hill
+    Copyright (C) 2015-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -29,8 +29,12 @@
 
 /******************************************************************************/
 
-var logger = self.logger = {};
+var logger = self.logger = {
+    ownerId: Date.now()
+};
+
 var messaging = vAPI.messaging;
+var activeTabId;
 
 /******************************************************************************/
 
@@ -42,12 +46,19 @@ var removeAllChildren = logger.removeAllChildren = function(node) {
 
 /******************************************************************************/
 
-var tabIdFromClassName = logger.tabIdFromClassName = function(className) {
-    var matches = className.match(/(?:^| )tab_([^ ]+)(?: |$)/);
-    if ( matches === null ) {
-        return '';
+var tabIdFromClassName = function(className) {
+    var matches = className.match(/\btab_([^ ]+)\b/);
+    if ( matches === null ) { return 0; }
+    if ( matches[1] === 'bts' ) { return -1; }
+    return parseInt(matches[1], 10);
+};
+
+var tabIdFromPageSelector = logger.tabIdFromPageSelector = function() {
+    var tabClass = uDom.nodeFromId('pageSelector').value;
+    if ( tabClass === 'tab_active' && activeTabId !== undefined ) {
+        return activeTabId;
     }
-    return matches[1];
+    return /^tab_\d+$/.test(tabClass) ? parseInt(tabClass.slice(4), 10) : 0;
 };
 
 /******************************************************************************/
@@ -70,8 +81,7 @@ var tdJunkyard = [];
 var firstVarDataCol = 2;  // currently, column 2 (0-based index)
 var lastVarDataIndex = 4; // currently, d0-d3
 var maxEntries = 5000;
-var noTabId = '';
-var allTabIds = {};
+var allTabIds = new Map();
 var allTabIdsToken;
 var hiddenTemplate = document.querySelector('#hiddenTemplate > span');
 var reRFC3986 = /^([^:\/?#]+:)?(\/\/[^\/?#]*)?([^?#]*)(\?[^#]*)?(#.*)?/;
@@ -97,16 +107,17 @@ var staticFilterTypes = {
     'css': 'stylesheet',
     'frame': 'subdocument',
     'ping': 'other',
+    'object_subrequest': 'object',
     'xhr': 'xmlhttprequest'
 };
 
 /******************************************************************************/
 
 var classNameFromTabId = function(tabId) {
-    if ( tabId === noTabId ) {
+    if ( tabId < 0 ) {
         return 'tab_bts';
     }
-    if ( tabId !== '' ) {
+    if ( tabId !== 0 ) {
         return 'tab_' + tabId;
     }
     return '';
@@ -343,7 +354,7 @@ var renderLogEntry = function(entry) {
     if ( entry.tab ) {
         tr.classList.add('tab');
         tr.classList.add(classNameFromTabId(entry.tab));
-        if ( entry.tab === noTabId ) {
+        if ( entry.tab < 0 ) {
             tr.cells[1].appendChild(createHiddenTextNode('bts'));
         }
     }
@@ -365,24 +376,20 @@ var logDate = new Date(),
 var renderLogEntries = function(response) {
     document.body.classList.toggle('colorBlind', response.colorBlind);
 
-    var entries = response.entries;
-    if ( entries.length === 0 ) {
-        return;
-    }
+    let entries = response.entries;
+    if ( entries.length === 0 ) { return; }
 
     // Preserve scroll position
-    var height = tbody.offsetHeight;
+    let height = tbody.offsetHeight;
 
-    var tabIds = response.tabIds;
-    var n = entries.length;
-    var entry, tr;
-    for ( var i = 0; i < n; i++ ) {
-        entry = entries[i];
-        tr = renderLogEntry(entries[i]);
+    let tabIds = allTabIds;
+    for ( let i = 0, n = entries.length; i < n; i++ ) {
+        let entry = entries[i];
+        let tr = renderLogEntry(entries[i]);
         // https://github.com/gorhill/uBlock/issues/1613#issuecomment-217637122
         // Unlikely, but it may happen: mark as void if associated tab no
         // longer exist.
-        if ( entry.tab && tabIds.hasOwnProperty(entry.tab) === false ) {
+        if ( entry.tab && tabIds.has(entry.tab) === false ) {
             tr.classList.remove('canMtx');
         }
     }
@@ -393,11 +400,9 @@ var renderLogEntries = function(response) {
     truncateLog(maxEntries);
 
     // Follow waterfall if not observing top of waterfall.
-    var yDelta = tbody.offsetHeight - height;
-    if ( yDelta === 0 ) {
-        return;
-    }
-    var container = uDom.nodeFromSelector('#netInspector .vscrollable');
+    let yDelta = tbody.offsetHeight - height;
+    if ( yDelta === 0 ) { return; }
+    let container = uDom.nodeFromSelector('#netInspector .vscrollable');
     if ( container.scrollTop !== 0 ) {
         container.scrollTop += yDelta;
     }
@@ -405,20 +410,35 @@ var renderLogEntries = function(response) {
 
 /******************************************************************************/
 
+let updateCurrentTabTitle = (function() {
+    let i18nCurrentTab = vAPI.i18n('loggerCurrentTab');
+
+    return function() {
+        let select = uDom.nodeFromId('pageSelector');
+        if ( select.value !== 'tab_active' ) { return; }
+        let opt0 = select.querySelector('[value="tab_active"]');
+        let opt1 = select.querySelector('[value="tab_' + activeTabId + '"]');
+        let text = i18nCurrentTab;
+        if ( opt1 !== null ) {
+            text += ' / ' + opt1.textContent;
+        }
+        opt0.textContent = text;
+    };
+})();
+
+/******************************************************************************/
+
 var synchronizeTabIds = function(newTabIds) {
+    var select = uDom.nodeFromId('pageSelector');
+    var selectValue = select.value;
+
     var oldTabIds = allTabIds;
-    var autoDeleteVoidRows = !!vAPI.localStorage.getItem('loggerAutoDeleteVoidRows');
+    var autoDeleteVoidRows = selectValue === 'tab_active';
     var rowVoided = false;
-    var trs;
-    for ( var tabId in oldTabIds ) {
-        if ( oldTabIds.hasOwnProperty(tabId) === false ) {
-            continue;
-        }
-        if ( newTabIds.hasOwnProperty(tabId) ) {
-            continue;
-        }
+    for ( var tabId of oldTabIds.keys() ) {
+        if ( newTabIds.has(tabId) ) { continue; }
         // Mark or remove voided rows
-        trs = uDom('.tab_' + tabId);
+        var trs = uDom('.tab_' + tabId);
         if ( autoDeleteVoidRows ) {
             toJunkyard(trs);
         } else {
@@ -431,24 +451,20 @@ var synchronizeTabIds = function(newTabIds) {
         }
     }
 
-    var select = uDom.nodeFromId('pageSelector');
-    var selectValue = select.value;
-    var tabIds = Object.keys(newTabIds).sort(function(a, b) {
-        return newTabIds[a].localeCompare(newTabIds[b]);
+    var tabIds = Array.from(newTabIds.keys()).sort(function(a, b) {
+        return newTabIds.get(a).localeCompare(newTabIds.get(b));
     });
     var option;
-    for ( var i = 0, j = 2; i < tabIds.length; i++ ) {
+    for ( var i = 0, j = 3; i < tabIds.length; i++ ) {
         tabId = tabIds[i];
-        if ( tabId === noTabId ) {
-            continue;
-        }
+        if ( tabId < 0 ) { continue; }
         option = select.options[j];
         if ( !option ) {
             option = document.createElement('option');
             select.appendChild(option);
         }
         // Truncate too long labels.
-        option.textContent = newTabIds[tabId].slice(0, 80);
+        option.textContent = newTabIds.get(tabId).slice(0, 80);
         option.value = classNameFromTabId(tabId);
         if ( option.value === selectValue ) {
             select.selectedIndex = j;
@@ -470,6 +486,8 @@ var synchronizeTabIds = function(newTabIds) {
 
     allTabIds = newTabIds;
 
+    updateCurrentTabTitle();
+
     return rowVoided;
 };
 
@@ -491,8 +509,17 @@ var truncateLog = function(size) {
 /******************************************************************************/
 
 var onLogBufferRead = function(response) {
-    // This tells us the behind-the-scene tab id
-    noTabId = response.noTabId;
+    if ( !response || response.unavailable ) {
+        readLogBufferAsync();
+        return;
+    }
+
+    // Tab id of currently active tab
+    let activeTabIdChanged = false;
+    if ( response.activeTabId ) {
+        activeTabIdChanged = response.activeTabId !== activeTabId;
+        activeTabId = response.activeTabId;
+    }
 
     // This may have changed meanwhile
     if ( response.maxEntries !== maxEntries ) {
@@ -500,16 +527,20 @@ var onLogBufferRead = function(response) {
         uDom('#maxEntries').val(maxEntries || '');
     }
 
+    if ( Array.isArray(response.tabIds) ) {
+        response.tabIds = new Map(response.tabIds);
+    }
+
     // Neuter rows for which a tab does not exist anymore
-    var rowVoided = false;
-    if ( response.tabIdsToken !== allTabIdsToken ) {
+    let rowVoided = false;
+    if ( response.tabIds !== undefined ) {
         rowVoided = synchronizeTabIds(response.tabIds);
         allTabIdsToken = response.tabIdsToken;
     }
 
-    // https://github.com/gorhill/uBlock/issues/507
-    // Ensure tab selector is in sync with URL hash
-    pageSelectorFromURLHash();
+    if ( activeTabIdChanged ) {
+        pageSelectorFromURLHash();
+    }
 
     renderLogEntries(response);
 
@@ -526,7 +557,7 @@ var onLogBufferRead = function(response) {
         tbody.querySelector('tr') === null
     );
 
-    vAPI.setTimeout(readLogBuffer, 1200);
+    readLogBufferAsync();
 };
 
 /******************************************************************************/
@@ -536,52 +567,85 @@ var onLogBufferRead = function(response) {
 // require a bit more code to ensure no multi time out events.
 
 var readLogBuffer = function() {
-    messaging.send('loggerUI', { what: 'readAll' }, onLogBufferRead);
+    if ( logger.ownerId === undefined ) { return; }
+    vAPI.messaging.send(
+        'loggerUI',
+        {
+            what: 'readAll',
+            ownerId: logger.ownerId,
+            tabIdsToken: allTabIdsToken
+        },
+        onLogBufferRead
+    );
 };
 
+var readLogBufferAsync = function() {
+    if ( logger.ownerId === undefined ) { return; }
+    vAPI.setTimeout(readLogBuffer, 1200);
+};
+ 
 /******************************************************************************/
 
-var pageSelectorChanged = function() {
-    window.location.replace('#' + uDom.nodeFromId('pageSelector').value);
+let pageSelectorChanged = function() {
+    let select = uDom.nodeFromId('pageSelector');
+    window.location.replace('#' + select.value);
     pageSelectorFromURLHash();
 };
 
-/******************************************************************************/
+let pageSelectorFromURLHash = (function() {
+    let lastTabClass = '';
+    let lastEffectiveTabClass = '';
+    let reActiveTabId = /^(tab_[^+]+)\+(.+)$/;
 
-var pageSelectorFromURLHash = (function() {
-    var lastHash = '';
+    let selectRows = function(tabClass) {
+        let effectiveTabClass = tabClass;
+        if ( tabClass === 'tab_active' ) {
+            if ( activeTabId === undefined ) { return; }
+            effectiveTabClass = 'tab_' + activeTabId;
+        }
+        if ( effectiveTabClass === lastEffectiveTabClass ) { return; }
+        lastEffectiveTabClass = effectiveTabClass;
+
+        document.dispatchEvent(new Event('tabIdChanged'));
+
+        let style = uDom.nodeFromId('tabFilterer');
+        let sheet = style.sheet;
+        while ( sheet.cssRules.length !== 0 )  {
+            sheet.deleteRule(0);
+        }
+        if ( effectiveTabClass === '' ) { return; }
+        sheet.insertRule(
+            '#netInspector tr:not(.' + effectiveTabClass + '):not(.tab_bts) ' +
+            '{display:none;}',
+            0
+        );
+
+        updateCurrentTabTitle();
+    };
 
     return function() {
-        var hash = window.location.hash;
-        if ( hash === lastHash ) {
-            return;
+        let tabClass = window.location.hash.slice(1);
+        let match = reActiveTabId.exec(tabClass);
+        if ( match !== null ) {
+            tabClass = match[1];
+            activeTabId = parseInt(match[2], 10) || undefined;
+            window.location.hash = '#' + match[1];
         }
+        selectRows(tabClass);
+        if ( tabClass === lastTabClass ) { return; }
+        lastTabClass = tabClass;
 
-        var tabClass = hash.slice(1);
-        var select = uDom.nodeFromId('pageSelector');
-        var option = select.querySelector('option[value="' + tabClass + '"]');
+        let select = uDom.nodeFromId('pageSelector');
+        let option = select.querySelector('option[value="' + tabClass + '"]');
         if ( option === null ) {
-            hash = window.location.hash = '';
+            window.location.hash = '';
             tabClass = '';
             option = select.options[0];
         }
 
-        lastHash = hash;
-
         select.selectedIndex = option.index;
         select.value = option.value;
 
-        var style = uDom.nodeFromId('tabFilterer');
-        var sheet = style.sheet;
-        while ( sheet.cssRules.length !== 0 )  {
-            sheet.deleteRule(0);
-        }
-        if ( tabClass !== '' ) {
-            sheet.insertRule(
-                '#netInspector tr:not(.' + tabClass + ') { display: none; }',
-                0
-            );
-        }
         uDom('.needtab').toggleClass(
             'disabled',
             tabClass === '' || tabClass === 'tab_bts'
@@ -591,13 +655,14 @@ var pageSelectorFromURLHash = (function() {
 
 /******************************************************************************/
 
-var reloadTab = function() {
-    var tabClass = uDom.nodeFromId('pageSelector').value;
-    var tabId = tabIdFromClassName(tabClass);
-    if ( tabId === 'bts' || tabId === '' ) {
-        return;
-    }
-    messaging.send('loggerUI', { what: 'reloadTab', tabId: tabId });
+var reloadTab = function(ev) {
+    var tabId = tabIdFromPageSelector();
+    if ( tabId === 0 ) { return; }
+    messaging.send('loggerUI', {
+        what: 'reloadTab',
+        tabId: tabId,
+        bypassCache: ev && (ev.ctrlKey || ev.metaKey || ev.shiftKey)
+    });
 };
 
 /******************************************************************************/
@@ -932,7 +997,7 @@ var netFilteringManager = (function() {
         // First, whether picker can be used
         dialog.querySelector('.picker').classList.toggle(
             'hide',
-            targetTabId === noTabId ||
+            targetTabId < 0 ||
             targetType !== 'image' ||
             /(?:^| )[dlsu]b(?: |$)/.test(targetRow.className)
         );
@@ -1480,13 +1545,17 @@ var toJunkyard = function(trs) {
 /******************************************************************************/
 
 var clearBuffer = function() {
-    var tabId = uDom.nodeFromId('pageSelector').value || null;
+    var tabClass = uDom.nodeFromId('pageSelector').value;
+    var btsAlso = tabClass === '' || tabClass === 'tab_bts';
     var tbody = document.querySelector('#netInspector tbody');
     var tr = tbody.lastElementChild;
     var trPrevious;
     while ( tr !== null ) {
         trPrevious = tr.previousElementSibling;
-        if ( tabId === null || tr.classList.contains(tabId) ) {
+        if (
+            (tr.clientHeight > 0) &&
+            (tr.classList.contains('tab_bts') === false || btsAlso)
+        ) {
             trJunkyard.push(tbody.removeChild(tr));
         }
         tr = trPrevious;
@@ -1514,8 +1583,13 @@ var cleanBuffer = function() {
 
 /******************************************************************************/
 
-var toggleCompactView = function() {
-    uDom.nodeFromId('netInspector').classList.toggle('compactView');
+var toggleVCompactView = function() {
+    uDom.nodeFromId('netInspector').classList.toggle('vCompact');
+    uDom('#netInspector .vExpanded').toggleClass('vExpanded');
+};
+
+var toggleVCompactRow = function(ev) {
+    ev.target.parentElement.classList.toggle('vExpanded');
 };
 
 /******************************************************************************/
@@ -1568,12 +1642,7 @@ var popupManager = (function() {
     var toggleOn = function(td) {
         var tr = td.parentNode;
         realTabId = localTabId = tabIdFromClassName(tr.className);
-        if ( realTabId === '' ) {
-            return;
-        }
-        if ( localTabId === 'bts' ) {
-            realTabId = noTabId;
-        }
+        if ( realTabId === 0 ) { return; }
 
         container = uDom.nodeFromId('popupContainer');
 
@@ -1639,20 +1708,47 @@ var popupManager = (function() {
 
 /******************************************************************************/
 
+var grabView = function() {
+    if ( logger.ownerId === undefined ) {
+        logger.ownerId = Date.now();
+    }
+    readLogBufferAsync();
+};
+
+var releaseView = function() {
+    if ( logger.ownerId === undefined ) { return; }
+    vAPI.messaging.send(
+        'loggerUI',
+        { what: 'releaseView', ownerId: logger.ownerId }
+    );
+    logger.ownerId = undefined;
+};
+
+window.addEventListener('pagehide', releaseView);
+window.addEventListener('pageshow', grabView);
+// https://bugzilla.mozilla.org/show_bug.cgi?id=1398625
+window.addEventListener('beforeunload', releaseView);
+
+/******************************************************************************/
+
 readLogBuffer();
 
 uDom('#pageSelector').on('change', pageSelectorChanged);
 uDom('#refresh').on('click', reloadTab);
 uDom('#showdom').on('click', toggleInspectors);
 
-uDom('#compactViewToggler').on('click', toggleCompactView);
+uDom('#netInspector .vCompactToggler').on('click', toggleVCompactView);
 uDom('#clean').on('click', cleanBuffer);
 uDom('#clear').on('click', clearBuffer);
 uDom('#maxEntries').on('change', onMaxEntriesChanged);
+uDom('#netInspector table').on('click', 'tr > td:nth-of-type(1)', toggleVCompactRow);
 uDom('#netInspector table').on('click', 'tr.canMtx > td:nth-of-type(2)', popupManager.toggleOn);
 uDom('#netInspector').on('click', 'tr.canLookup > td:nth-of-type(3)', reverseLookupManager.toggleOn);
 uDom('#netInspector').on('click', 'tr.cat_net > td:nth-of-type(4)', netFilteringManager.toggleOn);
 
+// https://github.com/gorhill/uBlock/issues/507
+// Ensure tab selector is in sync with URL hash
+pageSelectorFromURLHash();
 window.addEventListener('hashchange', pageSelectorFromURLHash);
 
 /******************************************************************************/
