@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2014-2017 Raymond Hill
+    Copyright (C) 2014-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -88,7 +88,10 @@ var onMessage = function(request, sender, callback) {
         break;
     }
 
-    var tabId = sender && sender.tab ? sender.tab.id : 0;
+    // The concatenation with the empty string ensure that the resulting value
+    // is a string. This is important since tab id values are assumed to be
+    // of string type.
+    var tabId = sender && sender.tab ? '' + sender.tab.id : 0;
 
     // Sync
     var response;
@@ -99,17 +102,11 @@ var onMessage = function(request, sender, callback) {
         break;
 
     case 'compileCosmeticFilterSelector':
-        response = µb.cosmeticFilteringEngine.compileSelector(request.selector);
+        response = µb.staticExtFilteringEngine.compileSelector(request.selector);
         break;
 
     case 'cosmeticFiltersInjected':
         µb.cosmeticFilteringEngine.addToSelectorCache(request);
-        /* falls through */
-    case 'cosmeticFiltersActivated':
-        // Net-based cosmetic filters are of no interest for logging purpose.
-        if ( µb.logger.isEnabled() && request.type !== 'net' ) {
-            µb.logCosmeticFilters(tabId);
-        }
         break;
 
     case 'createUserFilter':
@@ -120,11 +117,16 @@ var onMessage = function(request, sender, callback) {
 
     case 'forceUpdateAssets':
         µb.scheduleAssetUpdater(0);
-        µb.assets.updateStart({ delay: µb.hiddenSettings.manualUpdateAssetFetchPeriod || 2000 });
+        µb.assets.updateStart({
+            delay: µb.hiddenSettings.manualUpdateAssetFetchPeriod
+        });
         break;
 
     case 'getAppData':
-        response = {name: vAPI.app.name, version: vAPI.app.version};
+        response = {
+            name: chrome.runtime.getManifest().name,
+            version: vAPI.app.version
+        };
         break;
 
     case 'getDomainNames':
@@ -132,12 +134,16 @@ var onMessage = function(request, sender, callback) {
         break;
 
     case 'getWhitelist':
-        response = µb.stringFromWhitelist(µb.netWhitelist);
+        response = {
+            whitelist: µb.stringFromWhitelist(µb.netWhitelist),
+            reBadHostname: µb.reWhitelistBadHostname.source,
+            reHostnameExtractor: µb.reWhitelistHostnameExtractor.source
+        };
         break;
 
     case 'launchElementPicker':
         // Launched from some auxiliary pages, clear context menu coords.
-        µb.mouseX = µb.mouseY = -1;
+        µb.mouseEventRegister.x = µb.mouseEventRegister.y = -1;
         µb.elementPickerExec(request.tabId, request.targetURL, request.zap);
         break;
 
@@ -146,14 +152,15 @@ var onMessage = function(request, sender, callback) {
         break;
 
     case 'mouseClick':
-        µb.mouseX = request.x;
-        µb.mouseY = request.y;
-        µb.mouseURL = request.url;
+        µb.mouseEventRegister.tabId = tabId;
+        µb.mouseEventRegister.x = request.x;
+        µb.mouseEventRegister.y = request.y;
+        µb.mouseEventRegister.url = request.url;
         break;
 
     case 'reloadTab':
         if ( vAPI.isBehindTheSceneTabId(request.tabId) === false ) {
-            vAPI.tabs.reload(request.tabId);
+            vAPI.tabs.reload(request.tabId, request.bypassCache === true);
             if ( request.select && vAPI.tabs.select ) {
                 vAPI.tabs.select(request.tabId);
             }
@@ -426,7 +433,7 @@ var onMessage = function(request, sender, callback) {
         pageStore = µb.pageStoreFromTabId(request.tabId);
         if ( pageStore ) {
             pageStore.toggleNetFilteringSwitch(request.url, request.scope, request.state);
-            µb.updateBadgeAsync(request.tabId);
+            µb.updateToolbarIcon(request.tabId, 0x03);
         }
         break;
 
@@ -452,56 +459,6 @@ vAPI.messaging.listen('popupPanel', onMessage);
 
 /******************************************************************************/
 
-var µb = µBlock;
-
-/******************************************************************************/
-
-var tagNameToRequestTypeMap = {
-     'embed': 'object',
-    'iframe': 'sub_frame',
-       'img': 'image',
-    'object': 'object'
-};
-
-/******************************************************************************/
-
-// Evaluate many requests.
-
-// https://github.com/gorhill/uBlock/issues/1782
-//   Treat `data:` URIs as 1st-party resources.
-
-var filterRequests = function(pageStore, details) {
-    var requests = details.requests;
-    if ( µb.userSettings.collapseBlocked === false ) {
-        return requests;
-    }
-
-    //console.debug('messaging.js/contentscript-end.js: processing %d requests', requests.length);
-
-    var hostnameFromURI = µb.URI.hostnameFromURI,
-        redirectEngine = µb.redirectEngine,
-        punycodeURL = vAPI.punycodeURL;
-
-    // Create evaluation context
-    var context = pageStore.createContextFromFrameHostname(details.pageHostname),
-        request,
-        i = requests.length;
-    while ( i-- ) {
-        request = requests[i];
-        context.requestURL = punycodeURL(request.url);
-        context.requestHostname = hostnameFromURI(context.requestURL);
-        context.requestType = tagNameToRequestTypeMap[request.tag];
-        if ( pageStore.filterRequest(context) !== 1 ) { continue; }
-        // Redirected? (We do not hide redirected resources.)
-        request.collapse = redirectEngine.matches(context) !== true;
-    }
-
-    context.dispose();
-    return requests;
-};
-
-/******************************************************************************/
-
 var onMessage = function(request, sender, callback) {
     // Async
     switch ( request.what ) {
@@ -510,42 +467,70 @@ var onMessage = function(request, sender, callback) {
     }
 
     // Sync
-    var response;
+    var µb = µBlock,
+        response,
+        tabId, frameId,
+        pageStore = null;
 
-    var pageStore;
     if ( sender && sender.tab ) {
-        pageStore = µb.pageStoreFromTabId(sender.tab.id);
+        tabId = sender.tab.id;
+        frameId = sender.frameId;
+        pageStore = µb.pageStoreFromTabId(tabId);
     }
 
     switch ( request.what ) {
+    case 'getCollapsibleBlockedRequests':
+        response = {
+            id: request.id,
+            hash: request.hash,
+            netSelectorCacheCountMax:
+                µb.cosmeticFilteringEngine.netSelectorCacheCountMax
+        };
+        if (
+            µb.userSettings.collapseBlocked &&
+            pageStore &&
+            pageStore.getNetFilteringSwitch()
+        ) {
+            pageStore.getBlockedResources(request, response);
+        }
+        break;
+
     case 'retrieveContentScriptParameters':
-        if ( pageStore && pageStore.getNetFilteringSwitch() ) {
-            response = {
-                loggerEnabled: µb.logger.isEnabled(),
-                collapseBlocked: µb.userSettings.collapseBlocked,
-                noCosmeticFiltering: µb.cosmeticFilteringEngine.acceptedCount === 0 || pageStore.noCosmeticFiltering === true,
-                noGenericCosmeticFiltering: pageStore.noGenericCosmeticFiltering === true
-            };
-            response.specificCosmeticFilters = µb.cosmeticFilteringEngine.retrieveDomainSelectors(
-                request,
-                response.noCosmeticFiltering
-            );
+        if (
+            pageStore === null ||
+            pageStore.getNetFilteringSwitch() === false ||
+            !request.url
+        ) {
+            break;
+        }
+        response = {
+            collapseBlocked: µb.userSettings.collapseBlocked,
+            noCosmeticFiltering: pageStore.noCosmeticFiltering === true,
+            noGenericCosmeticFiltering:
+                pageStore.noGenericCosmeticFiltering === true
+        };
+        request.tabId = tabId;
+        request.frameId = frameId;
+        request.hostname = µb.URI.hostnameFromURI(request.url);
+        request.domain = µb.URI.domainFromHostname(request.hostname);
+        request.entity = µb.URI.entityFromDomain(request.domain);
+        response.specificCosmeticFilters =
+            µb.cosmeticFilteringEngine.retrieveSpecificSelectors(request, response);
+        if ( µb.canInjectScriptletsNow === false ) {
+            response.scriptlets = µb.scriptletFilteringEngine.retrieve(request);
+        }
+        if ( response.noCosmeticFiltering !== true ) {
+            µb.logCosmeticFilters(tabId, frameId);
         }
         break;
 
     case 'retrieveGenericCosmeticSelectors':
         if ( pageStore && pageStore.getGenericCosmeticFilteringSwitch() ) {
+            request.tabId = tabId;
+            request.frameId = frameId;
             response = {
-                result: µb.cosmeticFilteringEngine.retrieveGenericSelectors(request)
-            };
-        }
-        break;
-
-    case 'filterRequests':
-        if ( pageStore && pageStore.getNetFilteringSwitch() ) {
-            response = {
-                result: filterRequests(pageStore, request),
-                netSelectorCacheCountMax: µb.cosmeticFilteringEngine.netSelectorCacheCountMax
+                result: µb.cosmeticFilteringEngine
+                          .retrieveGenericSelectors(request)
             };
         }
         break;
@@ -604,15 +589,14 @@ var onMessage = function(request, sender, callback) {
             callback({
                 frameContent: this.responseText.replace(reStrings, replacer),
                 target: µb.epickerTarget,
-                clientX: µb.mouseX,
-                clientY: µb.mouseY,
+                clientX: µb.mouseEventRegister.x,
+                clientY: µb.mouseEventRegister.y,
                 zap: µb.epickerZap,
                 eprom: µb.epickerEprom
             });
 
             µb.epickerTarget = '';
-            µb.mouseX = -1;
-            µb.mouseY = -1;
+            µb.mouseEventRegister.x = µb.mouseEventRegister.y = -1;
         };
         xhr.send();
         return;
@@ -742,20 +726,12 @@ var backupUserData = function(callback) {
         version: vAPI.app.version,
         userSettings: µb.userSettings,
         selectedFilterLists: µb.selectedFilterLists,
-        hiddenSettingsString: µb.stringFromHiddenSettings(),
+        hiddenSettings: µb.hiddenSettings,
         netWhitelist: µb.stringFromWhitelist(µb.netWhitelist),
         dynamicFilteringString: µb.permanentFirewall.toString(),
         urlFilteringString: µb.permanentURLFiltering.toString(),
         hostnameSwitchesString: µb.hnSwitches.toString(),
-        userFilters: '',
-        // TODO(seamless migration):
-        // The following is strictly for convenience, to be minimally
-        // forward-compatible. This will definitely be removed in the
-        // short term, as I do not expect the need to install an older
-        // version of uBO to ever be needed beyond the short term.
-        // >>>>>>>>
-        filterLists: µb.oldDataFromNewListKeys(µb.selectedFilterLists)
-        // <<<<<<<<
+        userFilters: ''
     };
 
     var onUserFiltersReady = function(details) {
@@ -784,8 +760,14 @@ var restoreUserData = function(request) {
     var onAllRemoved = function() {
         µBlock.saveLocalSettings();
         vAPI.storage.set(userData.userSettings);
-        µb.hiddenSettingsFromString(userData.hiddenSettingsString || '');
+        var hiddenSettings = userData.hiddenSettings;
+        if ( hiddenSettings instanceof Object === false ) {
+            hiddenSettings = µBlock.hiddenSettingsFromString(
+                userData.hiddenSettingsString || ''
+            );
+        }
         vAPI.storage.set({
+            hiddenSettings: hiddenSettings,
             netWhitelist: userData.netWhitelist || '',
             dynamicFilteringString: userData.dynamicFilteringString || '',
             urlFilteringString: userData.urlFilteringString || '',
@@ -796,17 +778,8 @@ var restoreUserData = function(request) {
             lastBackupTime: 0
         });
         µb.assets.put(µb.userFiltersPath, userData.userFilters);
-
-        // 'filterLists' is available up to uBO v1.10.4, not beyond.
-        // 'selectedFilterLists' is available from uBO v1.11 and beyond.
-        var listKeys;
         if ( Array.isArray(userData.selectedFilterLists) ) {
-            listKeys = userData.selectedFilterLists;
-        } else if ( userData.filterLists instanceof Object ) {
-            listKeys = µb.newListKeysFromOldData(userData.filterLists);
-        }
-        if ( listKeys !== undefined ) {
-            µb.saveSelectedFilterLists(listKeys, restart);
+            µb.saveSelectedFilterLists(userData.selectedFilterLists, restart);
         } else {
             restart();
         }
@@ -820,12 +793,13 @@ var restoreUserData = function(request) {
     // storage
     vAPI.cacheStorage.clear();
     vAPI.storage.clear(onAllRemoved);
+    vAPI.localStorage.removeItem('immediateHiddenSettings');
 };
 
 var resetUserData = function() {
     vAPI.cacheStorage.clear();
     vAPI.storage.clear();
-    vAPI.localStorage.removeItem('hiddenSettings');
+    vAPI.localStorage.removeItem('immediateHiddenSettings');
 
     // Keep global counts, people can become quite attached to numbers
     µb.saveLocalSettings();
@@ -864,8 +838,7 @@ var getLists = function(callback) {
         ignoreGenericCosmeticFilters: µb.userSettings.ignoreGenericCosmeticFilters,
         netFilterCount: µb.staticNetFilteringEngine.getFilterCount(),
         parseCosmeticFilters: µb.userSettings.parseAllABPHideFilters,
-        userFiltersPath: µb.userFiltersPath,
-        aliases: µb.assets.listKeyAliases
+        userFiltersPath: µb.userFiltersPath
     };
     var onMetadataReady = function(entries) {
         r.cache = entries;
@@ -886,46 +859,57 @@ var getLists = function(callback) {
 
 var getRules = function() {
     return {
-        permanentRules: µb.permanentFirewall.toString() + '\n' + µb.permanentURLFiltering.toString(),
-        sessionRules: µb.sessionFirewall.toString() + '\n' + µb.sessionURLFiltering.toString(),
-        hnSwitches: µb.hnSwitches.toString()
+        permanentRules: µb.permanentFirewall.toArray().concat(
+                            µb.permanentURLFiltering.toArray()
+                        ),
+          sessionRules: µb.sessionFirewall.toArray().concat(
+                            µb.sessionURLFiltering.toArray()
+                        ),
+            hnSwitches: µb.hnSwitches.toArray()
     };
 };
 
-// Untangle firewall rules, url rules and switches.
-var untangleRules = function(s) {
-    var textEnd = s.length;
-    var lineBeg = 0, lineEnd;
-    var line;
-    var firewallRules = [];
-    var urlRules = [];
-    var switches = [];
-
-    while ( lineBeg < textEnd ) {
-        lineEnd = s.indexOf('\n', lineBeg);
-        if ( lineEnd < 0 ) {
-            lineEnd = s.indexOf('\r', lineBeg);
-            if ( lineEnd < 0 ) {
-                lineEnd = textEnd;
-            }
+var modifyRuleset = function(details) {
+    var swRuleset = µb.hnSwitches,
+        hnRuleset, urlRuleset;
+    if ( details.permanent ) {
+        hnRuleset = µb.permanentFirewall;
+        urlRuleset = µb.permanentURLFiltering;
+    } else {
+        hnRuleset = µb.sessionFirewall;
+        urlRuleset = µb.sessionURLFiltering;
+    }
+    var toRemove = new Set(details.toRemove.trim().split(/\s*[\n\r]+\s*/));
+    var rule, parts, _;
+    for ( rule of toRemove ) {
+        if ( rule === '' ) { continue; }
+        parts = rule.split(/\s+/);
+        _ = hnRuleset.removeFromRuleParts(parts) ||
+            swRuleset.removeFromRuleParts(parts) ||
+            urlRuleset.removeFromRuleParts(parts);
+    }
+    var toAdd = new Set(details.toAdd.trim().split(/\s*[\n\r]+\s*/));
+    for ( rule of toAdd ) {
+        if ( rule === '' ) { continue; }
+        parts = rule.split(/\s+/);
+        _ = hnRuleset.addFromRuleParts(parts) ||
+            swRuleset.addFromRuleParts(parts) ||
+            urlRuleset.addFromRuleParts(parts);
+    }
+    if ( details.permanent ) {
+        if ( hnRuleset.changed ) {
+            µb.savePermanentFirewallRules();
+            hnRuleset.changed = false;
         }
-        line = s.slice(lineBeg, lineEnd).trim();
-        lineBeg = lineEnd + 1;
-
-        if ( line.indexOf('://') !== -1 ) {
-            urlRules.push(line);
-        } else if ( line.indexOf(':') === -1 ) {
-            firewallRules.push(line);
-        } else {
-            switches.push(line);
+        if ( urlRuleset.changed ) {
+            µb.savePermanentURLFilteringRules();
+            urlRuleset.changed = false;
         }
     }
-
-    return {
-        firewallRules: firewallRules.join('\n'),
-        urlRules: urlRules.join('\n'),
-        switches: switches.join('\n')
-    };
+    if ( swRuleset.changed ) {
+        µb.saveHostnameSwitches();
+        swRuleset.changed = false;
+    }
 };
 
 /******************************************************************************/
@@ -960,6 +944,13 @@ var onMessage = function(request, sender, callback) {
         response = getRules();
         break;
 
+    case 'modifyRuleset':
+        // https://github.com/chrisaljoudi/uBlock/issues/772
+        µb.cosmeticFilteringEngine.removeFromSelectorCache('*');
+        modifyRuleset(request);
+        response = getRules();
+        break;
+
     case 'purgeAllCaches':
         if ( request.hard ) {
             µb.assets.remove(/./);
@@ -974,6 +965,7 @@ var onMessage = function(request, sender, callback) {
         // https://github.com/gorhill/uBlock/pull/2314#issuecomment-278716960
         if ( request.assetKey === 'ublock-filters' ) {
             µb.assets.purge('ublock-resources');
+            µb.redirectEngine.invalidateResourcesSelfie();
         }
         break;
 
@@ -989,34 +981,8 @@ var onMessage = function(request, sender, callback) {
         resetUserData();
         break;
 
-    case 'setSessionRules':
-        // https://github.com/chrisaljoudi/uBlock/issues/772
-        µb.cosmeticFilteringEngine.removeFromSelectorCache('*');
-        response = untangleRules(request.rules);
-        µb.sessionFirewall.fromString(response.firewallRules);
-        µb.sessionURLFiltering.fromString(response.urlRules);
-        µb.hnSwitches.fromString(response.switches);
-        µb.saveHostnameSwitches();
-        response = getRules();
-        break;
-
-    case 'setPermanentRules':
-        response = untangleRules(request.rules);
-        µb.permanentFirewall.fromString(response.firewallRules);
-        µb.savePermanentFirewallRules();
-        µb.permanentURLFiltering.fromString(response.urlRules);
-        µb.savePermanentURLFilteringRules();
-        µb.hnSwitches.fromString(response.switches);
-        µb.saveHostnameSwitches();
-        response = getRules();
-        break;
-
-    case 'validateWhitelistString':
-        response = µb.validateWhitelistString(request.raw);
-        break;
-
     case 'writeHiddenSettings':
-        µb.hiddenSettingsFromString(request.content);
+        µb.changeHiddenSettings(µb.hiddenSettingsFromString(request.content));
         break;
 
     default:
@@ -1041,7 +1007,39 @@ vAPI.messaging.listen('dashboard', onMessage);
 
 /******************************************************************************/
 
-var µb = µBlock;
+var µb = µBlock,
+    extensionOriginURL = vAPI.getURL('');
+
+/******************************************************************************/
+
+var getLoggerData = function(details, activeTabId, callback) {
+    let response = {
+        colorBlind: µb.userSettings.colorBlindFriendly,
+        entries: µb.logger.readAll(details.ownerId),
+        maxEntries: µb.userSettings.requestLogMaxEntries,
+        activeTabId: activeTabId,
+        tabIdsToken: µb.pageStoresToken
+    };
+    if ( µb.pageStoresToken !== details.tabIdsToken ) {
+        let tabIds = new Map();
+        for ( let entry of µb.pageStores ) {
+            let pageStore = entry[1];
+            if ( pageStore.rawURL.startsWith(extensionOriginURL) ) { continue; }
+            tabIds.set(entry[0], pageStore.title);
+        }
+        response.tabIds = Array.from(tabIds);
+    }
+    if ( activeTabId ) {
+        let pageStore = µb.pageStoreFromTabId(activeTabId);
+        if (
+            pageStore === null ||
+            pageStore.rawURL.startsWith(extensionOriginURL)
+        ) {
+            response.activeTabId = undefined;
+        }
+    }
+    callback(response);
+};
 
 /******************************************************************************/
 
@@ -1079,6 +1077,19 @@ var getURLFilteringData = function(details) {
 var onMessage = function(request, sender, callback) {
     // Async
     switch ( request.what ) {
+    case 'readAll':
+        if (
+            µb.logger.ownerId !== undefined &&
+            µb.logger.ownerId !== request.ownerId
+        ) {
+            callback({ unavailable: true });
+            return;
+        }
+        vAPI.tabs.get(null, function(tab) {
+            getLoggerData(request, tab && tab.id, callback);
+        });
+        return;
+
     default:
         break;
     }
@@ -1087,27 +1098,10 @@ var onMessage = function(request, sender, callback) {
     var response;
 
     switch ( request.what ) {
-    case 'readAll':
-        var tabIds = {}, pageStore;
-        var loggerURL = vAPI.getURL('logger-ui.html');
-        for ( var tabId in µb.pageStores ) {
-            pageStore = µb.pageStoreFromTabId(tabId);
-            if ( pageStore === null ) {
-                continue;
-            }
-            if ( pageStore.rawURL.startsWith(loggerURL) ) {
-                continue;
-            }
-            tabIds[tabId] = pageStore.title;
+    case 'releaseView':
+        if ( request.ownerId === µb.logger.ownerId ) {
+            µb.logger.ownerId = undefined;
         }
-        response = {
-            colorBlind: µb.userSettings.colorBlindFriendly,
-            entries: µb.logger.readAll(),
-            maxEntries: µb.userSettings.requestLogMaxEntries,
-            noTabId: vAPI.noTabId,
-            tabIds: tabIds,
-            tabIdsToken: µb.pageStoresToken
-        };
         break;
 
     case 'saveURLFilteringRules':
@@ -1143,8 +1137,6 @@ vAPI.messaging.listen('loggerUI', onMessage);
 
 })();
 
-// https://www.youtube.com/watch?v=3_WcygKJP1k
-
 /******************************************************************************/
 /******************************************************************************/
 
