@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2014-2016 Raymond Hill
+    Copyright (C) 2014-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -43,57 +43,22 @@ var µb = µBlock;
 /******************************************************************************/
 
 // To mitigate memory churning
-var netFilteringResultCacheEntryJunkyard = [];
-var netFilteringResultCacheEntryJunkyardMax = 200;
-
-/******************************************************************************/
-
-var NetFilteringResultCacheEntry = function(result, type) {
-    this.init(result, type);
-};
-
-/******************************************************************************/
-
-NetFilteringResultCacheEntry.prototype.init = function(result, type) {
-    this.result = result;
-    this.type = type;
-    this.time = Date.now();
-    return this;
-};
-
-/******************************************************************************/
-
-NetFilteringResultCacheEntry.prototype.dispose = function() {
-    this.result = this.type = '';
-    if ( netFilteringResultCacheEntryJunkyard.length < netFilteringResultCacheEntryJunkyardMax ) {
-        netFilteringResultCacheEntryJunkyard.push(this);
-    }
-};
-
-/******************************************************************************/
-
-NetFilteringResultCacheEntry.factory = function(result, type) {
-    if ( netFilteringResultCacheEntryJunkyard.length ) {
-        return netFilteringResultCacheEntryJunkyard.pop().init(result, type);
-    }
-    return new NetFilteringResultCacheEntry(result, type);
-};
-
-/******************************************************************************/
-/******************************************************************************/
-
-// To mitigate memory churning
-var netFilteringCacheJunkyard = [];
-var netFilteringCacheJunkyardMax = 10;
+var netFilteringCacheJunkyard = [],
+    netFilteringCacheJunkyardMax = 10;
 
 /******************************************************************************/
 
 var NetFilteringResultCache = function() {
+    this.boundPruneAsyncCallback = this.pruneAsyncCallback.bind(this);
     this.init();
 };
 
 /******************************************************************************/
 
+NetFilteringResultCache.prototype.shelfLife = 15 * 1000;
+
+/******************************************************************************/
+
 NetFilteringResultCache.factory = function() {
     var entry = netFilteringCacheJunkyard.pop();
     if ( entry === undefined ) {
@@ -107,18 +72,16 @@ NetFilteringResultCache.factory = function() {
 /******************************************************************************/
 
 NetFilteringResultCache.prototype.init = function() {
-    this.urls = Object.create(null);
-    this.count = 0;
-    this.shelfLife = 15 * 1000;
+    this.blocked = new Map();
+    this.results = new Map();
+    this.hash = 0;
     this.timer = null;
-    this.boundPruneAsyncCallback = this.pruneAsyncCallback.bind(this);
 };
 
 /******************************************************************************/
 
 NetFilteringResultCache.prototype.dispose = function() {
     this.empty();
-    this.boundPruneAsyncCallback = null;
     if ( netFilteringCacheJunkyard.length < netFilteringCacheJunkyardMax ) {
         netFilteringCacheJunkyard.push(this);
     }
@@ -127,32 +90,42 @@ NetFilteringResultCache.prototype.dispose = function() {
 
 /******************************************************************************/
 
-NetFilteringResultCache.prototype.add = function(context, result) {
-    var url = context.requestURL,
-        type = context.requestType,
-        key = type + ' ' + url,
-        entry = this.urls[key];
-    if ( entry !== undefined ) {
-        entry.result = result;
-        entry.type = type;
-        entry.time = Date.now();
-        return;
-    }
-    this.urls[key] = NetFilteringResultCacheEntry.factory(result, type);
-    if ( this.count === 0 ) {
+NetFilteringResultCache.prototype.rememberResult = function(context, result, logData) {
+    if ( this.results.size === 0 ) {
         this.pruneAsync();
     }
-    this.count += 1;
+    var key = context.pageHostname + ' ' + context.requestType + ' ' + context.requestURL;
+    this.results.set(key, {
+        result: result,
+        logData: logData,
+        tstamp: Date.now()
+    });
+    if ( result !== 1 ) { return; }
+    var now = Date.now();
+    this.blocked.set(key, now);
+    this.hash = now;
+};
+
+/******************************************************************************/
+
+NetFilteringResultCache.prototype.rememberBlock = function(details) {
+    if ( this.blocked.size === 0 ) {
+        this.pruneAsync();
+    }
+    var now = Date.now();
+    this.blocked.set(
+        details.pageHostname + ' ' + details.requestType + ' ' + details.requestURL,
+        now
+    );
+    this.hash = now;
 };
 
 /******************************************************************************/
 
 NetFilteringResultCache.prototype.empty = function() {
-    for ( var key in this.urls ) {
-        this.urls[key].dispose();
-    }
-    this.urls = Object.create(null);
-    this.count = 0;
+    this.blocked.clear();
+    this.results.clear();
+    this.hash = 0;
     if ( this.timer !== null ) {
         clearTimeout(this.timer);
         this.timer = null;
@@ -161,36 +134,6 @@ NetFilteringResultCache.prototype.empty = function() {
 
 /******************************************************************************/
 
-NetFilteringResultCache.prototype.compareEntries = function(a, b) {
-    return this.urls[b].time - this.urls[a].time;
-};
-
-/******************************************************************************/
-
-NetFilteringResultCache.prototype.prune = function() {
-    var keys = Object.keys(this.urls).sort(this.compareEntries.bind(this));
-    var obsolete = Date.now() - this.shelfLife;
-    var key, entry;
-    var i = keys.length;
-    while ( i-- ) {
-        key = keys[i];
-        entry = this.urls[key];
-        if ( entry.time > obsolete ) {
-            break;
-        }
-        entry.dispose();
-        delete this.urls[key];
-    }
-    this.count -= keys.length - i - 1;
-    if ( this.count > 0 ) {
-        this.pruneAsync();
-    }
-};
-
-// https://www.youtube.com/watch?v=hcVpbsDyOhM
-
-/******************************************************************************/
-
 NetFilteringResultCache.prototype.pruneAsync = function() {
     if ( this.timer === null ) {
         this.timer = vAPI.setTimeout(this.boundPruneAsyncCallback, this.shelfLife * 2);
@@ -199,13 +142,46 @@ NetFilteringResultCache.prototype.pruneAsync = function() {
 
 NetFilteringResultCache.prototype.pruneAsyncCallback = function() {
     this.timer = null;
-    this.prune();
+    var obsolete = Date.now() - this.shelfLife,
+        entry;
+    for ( entry of this.blocked ) {
+        if ( entry[1] <= obsolete ) {
+            this.results.delete(entry[0]);
+            this.blocked.delete(entry[0]);
+        }
+    }
+    for ( entry of this.results ) {
+        if ( entry[1].tstamp <= obsolete ) {
+            this.results.delete(entry[0]);
+        }
+    }
+    if ( this.blocked.size !== 0 || this.results.size !== 0 ) {
+        this.pruneAsync();
+    }
 };
 
 /******************************************************************************/
 
-NetFilteringResultCache.prototype.lookup = function(context) {
-    return this.urls[context.requestType + ' ' + context.requestURL] || undefined;
+NetFilteringResultCache.prototype.lookupResult = function(context) {
+    return this.results.get(
+        context.pageHostname + ' ' +
+        context.requestType + ' ' +
+        context.requestURL
+    );
+};
+
+/******************************************************************************/
+
+NetFilteringResultCache.prototype.lookupAllBlocked = function(hostname) {
+    var result = [],
+        pos;
+    for ( var entry of this.blocked ) {
+        pos = entry[0].indexOf(' ');
+        if ( entry[0].slice(0, pos) === hostname ) {
+            result[result.length] = entry[0].slice(pos + 1);
+        }
+    }
+    return result;
 };
 
 /******************************************************************************/
@@ -264,8 +240,8 @@ var pageStoreJunkyardMax = 10;
 
 /******************************************************************************/
 
-var PageStore = function(tabId) {
-    this.init(tabId);
+var PageStore = function(tabId, context) {
+    this.init(tabId, context);
     this.journal = [];
     this.journalTimer = null;
     this.journalLastCommitted = this.journalLastUncommitted = undefined;
@@ -274,19 +250,23 @@ var PageStore = function(tabId) {
 
 /******************************************************************************/
 
-PageStore.factory = function(tabId) {
+PageStore.factory = function(tabId, context) {
     var entry = pageStoreJunkyard.pop();
     if ( entry === undefined ) {
-        entry = new PageStore(tabId);
+        entry = new PageStore(tabId, context);
     } else {
-        entry.init(tabId);
+        entry.init(tabId, context);
     }
     return entry;
 };
 
 /******************************************************************************/
 
-PageStore.prototype.init = function(tabId) {
+// https://github.com/gorhill/uBlock/issues/3201
+//   The context is used to determine whether we report behavior change to the
+//   logger.
+
+PageStore.prototype.init = function(tabId, context) {
     var tabContext = µb.tabContextManager.mustLookup(tabId);
     this.tabId = tabId;
 
@@ -304,7 +284,8 @@ PageStore.prototype.init = function(tabId) {
     this.rawURL = tabContext.rawURL;
     this.hostnameToCountMap = new Map();
     this.contentLastModified = 0;
-    this.frames = Object.create(null);
+    this.frames = new Map();
+    this.logData = undefined;
     this.perLoadBlockedRequestCount = 0;
     this.perLoadAllowedRequestCount = 0;
     this.hiddenElementCount = ''; // Empty string means "unknown"
@@ -315,12 +296,19 @@ PageStore.prototype.init = function(tabId) {
     this.netFilteringCache = NetFilteringResultCache.factory();
     this.internalRedirectionCount = 0;
 
-    this.noCosmeticFiltering = µb.hnSwitches.evaluateZ('no-cosmetic-filtering', tabContext.rootHostname) === true;
-    if ( this.noCosmeticFiltering && µb.logger.isEnabled() ) {
+    this.noCosmeticFiltering = µb.hnSwitches.evaluateZ(
+        'no-cosmetic-filtering',
+        tabContext.rootHostname
+    ) === true;
+    if (
+        this.noCosmeticFiltering &&
+        µb.logger.isEnabled() &&
+        context === 'tabCommitted'
+    ) {
         µb.logger.writeOne(
             tabId,
             'cosmetic',
-            µb.hnSwitches.toResultString(),
+            µb.hnSwitches.toLogData(),
             'dom',
             tabContext.rawURL,
             this.tabHostname,
@@ -336,12 +324,16 @@ PageStore.prototype.init = function(tabId) {
             tabContext.normalURL,
             'generichide'
         );
-        this.noGenericCosmeticFiltering = result === false;
-        if ( result !== undefined && µb.logger.isEnabled() ) {
+        this.noGenericCosmeticFiltering = result === 2;
+        if (
+            result !== 0 &&
+            µb.logger.isEnabled() &&
+            context === 'tabCommitted'
+        ) {
             µb.logger.writeOne(
                 tabId,
                 'net',
-                µb.staticNetFilteringEngine.toResultString(true),
+                µb.staticNetFilteringEngine.toLogData(),
                 'generichide',
                 tabContext.rawURL,
                 this.tabHostname,
@@ -383,7 +375,7 @@ PageStore.prototype.reuse = function(context) {
     }
     this.disposeFrameStores();
     this.netFilteringCache = this.netFilteringCache.dispose();
-    this.init(this.tabId);
+    this.init(this.tabId, context);
     return this;
 };
 
@@ -418,27 +410,26 @@ PageStore.prototype.dispose = function() {
 /******************************************************************************/
 
 PageStore.prototype.disposeFrameStores = function() {
-    var frames = this.frames;
-    for ( var k in frames ) {
-        frames[k].dispose();
+    for ( var frameStore of this.frames.values() ) {
+        frameStore.dispose();
     }
-    this.frames = Object.create(null);
+    this.frames.clear();
 };
 
 /******************************************************************************/
 
 PageStore.prototype.getFrame = function(frameId) {
-    return this.frames[frameId] || null;
+    return this.frames.get(frameId) || null;
 };
 
 /******************************************************************************/
 
 PageStore.prototype.setFrame = function(frameId, frameURL) {
-    var frameStore = this.frames[frameId];
-    if ( frameStore ) {
+    var frameStore = this.frames.get(frameId);
+    if ( frameStore !== undefined ) {
         frameStore.init(frameURL);
     } else {
-        this.frames[frameId] = FrameStore.factory(frameURL);
+        this.frames.set(frameId, FrameStore.factory(frameURL));
     }
 };
 
@@ -453,8 +444,8 @@ PageStore.prototype.createContextFromPage = function() {
 
 PageStore.prototype.createContextFromFrameId = function(frameId) {
     var context = µb.tabContextManager.createContext(this.tabId);
-    var frameStore = this.frames[frameId];
-    if ( frameStore ) {
+    var frameStore = this.frames.get(frameId);
+    if ( frameStore !== undefined ) {
         context.pageHostname = frameStore.pageHostname;
         context.pageDomain = frameStore.pageDomain;
     } else {
@@ -508,10 +499,10 @@ PageStore.prototype.injectLargeMediaElementScriptlet = function() {
     µb.contextMenu.update(this.tabId);
 };
 
-PageStore.prototype.temporarilyAllowLargeMediaElements = function() {
+PageStore.prototype.temporarilyAllowLargeMediaElements = function(state) {
     this.largeMediaCount = 0;
     µb.contextMenu.update(this.tabId);
-    this.allowLargeMediaElementsUntil = Date.now() + 86400000;
+    this.allowLargeMediaElementsUntil = state ? Date.now() + 86400000 : 0;
     µb.scriptlets.injectDeep(this.tabId, 'load-large-media-all');
 };
 
@@ -525,7 +516,7 @@ PageStore.prototype.journalAddRequest = function(hostname, result) {
     if ( hostname === '' ) { return; }
     this.journal.push(
         hostname,
-        result.charCodeAt(1) === 0x62 /* 'b' */ ? 0x00000001 : 0x00010000
+        result === 1 ? 0x00000001 : 0x00010000
     );
     if ( this.journalTimer === null ) {
         this.journalTimer = vAPI.setTimeout(this.journalProcess.bind(this, true), 1000);
@@ -585,7 +576,7 @@ PageStore.prototype.journalProcess = function(fromTimer) {
     // https://github.com/chrisaljoudi/uBlock/issues/905#issuecomment-76543649
     //   No point updating the badge if it's not being displayed.
     if ( (aggregateCounts & 0xFFFF) && µb.userSettings.showIconBadge ) {
-        µb.updateBadgeAsync(this.tabId);
+        µb.updateToolbarIcon(this.tabId, 0x02);
     }
 
     // Everything before pivot does not originate from current page -- we still
@@ -604,62 +595,132 @@ PageStore.prototype.journalProcess = function(fromTimer) {
 /******************************************************************************/
 
 PageStore.prototype.filterRequest = function(context) {
-    var requestType = context.requestType;
-
-    // We want to short-term cache filtering results of collapsible types,
-    // because they are likely to be reused, from network request handler and
-    // from content script handler.
-    if ( 'image media object sub_frame'.indexOf(requestType) === -1 ) {
-        return this.filterRequestNoCache(context);
-    }
+    this.logData = undefined;
 
     if ( this.getNetFilteringSwitch() === false ) {
-        this.netFilteringCache.add(context, '');
-        return '';
+        return 0;
     }
 
-    var entry = this.netFilteringCache.lookup(context);
-    if ( entry !== undefined ) {
-        return entry.result;
+    var requestType = context.requestType;
+
+    if ( requestType === 'csp_report' && this.filterCSPReport(context) === 1 ) {
+        return 1;
+    }
+
+    if ( requestType.endsWith('font') && this.filterFont(context) === 1 ) {
+        return 1;
+    }
+
+    var cacheableResult = this.cacheableResults[requestType] === true;
+
+    if ( cacheableResult ) {
+        var entry = this.netFilteringCache.lookupResult(context);
+        if ( entry !== undefined ) {
+            this.logData = entry.logData;
+            return entry.result;
+        }
     }
 
     // Dynamic URL filtering.
-    µb.sessionURLFiltering.evaluateZ(context.rootHostname, context.requestURL, requestType);
-    var result = µb.sessionURLFiltering.toFilterString();
+    var result = µb.sessionURLFiltering.evaluateZ(context.rootHostname, context.requestURL, requestType);
+    if ( result !== 0 && µb.logger.isEnabled() ) {
+        this.logData = µb.sessionURLFiltering.toLogData();
+    }
 
     // Dynamic hostname/type filtering.
-    if ( result === '' && µb.userSettings.advancedUserEnabled ) {
-        µb.sessionFirewall.evaluateCellZY( context.rootHostname, context.requestHostname, requestType);
-        if ( µb.sessionFirewall.mustBlockOrAllow() ) {
-            result = µb.sessionFirewall.toFilterString();
+    if ( result === 0 && µb.userSettings.advancedUserEnabled ) {
+        result = µb.sessionFirewall.evaluateCellZY(context.rootHostname, context.requestHostname, requestType);
+        if ( result !== 0 && result !== 3 && µb.logger.isEnabled() ) {
+            this.logData = µb.sessionFirewall.toLogData();
         }
     }
 
-    // Static filtering: lowest filtering precedence.
-    if ( result === '' || result.charCodeAt(1) === 110 /* 'n' */ ) {
-        if ( µb.staticNetFilteringEngine.matchString(context) !== undefined ) {
-            result = µb.staticNetFilteringEngine.toResultString(µb.logger.isEnabled());
+    // Static filtering has lowest precedence.
+    if ( result === 0 || result === 3 ) {
+        result = µb.staticNetFilteringEngine.matchString(context);
+        if ( result !== 0 && µb.logger.isEnabled() ) {
+            this.logData = µb.staticNetFilteringEngine.toLogData();
         }
     }
 
-    this.netFilteringCache.add(context, result);
+    if ( cacheableResult ) {
+        this.netFilteringCache.rememberResult(context, result, this.logData);
+    } else if ( result === 1 && this.collapsibleResources[requestType] === true ) {
+        this.netFilteringCache.rememberBlock(context, true);
+    }
 
     return result;
 };
 
+PageStore.prototype.cacheableResults = {
+    sub_frame: true
+};
+
+PageStore.prototype.collapsibleResources = {
+    image: true,
+    media: true,
+    object: true,
+    sub_frame: true
+};
+
+/******************************************************************************/
+
+PageStore.prototype.filterCSPReport = function(context) {
+    if ( µb.hnSwitches.evaluateZ('no-csp-reports', context.requestHostname) ) {
+        if ( µb.logger.isEnabled() ) {
+            this.logData = µb.hnSwitches.toLogData();
+        }
+        return 1;
+    }
+    // https://github.com/gorhill/uBlock/issues/3140
+    //   Special handling of CSP reports if and only if these can't be filtered
+    //   natively.
+    if (
+        vAPI.net.nativeCSPReportFiltering !== true &&
+        this.internalRedirectionCount !== 0
+    ) {
+        if ( µb.logger.isEnabled() ) {
+            this.logData = {
+                result: 1,
+                source: 'global',
+                raw: 'no-spurious-csp-report'
+            };
+        }
+        return 1;
+    }
+    return 0;
+};
+
+/******************************************************************************/
+
+PageStore.prototype.filterFont = function(context) {
+    if ( context.requestType === 'font' ) {
+        this.remoteFontCount += 1;
+    }
+    if ( µb.hnSwitches.evaluateZ('no-remote-fonts', context.rootHostname) !== false ) {
+        if ( µb.logger.isEnabled() ) {
+            this.logData = µb.hnSwitches.toLogData();
+        }
+        return 1;
+    }
+    return 0;
+};
+
 /******************************************************************************/
 
 // The caller is responsible to check whether filtering is enabled or not.
 
 PageStore.prototype.filterLargeMediaElement = function(size) {
+    this.logData = undefined;
+
     if ( Date.now() < this.allowLargeMediaElementsUntil ) {
-        return;
+        return 0;
     }
     if ( µb.hnSwitches.evaluateZ('no-large-media', this.tabHostname) !== true ) {
-        return;
+        return 0;
     }
     if ( (size >>> 10) < µb.userSettings.largeMediaSize ) {
-        return;
+        return 0;
     }
 
     this.largeMediaCount += 1;
@@ -670,58 +731,42 @@ PageStore.prototype.filterLargeMediaElement = function(size) {
         );
     }
 
-    return µb.hnSwitches.toResultString();
-};
-
-/******************************************************************************/
-
-PageStore.prototype.filterRequestNoCache = function(context) {
-    if ( this.getNetFilteringSwitch() === false ) {
-        return '';
+    if ( µb.logger.isEnabled() ) {
+        this.logData = µb.hnSwitches.toLogData();
     }
 
-    var requestType = context.requestType,
-        result = '';
-
-    if ( requestType === 'csp_report' ) {
-        if ( this.internalRedirectionCount !== 0 ) {
-            result = 'gb:no-spurious-csp-report';
-        }
-    } else if ( requestType === 'font' ) {
-        if ( µb.hnSwitches.evaluateZ('no-remote-fonts', context.rootHostname) !== false ) {
-            result = µb.hnSwitches.toResultString();
-        }
-        this.remoteFontCount += 1;
-    }
-
-    // Dynamic URL filtering.
-    if ( result === '' ) {
-        µb.sessionURLFiltering.evaluateZ(context.rootHostname, context.requestURL, requestType);
-        result = µb.sessionURLFiltering.toFilterString();
-    }
-
-    // Dynamic hostname/type filtering.
-    if ( result === '' && µb.userSettings.advancedUserEnabled ) {
-        µb.sessionFirewall.evaluateCellZY(context.rootHostname, context.requestHostname, requestType);
-        if ( µb.sessionFirewall.mustBlockOrAllow() ) {
-            result = µb.sessionFirewall.toFilterString();
-        }
-    }
-
-    // Static filtering has lowest precedence.
-    if ( result === '' || result.charCodeAt(1) === 110 /* 'n' */ ) {
-        if ( µb.staticNetFilteringEngine.matchString(context) !== undefined ) {
-            result = µb.staticNetFilteringEngine.toResultString(µb.logger.isEnabled());
-        }
-    }
-
-    return result;
+    return 1;
 };
 
 // https://www.youtube.com/watch?v=drW8p_dTLD4
 
 /******************************************************************************/
 
+PageStore.prototype.getBlockedResources = function(request, response) {
+    var µburi = µb.URI,
+        normalURL = µb.normalizePageURL(this.tabId, request.frameURL),
+        frameHostname = µburi.hostnameFromURI(normalURL),
+        resources = request.resources;
+    // Force some resources to go through the filtering engine in order to
+    // populate the blocked-resources cache. This is required because for
+    // some resources it's not possible to detect whether they were blocked
+    // content script-side (i.e. `iframes` -- unlike `img`).
+    if ( Array.isArray(resources) && resources.length !== 0 ) {
+        var context = this.createContextFromFrameHostname(frameHostname);
+        for ( var resource of resources ) {
+            context.requestType = resource.type;
+            context.requestHostname = µburi.hostnameFromURI(resource.url);
+            context.requestURL = resource.url;
+            this.filterRequest(context);
+        }
+    }
+    if ( this.netFilteringCache.hash === response.hash ) { return; }
+    response.hash = this.netFilteringCache.hash;
+    response.blockedResources = this.netFilteringCache.lookupAllBlocked(frameHostname);
+};
+
+/******************************************************************************/
+
 return {
     factory: PageStore.factory
 };
