@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2014-2017 Raymond Hill
+    Copyright (C) 2014-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -34,41 +34,24 @@ if ( typeof popupFontSize === 'string' && popupFontSize !== 'unset' ) {
     document.body.style.setProperty('font-size', popupFontSize);
 }
 
-// Ensure the popup is properly sized as soon as possible. It is assume the DOM
-// content is ready at this point, which should be the case given where this
-// script file is included in the HTML file.
-
-var dfPaneVisibleStored = vAPI.localStorage.getItem('popupFirewallPane') === 'true';
-
-// No restriction on vertical size?
-if ( /[\?&]fullsize=1/.test(window.location.search) ) {
-    document.body.classList.add('fullsize');
+// https://github.com/gorhill/uBlock/issues/3032
+// Popup panel can be in one of two modes:
+// - not responsive: viewport is expected to adjust to popup panel size
+// - responsive: popup panel must adjust to viewport size -- this happens
+//   when the viewport is not resized by the browser to perfectly fits uBO's
+//   popup panel.
+if (
+    vAPI.webextFlavor.soup.has('mobile') ||
+    /[\?&]responsive=1/.test(window.location.search)
+) {
+    document.body.classList.add('responsive');
 }
 
-// Mobile device?
-if ( /[\?&]mobile=1/.test(window.location.search) ) {
-    document.body.classList.add('mobile');
-}
-
-// The padlock/eraser must be manually positioned:
-// - Its vertical position depends on the height of the popup title bar
-// - Its horizontal position depends on whether there is a vertical scrollbar.
-document.getElementById('rulesetTools').style.setProperty(
-    'top',
-    (document.getElementById('appinfo').getBoundingClientRect().bottom + 3) + 'px'
-);
-
-var positionRulesetTools = function() {
-    document.getElementById('rulesetTools').style.setProperty(
-        'left',
-        (document.getElementById('firewallContainer').getBoundingClientRect().left + 3) + 'px'
-    );
-};
-
 // https://github.com/chrisaljoudi/uBlock/issues/996
-// Experimental: mitigate glitchy popup UI: immediately set the firewall pane
-// visibility to its last known state. By default the pane is hidden.
-// Will remove if it makes no difference.
+// Experimental: mitigate glitchy popup UI: immediately set the firewall
+// pane visibility to its last known state. By default the pane is hidden.
+let dfPaneVisibleStored =
+    vAPI.localStorage.getItem('popupFirewallPane') === 'true';
 if ( dfPaneVisibleStored ) {
     document.getElementById('panes').classList.add('dfEnabled');
 }
@@ -84,7 +67,7 @@ var scopeToSrcHostnameMap = {
     '.': ''
 };
 var dfHotspots = null;
-var hostnameToSortableTokenMap = {};
+var hostnameToSortableTokenMap = new Map();
 var allDomains = {};
 var allDomainCount = 0;
 var allHostnameRows = [];
@@ -107,33 +90,51 @@ var reCyrillicAmbiguous = /[\u042c\u0430\u0433\u0435\u043e\u043f\u0440\u0441\u04
 
 /******************************************************************************/
 
+// The padlock/eraser must be manually positioned:
+// - Its vertical position depends on the height of the popup title bar
+// - Its horizontal position depends on whether there is a vertical scrollbar.
+
+var positionRulesetTools = function() {
+    var vpos = document.getElementById('appinfo')
+                       .getBoundingClientRect()
+                       .bottom + window.scrollY + 3;
+    var hpos = document.getElementById('firewallContainer')
+                       .getBoundingClientRect()
+                       .left + window.scrollX + 3;
+    var style = document.getElementById('rulesetTools').style;
+    style.setProperty('top', (vpos >>> 0) + 'px');
+    style.setProperty('left', (hpos >>> 0) + 'px');
+};
+
+/******************************************************************************/
+
 var cachePopupData = function(data) {
     popupData = {};
     scopeToSrcHostnameMap['.'] = '';
-    hostnameToSortableTokenMap = {};
+    hostnameToSortableTokenMap.clear();
 
     if ( typeof data !== 'object' ) {
         return popupData;
     }
     popupData = data;
     scopeToSrcHostnameMap['.'] = popupData.pageHostname || '';
-    var hostnameDict = popupData.hostnameDict;
+    let hostnameDict = popupData.hostnameDict;
     if ( typeof hostnameDict !== 'object' ) {
         return popupData;
     }
-    var domain, prefix;
-    for ( var hostname in hostnameDict ) {
-        if ( hostnameDict.hasOwnProperty(hostname) === false ) {
-            continue;
-        }
-        domain = hostnameDict[hostname].domain;
-        prefix = hostname.slice(0, 0 - domain.length);
+    for ( let hostname in hostnameDict ) {
+        if ( hostnameDict.hasOwnProperty(hostname) === false ) { continue; }
+        let domain = hostnameDict[hostname].domain;
+        let prefix = hostname.slice(0, 0 - domain.length - 1);
         // Prefix with space char for 1st-party hostnames: this ensure these
         // will come first in list.
         if ( domain === popupData.pageDomain ) {
             domain = '\u0020';
         }
-        hostnameToSortableTokenMap[hostname] = domain + prefix.split('.').reverse().join('.');
+        hostnameToSortableTokenMap.set(
+            hostname,
+            domain + ' ' + prefix.split('.').reverse().join('.')
+        );
     }
     return popupData;
 };
@@ -179,11 +180,11 @@ var formatNumber = function(count) {
 var rulekeyCompare = function(a, b) {
     var ha = a.slice(2, a.indexOf(' ', 2));
     if ( !reIP.test(ha) ) {
-        ha = hostnameToSortableTokenMap[ha] || ' ';
+        ha = hostnameToSortableTokenMap.get(ha) || ' ';
     }
     var hb = b.slice(2, b.indexOf(' ', 2));
     if ( !reIP.test(hb) ) {
-        hb = hostnameToSortableTokenMap[hb] || ' ';
+        hb = hostnameToSortableTokenMap.get(hb) || ' ';
     }
     var ca = ha.charCodeAt(0),
         cb = hb.charCodeAt(0);
@@ -309,12 +310,11 @@ var updateAllFirewallCells = function() {
         );
     }
 
-    positionRulesetTools();
-
-    uDom.nodeFromId('firewallContainer').classList.toggle(
-        'dirty',
-        popupData.matrixIsDirty === true
-    );
+    var dirty = popupData.matrixIsDirty === true;
+    if ( dirty ) {
+        positionRulesetTools();
+    }
+    uDom.nodeFromId('firewallContainer').classList.toggle('dirty', dirty);
 };
 
 /******************************************************************************/
@@ -393,25 +393,24 @@ var renderPrivacyExposure = function() {
 // Assume everything has to be done incrementally.
 
 var renderPopup = function() {
+    var elem, text;
+
     if ( popupData.tabTitle ) {
         document.title = popupData.appName + ' - ' + popupData.tabTitle;
     }
 
-    uDom('body')
-        .toggleClass('advancedUser', popupData.advancedUserEnabled)
-        .toggleClass(
-            'off',
-            (popupData.pageURL === '') ||
-            (!popupData.netFilteringSwitch) ||
-            (popupData.pageHostname === 'behind-the-scene' && !popupData.advancedUserEnabled)
-        );
+    elem = document.body;
+    elem.classList.toggle('advancedUser', popupData.advancedUserEnabled);
+    elem.classList.toggle(
+        'off',
+        popupData.pageURL === '' || !popupData.netFilteringSwitch
+    );
 
     // If you think the `=== true` is pointless, you are mistaken
     uDom.nodeFromId('gotoPick').classList.toggle('enabled', popupData.canElementPicker === true);
     uDom.nodeFromId('gotoZap').classList.toggle('enabled', popupData.canElementPicker === true);
 
-    var text,
-        blocked = popupData.pageBlockedRequestCount,
+    var blocked = popupData.pageBlockedRequestCount,
         total = popupData.pageAllowedRequestCount + blocked;
     if ( total === 0 ) {
         text = formatNumber(0);
@@ -431,15 +430,6 @@ var renderPopup = function() {
     }
     uDom.nodeFromId('total-blocked').textContent = text;
 
-    // https://github.com/gorhill/uBlock/issues/507
-    // Convenience: open the logger with current tab automatically selected
-    if ( popupData.tabId ) {
-        uDom.nodeFromSelector('#basicTools > a[href^="logger-ui.html"]').setAttribute(
-            'href',
-            'logger-ui.html#tab_' + popupData.tabId
-        );
-    }
-
     // This will collate all domains, touched or not
     renderPrivacyExposure();
 
@@ -478,18 +468,82 @@ var renderPopup = function() {
     }
 
     uDom.nodeFromId('panes').classList.toggle('dfEnabled', dfPaneVisible);
-    uDom('#firewallContainer')
-        .toggleClass('minimized', popupData.firewallPaneMinimized)
-        .toggleClass('colorBlind', popupData.colorBlindFriendly);
+
+    elem = uDom.nodeFromId('firewallContainer');
+    elem.classList.toggle('minimized', popupData.firewallPaneMinimized);
+    elem.classList.toggle('colorBlind', popupData.colorBlindFriendly);
 
     // Build dynamic filtering pane only if in use
     if ( dfPaneVisible ) {
         buildAllFirewallRows();
     }
+
+    renderTooltips();
 };
 
 /******************************************************************************/
 
+// https://github.com/gorhill/uBlock/issues/2889
+//   Use tooltip for ARIA purpose.
+
+var renderTooltips = function(selector) {
+    var elem, text;
+    for ( var entry of tooltipTargetSelectors ) {
+        if ( selector !== undefined && entry[0] !== selector ) { continue; }
+        text = vAPI.i18n(
+            entry[1].i18n +
+            (uDom.nodeFromSelector(entry[1].state) === null ? '1' : '2')
+        );
+        elem = uDom.nodeFromSelector(entry[0]);
+        elem.setAttribute('aria-label', text);
+        elem.setAttribute('data-tip', text);
+        if ( selector !== undefined ) {
+            uDom.nodeFromId('tooltip').textContent =
+                elem.getAttribute('data-tip');
+        }
+    }
+};
+
+var tooltipTargetSelectors = new Map([
+    [
+        '#switch',
+        {
+            state: 'body.off',
+            i18n: 'popupPowerSwitchInfo',
+        }
+    ],
+    [
+        '#no-popups',
+        {
+            state: '#no-popups.on',
+            i18n: 'popupTipNoPopups'
+        }
+    ],
+    [
+        '#no-large-media',
+        {
+            state: '#no-large-media.on',
+            i18n: 'popupTipNoLargeMedia'
+        }
+    ],
+    [
+        '#no-cosmetic-filtering',
+        {
+            state: '#no-cosmetic-filtering.on',
+            i18n: 'popupTipNoCosmeticFiltering'
+        }
+    ],
+    [
+        '#no-remote-fonts',
+        {
+            state: '#no-remote-fonts.on',
+            i18n: 'popupTipNoRemoteFonts'
+        }
+    ]
+]);
+
+/******************************************************************************/
+
 // All rendering code which need to be executed only once.
 
 var renderOnce = function() {
@@ -509,43 +563,63 @@ var renderOnce = function() {
     uDom.nodeFromId('appname').textContent = popupData.appName;
     uDom.nodeFromId('version').textContent = popupData.appVersion;
 
+    // https://github.com/uBlockOrigin/uBlock-issues/issues/22
+    if ( popupData.advancedUserEnabled !== true ) {
+        uDom('#firewallContainer [data-i18n-tip][data-src]').removeAttr('data-tip');
+    }
+
+    // https://github.com/gorhill/uBlock/issues/2274
+    //   Make use of the whole viewport when in responsive mode.
+    if ( document.body.classList.contains('responsive') ) { return; }
+
     // For large displays: we do not want the left pane -- optional and
     // hidden by defaut -- to dictate the height of the popup. The right pane
     // dictates the height of the popup, and the left pane will have a
     // scrollbar if ever its height is more than what is available.
     // For small displays: we use the whole viewport.
 
-    var rpane = uDom.nodeFromSelector('#panes > div:first-of-type'),
+    let rpane = uDom.nodeFromSelector('#panes > div:first-of-type'),
         lpane = uDom.nodeFromSelector('#panes > div:last-of-type');
 
-    var fillViewport = function() {
-        lpane.style.setProperty(
-            'height',
-            Math.max(
-                window.innerHeight - uDom.nodeFromSelector('#appinfo').offsetHeight,
-                rpane.offsetHeight
-            ) + 'px'
-        );
-        lpane.style.setProperty('width', (window.innerWidth - rpane.offsetWidth) + 'px');
+    lpane.style.setProperty('height', rpane.offsetHeight + 'px');
+
+    // Be prepared to fall into responsive mode if ever it is found the
+    // viewport is not a perfect match for the popup panel.
+
+    let resizeTimer;
+    let resize = function() {
+        resizeTimer = undefined;
+        // Do not use equality, fractional pixel dimension occurs and must
+        // be ignored.
+        // https://www.reddit.com/r/uBlockOrigin/comments/8qodpw/how_to_hide_the_info_shown_of_what_is_currently/e0lglrr/
+        //   Tolerance of 2px fixes the issue.
+        if (
+            Math.abs(document.body.offsetWidth - window.innerWidth) <= 2 &&
+            Math.abs(document.body.offsetHeight - window.innerHeight) <= 2
+        ) {
+            return;
+        }
+        document.body.classList.add('responsive');
+        lpane.style.removeProperty('height');
+        window.removeEventListener('resize', resizeAsync);
     };
-
-    // https://github.com/gorhill/uBlock/issues/2274
-    //   Make use of the whole viewport on mobile devices.
-    if ( document.body.classList.contains('mobile') ) {
-        fillViewport();
-        window.addEventListener('resize', fillViewport);
-        return;
-    }
-
-    if ( document.body.classList.contains('fullsize') === false ) {
-        lpane.style.setProperty('height', rpane.offsetHeight + 'px');
-    }
+    let resizeAsync = function() {
+        if ( resizeTimer !== undefined ) {
+            clearTimeout(resizeTimer);
+        }
+        resizeTimer = vAPI.setTimeout(resize, 67);
+    };
+    window.addEventListener('resize', resizeAsync);
+    resizeAsync();
 };
 
 /******************************************************************************/
 
 var renderPopupLazy = function() {
-    messaging.send('popupPanel', { what: 'getPopupLazyData', tabId: popupData.tabId });
+    messaging.send(
+        'popupPanel',
+        { what: 'getPopupLazyData', tabId: popupData.tabId }
+    );
 };
 
 var onPopupMessage = function(data) {
@@ -566,12 +640,7 @@ messaging.addChannelListener('popup', onPopupMessage);
 /******************************************************************************/
 
 var toggleNetFilteringSwitch = function(ev) {
-    if ( !popupData || !popupData.pageURL ) {
-        return;
-    }
-    if ( popupData.pageHostname === 'behind-the-scene' && !popupData.advancedUserEnabled ) {
-        return;
-    }
+    if ( !popupData || !popupData.pageURL ) { return; }
     messaging.send(
         'popupPanel',
         {
@@ -582,7 +651,7 @@ var toggleNetFilteringSwitch = function(ev) {
             tabId: popupData.tabId
         }
     );
-
+    renderTooltips('#switch');
     hashFromPopupData();
 };
 
@@ -618,18 +687,24 @@ var gotoPick = function() {
 /******************************************************************************/
 
 var gotoURL = function(ev) {
-    if ( this.hasAttribute('href') === false) {
-        return;
-    }
+    if ( this.hasAttribute('href') === false ) { return; }
 
     ev.preventDefault();
 
+    let url = this.getAttribute('href');
+    if (
+        url === 'logger-ui.html#tab_active' &&
+        typeof popupData.tabId === 'number'
+    ) {
+        url += '+' + popupData.tabId;
+    }
+
     messaging.send(
         'popupPanel',
         {
             what: 'gotoURL',
             details: {
-                url: this.getAttribute('href'),
+                url: url,
                 select: true,
                 index: -1,
                 shiftKey: ev.shiftKey
@@ -751,13 +826,14 @@ var setFirewallRuleHandler = function(ev) {
 
 /******************************************************************************/
 
-var reloadTab = function() {
+var reloadTab = function(ev) {
     messaging.send(
         'popupPanel',
         {
             what: 'reloadTab',
             tabId: popupData.tabId,
-            select: true
+            select: true,
+            bypassCache: ev.ctrlKey || ev.metaKey || ev.shiftKey
         }
     );
 
@@ -784,7 +860,7 @@ var toggleMinimize = function(ev) {
             {
                 what: 'gotoURL',
                 details: {
-                    url: 'popup.html?tabId=' + popupData.tabId + '&fullsize=1',
+                    url: 'popup.html?tabId=' + popupData.tabId + '&responsive=1',
                     select: true,
                     index: -1
                 }
@@ -848,9 +924,7 @@ var revertFirewallRules = function() {
 var toggleHostnameSwitch = function(ev) {
     var target = ev.currentTarget;
     var switchName = target.getAttribute('id');
-    if ( !switchName ) {
-        return;
-    }
+    if ( !switchName ) { return; }
     target.classList.toggle('on');
     messaging.send(
         'popupPanel',
@@ -862,6 +936,7 @@ var toggleHostnameSwitch = function(ev) {
             tabId: popupData.tabId
         }
     );
+    renderTooltips('#' + switchName);
     hashFromPopupData();
 };
 
@@ -987,31 +1062,31 @@ var onHideTooltip = function() {
 (function() {
     // If there's no tab id specified in the query string,
     // it will default to current tab.
-    var tabId = null;
+    let tabId = null;
 
     // Extract the tab id of the page this popup is for
-    var matches = window.location.search.match(/[\?&]tabId=([^&]+)/);
+    let matches = window.location.search.match(/[\?&]tabId=([^&]+)/);
     if ( matches && matches.length === 2 ) {
-        tabId = matches[1];
+        tabId = parseInt(matches[1], 10) || 0;
     }
     getPopupData(tabId);
-
-    uDom('#switch').on('click', toggleNetFilteringSwitch);
-    uDom('#gotoZap').on('click', gotoZap);
-    uDom('#gotoPick').on('click', gotoPick);
-    uDom('h2').on('click', toggleFirewallPane);
-    uDom('#refresh').on('click', reloadTab);
-    uDom('.hnSwitch').on('click', toggleHostnameSwitch);
-    uDom('#saveRules').on('click', saveFirewallRules);
-    uDom('#revertRules').on('click', revertFirewallRules);
-    uDom('[data-i18n="popupAnyRulePrompt"]').on('click', toggleMinimize);
-
-    uDom('body').on('mouseenter', '[data-tip]', onShowTooltip)
-                .on('mouseleave', '[data-tip]', onHideTooltip);
-
-    uDom('a[href]').on('click', gotoURL);
 })();
 
+uDom('#switch').on('click', toggleNetFilteringSwitch);
+uDom('#gotoZap').on('click', gotoZap);
+uDom('#gotoPick').on('click', gotoPick);
+uDom('h2').on('click', toggleFirewallPane);
+uDom('#refresh').on('click', reloadTab);
+uDom('.hnSwitch').on('click', toggleHostnameSwitch);
+uDom('#saveRules').on('click', saveFirewallRules);
+uDom('#revertRules').on('click', revertFirewallRules);
+uDom('[data-i18n="popupAnyRulePrompt"]').on('click', toggleMinimize);
+
+uDom('body').on('mouseenter', '[data-tip]', onShowTooltip)
+            .on('mouseleave', '[data-tip]', onHideTooltip);
+
+uDom('a[href]').on('click', gotoURL);
+
 /******************************************************************************/
 
 })();
