@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2014-2016 Raymond Hill
+    Copyright (C) 2014-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -118,12 +118,11 @@
 
 /******************************************************************************/
 
-if ( typeof vAPI !== 'object' ) {
-    return;
-}
-
-// don't run in frames
-if ( window.top !== window ) {
+if (
+    window.top !== window ||
+    typeof vAPI !== 'object' ||
+    vAPI.domFilterer instanceof Object === false
+) {
     return;
 }
 
@@ -131,8 +130,8 @@ var pickerRoot = document.getElementById(vAPI.sessionId);
 if ( pickerRoot ) {
     return;
 }
+
 var pickerBody = null;
-var pickerStyle = null;
 var svgOcean = null;
 var svgIslands = null;
 var svgRoot = null;
@@ -463,27 +462,21 @@ var filterTypes = {
 // Extract the best possible cosmetic filter, i.e. as specific as possible.
 
 // https://github.com/gorhill/uBlock/issues/1725
-// Also take into account the `src` attribute for `img` elements -- and limit
-// the value to the 1024 first characters.
+//   Also take into account the `src` attribute for `img` elements -- and limit
+//   the value to the 1024 first characters.
 
 var cosmeticFilterFromElement = function(elem) {
-    if ( elem === null ) {
-        return 0;
-    }
-    if ( elem.nodeType !== 1 ) {
-        return 0;
-    }
+    if ( elem === null ) { return 0; }
+    if ( elem.nodeType !== 1 ) { return 0; }
 
     if ( candidateElements.indexOf(elem) === -1 ) {
         candidateElements.push(elem);
     }
 
-    var tagName = elem.localName;
-    var selector = '';
-    var v, i;
+    let selector = '';
 
     // Id
-    v = typeof elem.id === 'string' && CSS.escape(elem.id);
+    let v = typeof elem.id === 'string' && CSS.escape(elem.id);
     if ( v ) {
         selector = '#' + v;
     }
@@ -491,18 +484,20 @@ var cosmeticFilterFromElement = function(elem) {
     // Class(es)
     v = elem.classList;
     if ( v ) {
-        i = v.length || 0;
+        let i = v.length || 0;
         while ( i-- ) {
             selector += '.' + CSS.escape(v.item(i));
         }
     }
 
     // Tag name
+    let tagName = elem.localName;
+
+    // Use attributes if still no selector found.
     // https://github.com/gorhill/uBlock/issues/1901
-    // Trim attribute value, this may help in case of malformed HTML.
+    //   Trim attribute value, this may help in case of malformed HTML.
     if ( selector === '' ) {
-        selector = tagName;
-        var attributes = [], attr;
+        let attributes = [], attr;
         switch ( tagName ) {
         case 'a':
             v = elem.getAttribute('href');
@@ -530,13 +525,11 @@ var cosmeticFilterFromElement = function(elem) {
             break;
         }
         while ( (attr = attributes.pop()) ) {
-            if ( attr.v.length === 0 ) {
-                continue;
-            }
+            if ( attr.v.length === 0 ) { continue; }
             v = elem.getAttribute(attr.k);
             if ( attr.v === v ) {
                 selector += '[' + attr.k + '="' + attr.v + '"]';
-            } else if ( v.lastIndexOf(attr.v, 0) === 0 ) {
+            } else if ( v.startsWith(attr.v) ) {
                 selector += '[' + attr.k + '^="' + attr.v + '"]';
             } else {
                 selector += '[' + attr.k + '*="' + attr.v + '"]';
@@ -544,19 +537,31 @@ var cosmeticFilterFromElement = function(elem) {
         }
     }
 
+    // https://github.com/uBlockOrigin/uBlock-issues/issues/17
+    //   If selector is ambiguous at this point, add the element name to
+    //   further narrow it down.
+    let parentNode = elem.parentNode;
+    if (
+        selector === '' ||
+        safeQuerySelectorAll(parentNode, cssScope + selector).length > 1
+    ) {
+        selector = tagName + selector;
+    }
+
     // https://github.com/chrisaljoudi/uBlock/issues/637
-    // If the selector is still ambiguous at this point, further narrow using
-    // `nth-of-type`. It is preferable to use `nth-of-type` as opposed to
-    // `nth-child`, as `nth-of-type` is less volatile.
-    var parentNode = elem.parentNode;
+    //   If the selector is still ambiguous at this point, further narrow using
+    //   `nth-of-type`. It is preferable to use `nth-of-type` as opposed to
+    //   `nth-child`, as `nth-of-type` is less volatile.
     if ( safeQuerySelectorAll(parentNode, cssScope + selector).length > 1 ) {
-        i = 1;
+        let i = 1;
         while ( elem.previousSibling !== null ) {
             elem = elem.previousSibling;
-            if ( typeof elem.localName !== 'string' || elem.localName !== tagName ) {
-                continue;
+            if (
+                typeof elem.localName === 'string' &&
+                elem.localName === tagName
+            ) {
+                i++;
             }
-            i++;
         }
         selector += ':nth-of-type(' + i + ')';
     }
@@ -583,7 +588,7 @@ var filtersFrom = function(x, y) {
     candidateElements.length = 0;
 
     // We need at least one element.
-    var first = null;
+    let first = null;
     if ( typeof x === 'number' ) {
         first = elementFromPoint(x, y);
     } else if ( x instanceof HTMLElement ) {
@@ -597,23 +602,29 @@ var filtersFrom = function(x, y) {
     }
 
     // Cosmetic filter candidates from ancestors.
-    var elem = first;
+    let elem = first;
     while ( elem && elem !== document.body ) {
         cosmeticFilterFromElement(elem);
         elem = elem.parentNode;
     }
     // The body tag is needed as anchor only when the immediate child
-    // uses`nth-of-type`.
-    var i = cosmeticFilterCandidates.length;
-    if ( i !== 0 && cosmeticFilterCandidates[i-1].indexOf(':nth-of-type(') !== -1 ) {
-        cosmeticFilterCandidates.push('##body');
+    // uses `nth-of-type`.
+    let i = cosmeticFilterCandidates.length;
+    if ( i !== 0 ) {
+        let selector = cosmeticFilterCandidates[i-1];
+        if (
+            selector.indexOf(':nth-of-type(') !== -1 &&
+            safeQuerySelectorAll(document.body, selector).length > 1
+        ) {
+            cosmeticFilterCandidates.push('##body');
+        }
     }
 
     // https://github.com/gorhill/uBlock/issues/1545
     // Network filter candidates from all other elements found at point (x, y).
     if ( typeof x === 'number' ) {
-        var attrName = pickerRoot.id + '-clickblind';
-        var previous;
+        let attrName = pickerRoot.id + '-clickblind';
+        let previous;
         elem = first;
         while ( elem !== null ) {
             previous = elem;
@@ -624,7 +635,7 @@ var filtersFrom = function(x, y) {
             }
             netFilterFromElement(elem);
         }
-        var elems = document.querySelectorAll('[' + attrName + ']');
+        let elems = document.querySelectorAll('[' + attrName + ']');
         i = elems.length;
         while ( i-- ) {
             elems[i].removeAttribute(attrName);
@@ -769,7 +780,7 @@ var filterToDOMInterface = (function() {
         var elems;
         if ( o.style ) {
             elems = document.querySelectorAll(o.style[0]);
-            lastAction = o.style.join(' ');
+            lastAction = o.style[0] + ' {' + o.style[1] + '}';
         } else if ( o.tasks ) {
             elems = vAPI.domFilterer.createProceduralFilter(o).exec();
         }
@@ -1015,35 +1026,34 @@ var onCandidateChanged = (function() {
 /******************************************************************************/
 
 var candidateFromFilterChoice = function(filterChoice) {
-    var slot = filterChoice.slot;
-    var filters = filterChoice.filters;
-    var filter = filters[slot];
+    let slot = filterChoice.slot;
+    let filters = filterChoice.filters;
+    let filter = filters[slot];
 
-    if ( filter === undefined ) {
-        return '';
-    }
+    if ( filter === undefined ) { return ''; }
 
     // For net filters there no such thing as a path
-    if ( filter.lastIndexOf('##', 0) !== 0 ) {
-        return filter;
-    }
+    if ( filter.startsWith('##') === false ) { return filter; }
 
     // At this point, we have a cosmetic filter
 
     // Modifier means "target broadly". Hence:
     // - Do not compute exact path.
     // - Discard narrowing directives.
+    // - Remove the id if one or more classes exist
+    //   TODO: should remove tag name too? ¯\_(ツ)_/¯
     if ( filterChoice.modifier ) {
         filter = filter.replace(/:nth-of-type\(\d+\)/, '');
-        // Remove the id if one or more classes exist.
-        if ( filter.charAt(2) === '#' && filter.indexOf('.') !== -1 ) {
-            filter = filter.replace(/#[^#.]+/, '');
+        if ( filter.indexOf('.') !== -1 ) {
+            if ( filter.charAt(2) === '#' ) {
+                filter = filter.replace(/#[^#.]+/, '');
+            }
         }
         return filter;
     }
 
     // Return path: the target element, then all siblings prepended
-    var selector = '', joiner = '';
+    let selector = '', joiner = '';
     for ( ; slot < filters.length; slot++ ) {
         filter = filters[slot];
         // Remove all classes when an id exists.
@@ -1052,19 +1062,17 @@ var candidateFromFilterChoice = function(filterChoice) {
         }
         selector = filter.slice(2) + joiner + selector;
         // Stop at any element with an id: these are unique in a web page
-        if ( filter.lastIndexOf('###', 0) === 0 ) {
-            break;
-        }
+        if ( filter.startsWith('###') ) { break; }
         // Stop if current selector matches only one element on the page
-        if ( document.querySelectorAll(selector).length === 1 ) {
-            break;
-        }
+        if ( document.querySelectorAll(selector).length === 1 ) { break; }
         joiner = ' > ';
     }
 
     // https://github.com/gorhill/uBlock/issues/2519
+    // https://github.com/uBlockOrigin/uBlock-issues/issues/17
     if (
         slot === filters.length &&
+        selector.startsWith('body > ') === false &&
         document.querySelectorAll(selector).length > 1
     ) {
         selector = 'body > ' + selector;
@@ -1093,7 +1101,15 @@ var filterChoiceFromEvent = function(ev) {
 /******************************************************************************/
 
 var onDialogClicked = function(ev) {
-    if ( ev.target === null ) {
+    if ( ev.isTrusted === false ) { return; }
+
+    // If the dialog is hidden, clicking on it force it to become visible.
+    if ( dialog.classList.contains('hide') ) {
+        dialog.classList.add('show');
+        dialog.classList.remove('hide');
+    }
+
+    else if ( ev.target === null ) {
         /* do nothing */
     }
 
@@ -1162,6 +1178,11 @@ var showDialog = function(options) {
 
     options = options || {};
 
+    // Typically the dialog will be forced to be visible when using a
+    // touch-aware device.
+    dialog.classList.toggle('show', options.show === true);
+    dialog.classList.remove('hide');
+
     // Create lists of candidate filters
     var populate = function(src, des) {
         var root = dialog.querySelector(des);
@@ -1263,43 +1284,81 @@ var onSvgHovered = (function() {
     };
 })();
 
-/******************************************************************************/
+/*******************************************************************************
+
+    Swipe right:
+        If picker not paused: quit picker
+        If picker paused and dialog visible: hide dialog
+        If picker paused and dialog not visible: quit picker
+
+    Swipe left:
+        If picker paused and dialog not visible: show dialog
+
+*/
 
 var onSvgTouchStartStop = (function() {
     var startX,
         startY;
     return function onTouch(ev) {
-        ev.preventDefault();
         if ( ev.type === 'touchstart' ) {
-            startX = ev.touches[0].pageX;
-            startY = ev.touches[0].pageY;
+            startX = ev.touches[0].screenX;
+            startY = ev.touches[0].screenY;
             return;
         }
         if ( startX === undefined ) { return; }
-        var stopX = ev.changedTouches[0].pageX,
-            stopY = ev.changedTouches[0].pageY,
+        if ( ev.cancelable === false ) { return; }
+        var stopX = ev.changedTouches[0].screenX,
+            stopY = ev.changedTouches[0].screenY,
+            angle = Math.abs(Math.atan2(stopY - startY, stopX - startX)),
             distance = Math.sqrt(
                 Math.pow(stopX - startX, 2),
                 Math.pow(stopY - startY, 2)
             );
-        // Swipe = exit element zapper/picker.
-        if ( distance > 32 ) {
-            stopPicker();
+        // Interpret touch events as a click events if swipe is not valid.
+        if ( distance < 32 ) {
+            onSvgClicked({
+                type: 'touch',
+                target: ev.target,
+                clientX: ev.changedTouches[0].pageX,
+                clientY: ev.changedTouches[0].pageY,
+                isTrusted: ev.isTrusted
+            });
+            ev.preventDefault();
             return;
         }
-        // Interpret touch event as a click.
-        onSvgClicked({
-            type: 'click',
-            target: ev.target,
-            clientX: startX,
-            clientY: startY
-        });
+        if ( distance < 64 ) { return; }
+        var angleUpperBound = Math.PI * 0.25 * 0.5,
+            swipeRight = angle < angleUpperBound;
+        if ( swipeRight === false && angle < Math.PI - angleUpperBound ) {
+            return;
+        }
+        ev.preventDefault();
+        // Swipe left.
+        if ( swipeRight === false ) {
+            if ( pickerBody.classList.contains('paused') ) {
+                dialog.classList.remove('hide');
+                dialog.classList.add('show');
+            }
+            return;
+        }
+        // Swipe right.
+        if (
+            pickerBody.classList.contains('paused') &&
+            dialog.classList.contains('show')
+        ) {
+            dialog.classList.remove('show');
+            dialog.classList.add('hide');
+            return;
+        }
+        stopPicker();
     };
 })();
 
 /******************************************************************************/
 
 var onSvgClicked = function(ev) {
+    if ( ev.isTrusted === false ) { return; }
+
     // If zap mode, highlight element under mouse, this makes the zapper usable
     // on touch screens.
     if ( pickerBody.classList.contains('zap') ) {
@@ -1330,7 +1389,7 @@ var onSvgClicked = function(ev) {
     if ( filtersFrom(ev.clientX, ev.clientY) === 0 ) {
         return;
     }
-    showDialog();
+    showDialog({ show: ev.type === 'touch' });
 };
 
 /******************************************************************************/
@@ -1397,14 +1456,15 @@ var stopPicker = function() {
     candidateElements = [];
     bestCandidateFilter = null;
 
-    if ( pickerRoot === null ) {
-        return;
-    }
+    if ( pickerRoot === null ) { return; }
 
     // https://github.com/gorhill/uBlock/issues/2060
-    if ( vAPI.userCSS ) {
-        vAPI.userCSS.remove(pickerStyle.textContent);
+    if ( vAPI.domFilterer instanceof Object ) {
+        vAPI.userStylesheet.remove(pickerCSS1);
+        vAPI.userStylesheet.remove(pickerCSS2);
+        vAPI.userStylesheet.apply();
     }
+    vAPI.domFilterer.unexcludeNode(pickerRoot);
 
     window.removeEventListener('scroll', onScrolled, true);
     pickerRoot.contentWindow.removeEventListener('keydown', onKeyPressed, true);
@@ -1414,9 +1474,8 @@ var stopPicker = function() {
     svgRoot.removeEventListener('click', onSvgClicked);
     svgRoot.removeEventListener('touchstart', onSvgTouchStartStop);
     svgRoot.removeEventListener('touchend', onSvgTouchStartStop);
-    pickerStyle.parentNode.removeChild(pickerStyle);
     pickerRoot.parentNode.removeChild(pickerRoot);
-    pickerRoot.onload = null;
+    pickerRoot.removeEventListener('load', stopPicker);
     pickerRoot =
     pickerBody =
     dialog =
@@ -1429,7 +1488,7 @@ var stopPicker = function() {
 /******************************************************************************/
 
 var startPicker = function(details) {
-    pickerRoot.onload = stopPicker;
+    pickerRoot.addEventListener('load', stopPicker);
 
     var frameDoc = pickerRoot.contentDocument;
     var parsedDom = (new DOMParser()).parseFromString(
@@ -1534,9 +1593,22 @@ var startPicker = function(details) {
 
 /******************************************************************************/
 
+var bootstrapPicker = function() {
+    pickerRoot.removeEventListener('load', bootstrapPicker);
+    vAPI.shutdown.add(stopPicker);
+    vAPI.messaging.send(
+        'elementPicker',
+        { what: 'elementPickerArguments' },
+        startPicker
+    );
+};
+
+/******************************************************************************/
+
 pickerRoot = document.createElement('iframe');
 pickerRoot.id = vAPI.sessionId;
-pickerRoot.style.cssText = [
+
+var pickerCSSStyle = [
     'background: transparent',
     'border: 0',
     'border-radius: 0',
@@ -1546,6 +1618,7 @@ pickerRoot.style.cssText = [
     'left: 0',
     'margin: 0',
     'max-height: none',
+    'max-width: none',
     'opacity: 1',
     'outline: 0',
     'padding: 0',
@@ -1556,43 +1629,52 @@ pickerRoot.style.cssText = [
     'z-index: 2147483647',
     ''
 ].join(' !important;');
+pickerRoot.style.cssText = pickerCSSStyle;
 
-// https://github.com/gorhill/uBlock/issues/1529
-// In addition to inline styles, harden the element picker styles by using
-// a dedicated style tag.
-pickerStyle = document.createElement('style');
-pickerStyle.textContent = [
+var pickerCSS1 = [
     '#' + pickerRoot.id + ' {',
-        pickerRoot.style.cssText,
-    '}',
+        pickerCSSStyle,
+    '}'
+].join('\n');
+var pickerCSS2 = [
     '[' + pickerRoot.id + '-clickblind] {',
         'pointer-events: none !important;',
-    '}',
-    ''
+    '}'
 ].join('\n');
-document.documentElement.appendChild(pickerStyle);
+
+// https://github.com/gorhill/uBlock/issues/1529
+//   In addition to inline styles, harden the element picker styles by using
+//   dedicated CSS rules.
+vAPI.userStylesheet.add(pickerCSS1);
+vAPI.userStylesheet.add(pickerCSS2);
+vAPI.userStylesheet.apply();
 
 // https://github.com/gorhill/uBlock/issues/2060
-if ( vAPI.domFilterer ) {
-    pickerRoot[vAPI.domFilterer.getExcludeId()] = true;
-}
-if ( vAPI.userCSS ) {
-    vAPI.userCSS.add(pickerStyle.textContent);
-}
-
-pickerRoot.onload = function() {
-    vAPI.shutdown.add(stopPicker);
-    vAPI.messaging.send(
-        'elementPicker',
-        { what: 'elementPickerArguments' },
-        startPicker
-    );
-};
+vAPI.domFilterer.excludeNode(pickerRoot);
 
+pickerRoot.addEventListener('load', bootstrapPicker);
 document.documentElement.appendChild(pickerRoot);
 
-// https://www.youtube.com/watch?v=sociXdKnyr8
-
 /******************************************************************************/
 
 })();
+
+
+
+
+
+
+
+
+/*******************************************************************************
+
+    DO NOT:
+    - Remove the following code
+    - Add code beyond the following code
+    Reason:
+    - https://github.com/gorhill/uBlock/pull/3721
+    - uBO never uses the return value from injected content scripts
+
+**/
+
+void 0;
