@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2014-2017 Raymond Hill
+    Copyright (C) 2014-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -20,7 +20,7 @@
 */
 
 /* jshint bitwise: false */
-/* global punycode */
+/* global punycode, HNTrieBuilder */
 
 'use strict';
 
@@ -58,6 +58,7 @@ var typeNameToTypeValue = {
         'stylesheet':  1 << 4,
              'image':  2 << 4,
             'object':  3 << 4,
+ 'object_subrequest':  3 << 4,
             'script':  4 << 4,
     'xmlhttprequest':  5 << 4,
          'sub_frame':  6 << 4,
@@ -69,10 +70,12 @@ var typeNameToTypeValue = {
           'popunder': 12 << 4,
         'main_frame': 13 << 4,  // start of 1st-party-only behavorial filtering
        'generichide': 14 << 4,
-     'inline-script': 15 << 4,
-              'data': 16 << 4,  // special: a generic data holder
-          'redirect': 17 << 4,
-            'webrtc': 18 << 4
+       'inline-font': 15 << 4,
+     'inline-script': 16 << 4,
+              'data': 17 << 4,  // special: a generic data holder
+          'redirect': 18 << 4,
+            'webrtc': 19 << 4,
+       'unsupported': 20 << 4
 };
 var otherTypeBitValue = typeNameToTypeValue.other;
 
@@ -91,18 +94,14 @@ var typeValueToTypeName = {
     12: 'popunder',
     13: 'document',
     14: 'generichide',
-    15: 'inline-script',
-    16: 'data',
-    17: 'redirect',
-    18: 'webrtc'
+    15: 'inline-font',
+    16: 'inline-script',
+    17: 'data',
+    18: 'redirect',
+    19: 'webrtc',
+    20: 'unsupported'
 };
 
-// All network request types to bitmap
-//   bring origin to 0 (from 4 -- see typeNameToTypeValue)
-//   left-shift 1 by the above-calculated value
-//   subtract 1 to set all type bits
-var allNetRequestTypesBitmap = (1 << (otherTypeBitValue >>> 4)) - 1;
-
 var BlockAnyTypeAnyParty = BlockAction | AnyType | AnyParty;
 var BlockAnyType = BlockAction | AnyType;
 var BlockAnyParty = BlockAction | AnyParty;
@@ -200,7 +199,11 @@ var rawToRegexStr = function(s, anchor) {
                  .replace(me.escape3, '')
                  .replace(me.escape4, '[^ ]*?');
     if ( anchor & 0x4 ) {
-        reStr = '[0-9a-z.-]*?' + reStr;
+        reStr = (
+                    reStr.startsWith('\\.') ?
+                        rawToRegexStr.reTextHostnameAnchor2 :
+                        rawToRegexStr.reTextHostnameAnchor1
+                ) + reStr;
     } else if ( anchor & 0x2 ) {
         reStr = '^' + reStr;
     }
@@ -213,6 +216,8 @@ rawToRegexStr.escape1 = /[.+?${}()|[\]\\]/g;
 rawToRegexStr.escape2 = /\^/g;
 rawToRegexStr.escape3 = /^\*|\*$/g;
 rawToRegexStr.escape4 = /\*/g;
+rawToRegexStr.reTextHostnameAnchor1 = '^[a-z-]+://(?:[^/?#]+\\.)?';
+rawToRegexStr.reTextHostnameAnchor2 = '^[a-z-]+://(?:[^/?#]+)?';
 
 var filterFingerprinter = µb.CompiledLineWriter.fingerprint;
 
@@ -232,9 +237,9 @@ var toLogDataInternal = function(categoryBits, tokenHash, filter) {
     } else if ( categoryBits & 0x004 ) {
         opts.push('first-party');
     }
-    var type = (categoryBits >>> 4) & 0x1F;
-    if ( type !== 0 && type !== 16 /* data */ ) {
-        opts.push(typeValueToTypeName[type]);
+    var type = categoryBits & 0x1F0;
+    if ( type !== 0 && type !== typeNameToTypeValue.data ) {
+        opts.push(typeValueToTypeName[type >>> 4]);
     }
     if ( logData.opts !== undefined ) {
         opts.push(logData.opts);
@@ -449,7 +454,7 @@ FilterPlainHostname.prototype.match = function() {
 FilterPlainHostname.prototype.logData = function() {
     return {
         raw: '||' + this.s + '^',
-        regex: rawToRegexStr(this.s, 0x4),
+        regex: rawToRegexStr(this.s + '^'),
         compiled: this.compile()
     };
 };
@@ -534,6 +539,38 @@ registerFilterClass(FilterPlainRightAnchored);
 
 /******************************************************************************/
 
+var FilterExactMatch = function(s) {
+    this.s = s;
+};
+
+FilterExactMatch.prototype.match = function(url) {
+    return url === this.s;
+};
+
+FilterExactMatch.prototype.logData = function() {
+    return {
+        raw: '|' + this.s + '|',
+        regex: rawToRegexStr(this.s, 0x3),
+        compiled: this.compile()
+    };
+};
+
+FilterExactMatch.prototype.compile = function() {
+    return [ this.fid, this.s ];
+};
+
+FilterExactMatch.compile = function(details) {
+    return [ FilterExactMatch.fid, details.f ];
+};
+
+FilterExactMatch.load = function(args) {
+    return new FilterExactMatch(args[1]);
+};
+
+registerFilterClass(FilterExactMatch);
+
+/******************************************************************************/
+
 var FilterPlainHnAnchored = function(s) {
     this.s = s;
 };
@@ -623,14 +660,13 @@ FilterGenericHnAnchored.prototype.match = function(url) {
     if ( this.re === null ) {
         this.re = new RegExp(rawToRegexStr(this.s, this.anchor));
     }
-    var matchStart = url.search(this.re);
-    return matchStart !== -1 && isHnAnchored(url, matchStart);
+    return this.re.test(url);
 };
 
 FilterGenericHnAnchored.prototype.logData = function() {
     var out = {
         raw: '||' + this.s,
-        regex: this.re.source,
+        regex: rawToRegexStr(this.s, this.anchor & ~0x4),
         compiled: this.compile()
     };
     return out;
@@ -851,9 +887,9 @@ FilterOriginHitSet.prototype = Object.create(FilterOrigin.prototype, {
     matchOrigin: {
         value: function() {
             if ( this.oneOf === null ) {
-                this.oneOf = new RegExp('(?:^|\\.)(?:' + this.domainOpt.replace(/\./g, '\\.') + ')$');
+                this.oneOf = HNTrieBuilder.fromDomainOpt(this.domainOpt);
             }
-            return this.oneOf.test(pageHostnameRegister);
+            return this.oneOf.matches(pageHostnameRegister);
         }
     },
 });
@@ -883,9 +919,9 @@ FilterOriginMissSet.prototype = Object.create(FilterOrigin.prototype, {
     matchOrigin: {
         value: function() {
             if ( this.noneOf === null ) {
-                this.noneOf = new RegExp('(?:^|\\.)(?:' + this.domainOpt.replace(/~/g, '').replace(/\./g, '\\.') + ')$');
+                this.noneOf = HNTrieBuilder.fromDomainOpt(this.domainOpt.replace(/~/g, ''));
             }
-            return this.noneOf.test(pageHostnameRegister) === false;
+            return this.noneOf.matches(pageHostnameRegister) === false;
         }
     },
 });
@@ -918,15 +954,15 @@ FilterOriginMixedSet.prototype = Object.create(FilterOrigin.prototype, {
                 i = hostnames.length,
                 hostname;
             while ( i-- ) {
-                hostname = hostnames[i].replace(/\./g, '\\.');
+                hostname = hostnames[i];
                 if ( hostname.charCodeAt(0) === 0x7E /* '~' */ ) {
                     noneOf.push(hostname.slice(1));
                 } else {
                     oneOf.push(hostname);
                 }
             }
-            this.oneOf = new RegExp('(?:^|\\.)(?:' + oneOf.join('|') + ')$');
-            this.noneOf = new RegExp('(?:^|\\.)(?:' + noneOf.join('|') + ')$');
+            this.oneOf = HNTrieBuilder.fromIterable(oneOf);
+            this.noneOf = HNTrieBuilder.fromIterable(noneOf);
         }
     },
     toDomainOpt: {
@@ -938,7 +974,8 @@ FilterOriginMixedSet.prototype = Object.create(FilterOrigin.prototype, {
         value: function() {
             if ( this.oneOf === null ) { this.init(); }
             var needle = pageHostnameRegister;
-            return this.oneOf.test(needle) && this.noneOf.test(needle) === false;
+            return this.oneOf.matches(needle) &&
+                   this.noneOf.matches(needle) === false;
         }
     },
 });
@@ -1102,12 +1139,12 @@ FilterHostnameDict.prototype.logData = function() {
 };
 
 FilterHostnameDict.prototype.compile = function() {
-    return [ this.fid, µb.setToArray(this.dict) ];
+    return [ this.fid, Array.from(this.dict) ];
 };
 
 FilterHostnameDict.load = function(args) {
     var f = new FilterHostnameDict();
-    f.dict = µb.setFromArray(args[1]);
+    f.dict = new Set(args[1]);
     return f;
 };
 
@@ -1168,6 +1205,10 @@ FilterPair.prototype.remove = function(fdata) {
     if ( arrayStrictEquals(this.f1.compile(), fdata) === true ) {
         this.f1 = this.f2;
     }
+    // https://github.com/uBlockOrigin/uBlock-issues/issues/84
+    if ( this.f1 === undefined ) {
+        console.log(JSON.stringify(fdata));
+    }
 };
 
 FilterPair.prototype.match = function(url, tokenBeg) {
@@ -1192,11 +1233,17 @@ FilterPair.prototype.compile = function() {
 
 FilterPair.prototype.upgrade = function(a) {
     var bucket = new FilterBucket(this.f1, this.f2, a);
-    this.f1 = this.f2 = this.f = null;
+    this.f1 = this.f2 = undefined;
+    this.f = null;
     FilterPair.available = this;
     return bucket;
 };
 
+FilterPair.prototype.downgrade = function() {
+    if ( this.f2 !== undefined ) { return this; }
+    if ( this.f1 !== undefined ) { return this.f1; }
+};
+
 FilterPair.load = function(args) {
     var f1 = filterFromCompiledData(args[1]),
         f2 = filterFromCompiledData(args[2]),
@@ -1292,7 +1339,11 @@ FilterBucket.prototype.compile = function() {
 };
 
 FilterBucket.prototype.downgrade = function() {
-    return new FilterPair(this.filters[0], this.filters[1]);
+    if ( this.filters.length > 2 ) { return this; }
+    if ( this.filters.length === 2 ) {
+        return new FilterPair(this.filters[0], this.filters[1]);
+    }
+    if ( this.filters.length === 1 ) { return this.filters[0]; }
 };
 
 FilterBucket.load = function(args) {
@@ -1325,6 +1376,12 @@ var FilterParser = function() {
     this.reBadCSP = /(?:^|;)\s*report-(?:to|uri)\b/;
     this.domainOpt = '';
     this.noTokenHash = µb.urlTokenizer.tokenHashFromString('*');
+    this.unsupportedTypeBit = this.bitFromType('unsupported');
+    // All network request types to bitmap
+    //   bring origin to 0 (from 4 -- see typeNameToTypeValue)
+    //   left-shift 1 by the above-calculated value
+    //   subtract 1 to set all type bits
+    this.allNetRequestTypeBits = (1 << (otherTypeBitValue >>> 4)) - 1;
     this.reset();
 };
 
@@ -1335,25 +1392,30 @@ var FilterParser = function() {
 
 FilterParser.prototype.toNormalizedType = {
             'beacon': 'other',
+               'css': 'stylesheet',
               'data': 'data',
           'document': 'main_frame',
           'elemhide': 'generichide',
               'font': 'font',
+             'frame': 'sub_frame',
+      'genericblock': 'unsupported',
        'generichide': 'generichide',
              'image': 'image',
+       'inline-font': 'inline-font',
      'inline-script': 'inline-script',
              'media': 'media',
             'object': 'object',
-             'other': 'other',
  'object-subrequest': 'object',
+             'other': 'other',
               'ping': 'other',
           'popunder': 'popunder',
              'popup': 'popup',
             'script': 'script',
         'stylesheet': 'stylesheet',
        'subdocument': 'sub_frame',
+               'xhr': 'xmlhttprequest',
     'xmlhttprequest': 'xmlhttprequest',
-            'webrtc': 'webrtc',
+            'webrtc': 'unsupported',
          'websocket': 'websocket'
 };
 
@@ -1405,16 +1467,16 @@ FilterParser.prototype.parseTypeOption = function(raw, not) {
     }
 
     // Non-discrete network types can't be negated.
-    if ( (typeBit & allNetRequestTypesBitmap) === 0 ) {
+    if ( (typeBit & this.allNetRequestTypeBits) === 0 ) {
         return;
     }
 
     // Negated type: set all valid network request type bits to 1
     if (
-        (typeBit & allNetRequestTypesBitmap) !== 0 &&
-        (this.types & allNetRequestTypesBitmap) === 0
+        (typeBit & this.allNetRequestTypeBits) !== 0 &&
+        (this.types & this.allNetRequestTypeBits) === 0
     ) {
-        this.types |= allNetRequestTypesBitmap;
+        this.types |= this.allNetRequestTypeBits;
     }
     this.types &= ~typeBit;
 };
@@ -1465,7 +1527,7 @@ FilterParser.prototype.parseOptions = function(s) {
         if ( not ) {
             opt = opt.slice(1);
         }
-        if ( opt === 'third-party' ) {
+        if ( opt === 'third-party' || opt === '3p' ) {
             this.parsePartyOption(false, not);
             continue;
         }
@@ -1508,7 +1570,7 @@ FilterParser.prototype.parseOptions = function(s) {
             this.important = Important;
             continue;
         }
-        if ( opt === 'first-party' ) {
+        if ( opt === 'first-party' || opt === '1p' ) {
             this.parsePartyOption(true, not);
             continue;
         }
@@ -1557,7 +1619,10 @@ FilterParser.prototype.translate = function() {
         this.dataStr = "connect-src https: http:";
         // https://bugs.chromium.org/p/chromium/issues/detail?id=669086
         // TODO: remove when most users are beyond Chromium v56
-        if ( vAPI.chromiumVersion < 57 ) {
+        if (
+            vAPI.webextFlavor.soup.has('chromium') &&
+            vAPI.webextFlavor.major < 57
+        ) {
             this.dataStr += '; frame-src *';
         }
         return;
@@ -1680,12 +1745,22 @@ FilterParser.prototype.parse = function(raw) {
         pos = s.lastIndexOf('$');
         if ( pos !== -1 ) {
             // https://github.com/gorhill/uBlock/issues/952
-            // Discard Adguard-specific `$$` filters.
+            //   Discard Adguard-specific `$$` filters.
             if ( s.indexOf('$$') !== -1 ) {
                 this.unsupported = true;
                 return this;
             }
             this.parseOptions(s.slice(pos + 1));
+            // https://github.com/gorhill/uBlock/issues/2283
+            //   Abort if type is only for unsupported types, otherwise
+            //   toggle off `unsupported` bit.
+            if ( this.types & this.unsupportedTypeBit ) {
+                this.types &= ~this.unsupportedTypeBit;
+                if ( this.types === 0 ) {
+                    this.unsupported = true;
+                    return this;
+                }
+            }
             s = s.slice(0, pos);
         }
     }
@@ -1756,12 +1831,18 @@ FilterParser.prototype.parse = function(raw) {
 
     // https://github.com/gorhill/uBlock/issues/1669#issuecomment-224822448
     // remove pointless leading *.
+    // https://github.com/gorhill/uBlock/issues/3034
+    // - We can remove anchoring if we need to match all at the start.
     if ( s.startsWith('*') ) {
-        s = s.replace(/^\*+([^%0-9a-z])/, '$1');
+        s = s.replace(/^\*+([^%0-9a-z])/i, '$1');
+        this.anchor &= ~0x6;
     }
     // remove pointless trailing *
+    // https://github.com/gorhill/uBlock/issues/3034
+    // - We can remove anchoring if we need to match all at the end.
     if ( s.endsWith('*') ) {
-        s = s.replace(/([^%0-9a-z])\*+$/, '$1');
+        s = s.replace(/([^%0-9a-z])\*+$/i, '$1');
+        this.anchor &= ~0x1;
     }
 
     // nothing left?
@@ -1956,27 +2037,27 @@ FilterContainer.prototype.freeze = function() {
 /******************************************************************************/
 
 FilterContainer.prototype.toSelfie = function() {
-    var categoriesToSelfie = function(categoryMap) {
-        var categoryEntries = [];
-        for ( var categoryEntry of categoryMap ) {
-            var tokenEntries = [];
-            for ( var tokenEntry of categoryEntry[1] ) {
+    let categoriesToSelfie = function(categoryMap) {
+        let selfie = [];
+        for ( let categoryEntry of categoryMap ) {
+            let tokenEntries = [];
+            for ( let tokenEntry of categoryEntry[1] ) {
                 tokenEntries.push([ tokenEntry[0], tokenEntry[1].compile() ]);
             }
-            categoryEntries.push([ categoryEntry[0], tokenEntries ]);
+            selfie.push([ categoryEntry[0], tokenEntries ]);
         }
-        return JSON.stringify(categoryEntries);
+        return selfie;
     };
 
-    var dataFiltersToSelfie = function(dataFilters) {
-        var selfie = [];
-        for ( var entry of dataFilters.values() ) {
+    let dataFiltersToSelfie = function(dataFilters) {
+        let selfie = [];
+        for ( let entry of dataFilters.values() ) {
             do {
                 selfie.push(entry.compile());
                 entry = entry.next;
             } while ( entry !== undefined );
         }
-        return JSON.stringify(selfie);
+        return selfie;
     };
 
     return {
@@ -2002,28 +2083,17 @@ FilterContainer.prototype.fromSelfie = function(selfie) {
     this.blockFilterCount = selfie.blockFilterCount;
     this.discardedCount = selfie.discardedCount;
 
-    var entries;
-
-    var categoryMap = new Map();
-    entries = JSON.parse(selfie.categories);
-    for ( var i = 0, ni = entries.length; i < ni; i++ ) {
-        var categoryEntry = entries[i],
-            tokenMap = new Map();
-        var tokenEntries = categoryEntry[1];
-        for ( var j = 0, nj = tokenEntries.length; j < nj; j++ ) {
-            var tokenEntry = tokenEntries[j];
+    for ( let categoryEntry of selfie.categories ) {
+        let tokenMap = new Map();
+        for ( let tokenEntry of categoryEntry[1] ) {
             tokenMap.set(tokenEntry[0], filterFromCompiledData(tokenEntry[1]));
         }
-        categoryMap.set(categoryEntry[0], tokenMap);
+        this.categories.set(categoryEntry[0], tokenMap);
     }
-    this.categories = categoryMap;
 
-    entries = JSON.parse(selfie.dataFilters);
-    var entry, bucket;
-    i = entries.length;
-    while ( i-- ) {
-        entry = FilterDataHolderEntry.load(entries[i]);
-        bucket = this.dataFilters.get(entry.tokenHash);
+    for ( let dataEntry of selfie.dataFilters ) {
+        let entry = FilterDataHolderEntry.load(dataEntry);
+        let bucket = this.dataFilters.get(entry.tokenHash);
         if ( bucket !== undefined ) {
             entry.next = bucket;
         }
@@ -2055,6 +2125,9 @@ FilterContainer.prototype.compile = function(raw, writer) {
         return false;
     }
 
+    // 0 = network filters
+    writer.select(0);
+
     // Pure hostnames, use more efficient dictionary lookup
     // https://github.com/chrisaljoudi/uBlock/issues/665
     // Create a dict keyed on request type etc.
@@ -2098,6 +2171,8 @@ FilterContainer.prototype.compile = function(raw, writer) {
         fdata = FilterPlainLeftAnchored.compile(parsed);
     } else if ( parsed.anchor === 0x1 ) {
         fdata = FilterPlainRightAnchored.compile(parsed);
+    } else if ( parsed.anchor === 0x3 ) {
+        fdata = FilterExactMatch.compile(parsed);
     } else if ( parsed.tokenBeg === 0 ) {
         fdata = FilterPlainPrefix0.compile(parsed);
     } else if ( parsed.tokenBeg === 1 ) {
@@ -2201,9 +2276,13 @@ FilterContainer.prototype.fromCompiledContent = function(reader) {
         filterPairId = FilterPair.fid,
         filterBucketId = FilterBucket.fid,
         filterDataHolderId = FilterDataHolder.fid,
+        redirectTypeValue = typeNameToTypeValue.redirect,
         args, bits, bucket, entry,
         tokenHash, fdata, fingerprint;
 
+    // 0 = network filters
+    reader.select(0);
+
     while ( reader.next() === true ) {
         args = reader.args();
         bits = args[0];
@@ -2215,7 +2294,7 @@ FilterContainer.prototype.fromCompiledContent = function(reader) {
 
         // Special cases: delegate to more specialized engines.
         // Redirect engine.
-        if ( (bits >>> 4) === 17 ) {
+        if ( (bits & 0x1F0) === redirectTypeValue ) {
             µb.redirectEngine.fromCompiledRule(args[1]);
             continue;
         }
@@ -2306,36 +2385,24 @@ FilterContainer.prototype.removeBadFilters = function() {
         entry = bucket.get(tokenHash);
         if ( entry === undefined ) { continue; }
         fdata = args[2];
-        if ( entry.fid === filterPairId ) {
+        if ( entry.fid === filterPairId || entry.fid === filterBucketId ) {
             entry.remove(fdata);
-            if ( entry.size === 1 ) {
-                bucket.set(tokenHash, entry.f1);
+            entry = entry.downgrade();
+            if ( entry !== undefined ) {
+                bucket.set(tokenHash, entry);
+            } else {
+                bucket.delete(tokenHash);
             }
-            continue;
-        }
-        if ( entry.fid === filterBucketId ) {
-            entry.remove(fdata);
-            if ( entry.size === 2 ) {
-                bucket.set(tokenHash, entry.downgrade());
-            }
-            continue;
-        }
-        if ( entry.fid === filterHostnameDictId ) {
+        } else if ( entry.fid === filterHostnameDictId ) {
             entry.remove(fdata);
             if ( entry.size === 0 ) {
                 bucket.delete(tokenHash);
-                if ( bucket.size === 0 ) {
-                    this.categories.delete(bits);
-                }
             }
-            continue;
-        }
-        if ( arrayStrictEquals(entry.compile(), fdata) === true ) {
+        } else if ( arrayStrictEquals(entry.compile(), fdata) ) {
             bucket.delete(tokenHash);
-            if ( bucket.size === 0 ) {
-                this.categories.delete(bits);
-            }
-            continue;
+        }
+        if ( bucket.size === 0 ) {
+            this.categories.delete(bits);
         }
     }
 };
