@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2014-2017 Raymond Hill
+    Copyright (C) 2014-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -19,7 +19,7 @@
     Home: https://github.com/gorhill/uBlock
 */
 
-/* global objectAssign, punycode, publicSuffixList */
+/* global punycode, publicSuffixList */
 
 'use strict';
 
@@ -80,10 +80,73 @@
 
 /******************************************************************************/
 
-// For now, only boolean type is supported.
+µBlock.loadHiddenSettings = function() {
+    var onLoaded = function(bin) {
+        if ( bin instanceof Object === false ) { return; }
+        var µb = µBlock,
+            hs = bin.hiddenSettings;
+        // Remove following condition once 1.15.12+ is widespread.
+        if (
+            hs instanceof Object === false &&
+            typeof bin.hiddenSettingsString === 'string'
+        ) {
+            vAPI.storage.remove('hiddenSettingsString');
+            hs = µBlock.hiddenSettingsFromString(bin.hiddenSettingsString);
+        }
+        if ( hs instanceof Object ) {
+            var hsDefault = µb.hiddenSettingsDefault;
+            for ( var key in hsDefault ) {
+                if (
+                    hsDefault.hasOwnProperty(key) &&
+                    hs.hasOwnProperty(key) &&
+                    typeof hs[key] === typeof hsDefault[key]
+                ) {
+                    µb.hiddenSettings[key] = hs[key];
+                }
+            }
+            // To remove once 1.15.26 is widespread. The reason is to ensure
+            // the change in the following commit is taken into account:
+            // https://github.com/gorhill/uBlock/commit/8071321e9104
+            if ( hs.manualUpdateAssetFetchPeriod === 2000 ) {
+                µb.hiddenSettings.manualUpdateAssetFetchPeriod =
+                    µb.hiddenSettingsDefault.manualUpdateAssetFetchPeriod;
+                hs.manualUpdateAssetFetchPeriod = undefined;
+                µb.saveHiddenSettings();
+            }
+        }
+        if ( vAPI.localStorage.getItem('immediateHiddenSettings') === null ) {
+            µb.saveImmediateHiddenSettings();
+        }
+    };
+
+    vAPI.storage.get(
+        [ 'hiddenSettings', 'hiddenSettingsString'],
+        onLoaded
+    );
+};
+
+// Note: Save only the settings which values differ from the default ones.
+// This way the new default values in the future will properly apply for those
+// which were not modified by the user.
+
+µBlock.saveHiddenSettings = function(callback) {
+    var bin = { hiddenSettings: {} };
+    for ( var prop in this.hiddenSettings ) {
+        if (
+            this.hiddenSettings.hasOwnProperty(prop) &&
+            this.hiddenSettings[prop] !== this.hiddenSettingsDefault[prop]
+        ) {
+            bin.hiddenSettings[prop] = this.hiddenSettings[prop];
+        }
+    }
+    vAPI.storage.set(bin, callback);
+    this.saveImmediateHiddenSettings();
+};
+
+/******************************************************************************/
 
 µBlock.hiddenSettingsFromString = function(raw) {
-    var out = objectAssign({}, this.hiddenSettingsDefault),
+    var out = Object.assign({}, this.hiddenSettingsDefault),
         lineIter = new this.LineIterator(raw),
         line, matches, name, value;
     while ( lineIter.eot() === false ) {
@@ -114,19 +177,13 @@
             break;
         }
     }
-    this.hiddenSettings = out;
-    vAPI.localStorage.setItem('hiddenSettings', JSON.stringify(out));
-    vAPI.storage.set({ hiddenSettingsString: this.stringFromHiddenSettings() });
+    return out;
 };
 
-/******************************************************************************/
-
 µBlock.stringFromHiddenSettings = function() {
     var out = [],
-        keys = Object.keys(this.hiddenSettings).sort(),
-        key;
-    for ( var i = 0; i < keys.length; i++ ) {
-        key = keys[i];
+        keys = Object.keys(this.hiddenSettings).sort();
+    for ( var key of keys ) {
         out.push(key + ' ' + this.hiddenSettings[key]);
     }
     return out.join('\n');
@@ -134,6 +191,24 @@
 
 /******************************************************************************/
 
+// These settings must be available immediately on startup, without delay
+// through the vAPI.localStorage. Add/remove settings as needed.
+
+µBlock.saveImmediateHiddenSettings = function() {
+    vAPI.localStorage.setItem(
+        'immediateHiddenSettings',
+        JSON.stringify({
+            suspendTabsUntilReady: this.hiddenSettings.suspendTabsUntilReady,
+            userResourcesLocation: this.hiddenSettings.userResourcesLocation
+        })
+    );
+};
+
+// Do this here to have these hidden settings loaded ASAP.
+µBlock.loadHiddenSettings();
+
+/******************************************************************************/
+
 µBlock.savePermanentFirewallRules = function() {
     this.keyvalSetOne('dynamicFilteringString', this.permanentFirewall.toString());
 };
@@ -168,29 +243,21 @@
 
 µBlock.loadSelectedFilterLists = function(callback) {
     var µb = this;
-    vAPI.storage.get([ 'selectedFilterLists', 'remoteBlacklists' ], function(bin) {
-        if ( !bin || !bin.selectedFilterLists && !bin.remoteBlacklists ) {
-            // Select default filter lists if first-time launch.
+    vAPI.storage.get('selectedFilterLists', function(bin) {
+        // Select default filter lists if first-time launch.
+        if ( !bin || Array.isArray(bin.selectedFilterLists) === false ) {
             µb.assets.metadata(function(availableLists) {
-                µb.saveSelectedFilterLists(µb.autoSelectRegionalFilterLists(availableLists));
+                µb.saveSelectedFilterLists(
+                    µb.autoSelectRegionalFilterLists(availableLists)
+                );
                 callback();
             });
             return;
         }
-        var listKeys = [];
-        if ( bin.selectedFilterLists ) {
-            listKeys = bin.selectedFilterLists;
-        } else if ( bin.remoteBlacklists ) {
-            var oldListKeys = µb.newListKeysFromOldData(bin.remoteBlacklists);
-            if ( oldListKeys.sort().join() !== listKeys.sort().join() ) {
-                listKeys = oldListKeys;
-                µb.saveSelectedFilterLists(listKeys);
-            }
-            // TODO(seamless migration):
-            // Uncomment when all have moved to v1.11 and beyond.
-            //vAPI.storage.remove('remoteBlacklists');
-        }
-        µb.selectedFilterLists = listKeys;
+        // TODO: Removes once 1.1.15 is in widespread use.
+        // https://github.com/gorhill/uBlock/issues/3383
+        vAPI.storage.remove('remoteBlacklists');
+        µb.selectedFilterLists = bin.selectedFilterLists;
         callback();
     });
 };
@@ -211,65 +278,14 @@
             this.removeFilterList(oldKeys[i]);
         }
     }
-    newKeys = this.setToArray(newSet);
+    newKeys = Array.from(newSet);
     var bin = {
-        selectedFilterLists: newKeys,
-        remoteBlacklists: this.oldDataFromNewListKeys(newKeys)
+        selectedFilterLists: newKeys
     };
     this.selectedFilterLists = newKeys;
     vAPI.storage.set(bin, callback);
 };
 
-// TODO(seamless migration):
-// Remove when all have moved to v1.11 and beyond.
-// >>>>>>>>
-µBlock.newListKeysFromOldData = function(oldLists) {
-    var aliases = this.assets.listKeyAliases,
-        listKeys = [], newKey;
-    for ( var oldKey in oldLists ) {
-        if ( oldLists[oldKey].off !== true ) {
-            newKey = aliases[oldKey];
-            listKeys.push(newKey ? newKey : oldKey);
-        }
-    }
-    return listKeys;
-};
-
-µBlock.oldDataFromNewListKeys = function(selectedFilterLists) {
-    var µb = this,
-        remoteBlacklists = {};
-    var reverseAliases = Object.keys(this.assets.listKeyAliases).reduce(
-        function(a, b) {
-            a[µb.assets.listKeyAliases[b]] = b; return a;
-        },
-        {}
-    );
-    remoteBlacklists = selectedFilterLists.reduce(
-        function(a, b) {
-            a[reverseAliases[b] || b] = { off: false };
-            return a;
-        },
-        {}
-    );
-    remoteBlacklists = Object.keys(µb.assets.listKeyAliases).reduce(
-        function(a, b) {
-            var aliases = µb.assets.listKeyAliases;
-            if (
-                b.startsWith('assets/') &&
-                aliases[b] !== 'public_suffix_list.dat' &&
-                aliases[b] !== 'ublock-resources' &&
-                !a[b]
-            ) {
-                a[b] = { off: true };
-            }
-            return a;
-        },
-        remoteBlacklists
-    );
-    return remoteBlacklists;
-};
-// <<<<<<<<
-
 /******************************************************************************/
 
 µBlock.applyFilterListSelection = function(details, callback) {
@@ -295,7 +311,7 @@
             return haystack.replace(
                 new RegExp(
                     '(^|\\n)' +
-                    needle.replace(/[.*+?^${}()|[\]\\]/g, '\\$&') +
+                    µb.escapeRegex(needle) +
                     '(\\n|$)', 'g'),
                 '\n'
             ).trim();
@@ -342,10 +358,10 @@
             }
             selectedListKeySet.add(assetKey);
         }
-        externalLists = this.setToArray(importedSet).sort().join('\n');
+        externalLists = Array.from(importedSet).sort().join('\n');
     }
 
-    var result = this.setToArray(selectedListKeySet);
+    var result = Array.from(selectedListKeySet);
     if ( externalLists !== this.userSettings.externalLists ) {
         this.userSettings.externalLists = externalLists;
         vAPI.storage.set({ externalLists: externalLists });
@@ -371,7 +387,7 @@
         }
         out.add(location);
     }
-    return this.setToArray(out);
+    return Array.from(out);
 };
 
 /******************************************************************************/
@@ -411,7 +427,7 @@
         vAPI.storage.set({ 'availableFilterLists': µb.availableFilterLists });
         µb.staticNetFilteringEngine.freeze();
         µb.redirectEngine.freeze();
-        µb.cosmeticFilteringEngine.freeze();
+        µb.staticExtFilteringEngine.freeze();
         µb.selfieManager.destroy();
     };
 
@@ -439,7 +455,7 @@
             selectedListKeys.push(key);
             continue;
         }
-        if ( this.matchCurrentLanguage(list.lang) ) {
+        if ( this.listMatchesEnvironment(list) ) {
             selectedListKeys.push(key);
             list.off = false;
         }
@@ -456,7 +472,7 @@
 
     // User filter list.
     newAvailableLists[this.userFiltersPath] = {
-        group: 'default',
+        group: 'user',
         title: vAPI.i18n('1pPageName')
     };
 
@@ -559,7 +575,7 @@
             if ( entries.hasOwnProperty(assetKey) === false ) { continue; }
             entry = entries[assetKey];
             if ( entry.content !== 'filters' ) { continue; }
-            newAvailableLists[assetKey] = objectAssign({}, entry);
+            newAvailableLists[assetKey] = Object.assign({}, entry);
         }
 
         // Load set of currently selected filter lists.
@@ -597,8 +613,6 @@
     }
     this.loadingFilterLists = true;
 
-    //quickProfiler.start('µBlock.loadFilterLists()');
-
     var µb = this,
         filterlistsCount = 0,
         loadedListKeys = [];
@@ -609,12 +623,10 @@
 
     var onDone = function() {
         µb.staticNetFilteringEngine.freeze();
-        µb.cosmeticFilteringEngine.freeze();
+        µb.staticExtFilteringEngine.freeze();
         µb.redirectEngine.freeze();
         vAPI.storage.set({ 'availableFilterLists': µb.availableFilterLists });
 
-        //quickProfiler.stop(0);
-
         vAPI.messaging.broadcast({
             what: 'staticFilteringDataChanged',
             parseCosmeticFilters: µb.userSettings.parseAllABPHideFilters,
@@ -630,14 +642,16 @@
 
     var applyCompiledFilters = function(assetKey, compiled) {
         var snfe = µb.staticNetFilteringEngine,
-            cfe = µb.cosmeticFilteringEngine,
-            acceptedCount = snfe.acceptedCount + cfe.acceptedCount,
-            discardedCount = snfe.discardedCount + cfe.discardedCount;
+            sxfe = µb.staticExtFilteringEngine,
+            acceptedCount = snfe.acceptedCount + sxfe.acceptedCount,
+            discardedCount = snfe.discardedCount + sxfe.discardedCount;
         µb.applyCompiledFilters(compiled, assetKey === µb.userFiltersPath);
         if ( µb.availableFilterLists.hasOwnProperty(assetKey) ) {
             var entry = µb.availableFilterLists[assetKey];
-            entry.entryCount = snfe.acceptedCount + cfe.acceptedCount - acceptedCount;
-            entry.entryUsedCount = entry.entryCount - (snfe.discardedCount + cfe.discardedCount - discardedCount);
+            entry.entryCount = snfe.acceptedCount + sxfe.acceptedCount -
+                acceptedCount;
+            entry.entryUsedCount = entry.entryCount -
+                (snfe.discardedCount + sxfe.discardedCount - discardedCount);
         }
         loadedListKeys.push(assetKey);
     };
@@ -654,7 +668,7 @@
         µb.availableFilterLists = lists;
 
         µb.redirectEngine.reset();
-        µb.cosmeticFilteringEngine.reset();
+        µb.staticExtFilteringEngine.reset();
         µb.staticNetFilteringEngine.reset();
         µb.selfieManager.destroy();
         µb.staticFilteringReverseLookup.resetLists();
@@ -729,6 +743,9 @@
 
 /******************************************************************************/
 
+// https://github.com/gorhill/uBlock/issues/3406
+//   Lower minimum update period to 1 day.
+
 µBlock.extractFilterListMetadata = function(assetKey, raw) {
     var listEntry = this.availableFilterLists[assetKey];
     if ( listEntry === undefined ) { return; }
@@ -738,7 +755,7 @@
     // https://github.com/gorhill/uBlock/issues/313
     // Always try to fetch the name if this is an external filter list.
     if ( listEntry.title === '' || listEntry.group === 'custom' ) {
-        matches = head.match(/(?:^|\n)!\s*Title:([^\n]+)/i);
+        matches = head.match(/(?:^|\n)(?:!|# )[\t ]*Title[\t ]*:([^\n]+)/i);
         if ( matches !== null ) {
             // https://bugs.chromium.org/p/v8/issues/detail?id=2869
             // JSON.stringify/JSON.parse is to work around String.slice()
@@ -748,9 +765,12 @@
         }
     }
     // Extract update frequency information
-    matches = head.match(/(?:^|\n)![\t ]*Expires:[\t ]*([\d]+)[\t ]*days?/i);
+    matches = head.match(/(?:^|\n)(?:!|# )[\t ]*Expires[\t ]*:[\t ]*(\d+)[\t ]*(h)?/i);
     if ( matches !== null ) {
-        v = Math.max(parseInt(matches[1], 10), 2);
+        v = Math.max(parseInt(matches[1], 10), 1);
+        if ( matches[2] !== undefined ) {
+            v = Math.ceil(v / 24);
+        }
         if ( v !== listEntry.updateAfter ) {
             this.assets.registerAssetSource(assetKey, { updateAfter: v });
         }
@@ -771,28 +791,25 @@
 /******************************************************************************/
 
 µBlock.compileFilters = function(rawText) {
-    var networkFilters = new this.CompiledLineWriter(),
-        cosmeticFilters = new this.CompiledLineWriter();
+    var writer = new this.CompiledLineWriter();
 
     // Useful references:
     //    https://adblockplus.org/en/filter-cheatsheet
     //    https://adblockplus.org/en/filters
     var staticNetFilteringEngine = this.staticNetFilteringEngine,
-        cosmeticFilteringEngine = this.cosmeticFilteringEngine,
+        staticExtFilteringEngine = this.staticExtFilteringEngine,
         reIsWhitespaceChar = /\s/,
         reMaybeLocalIp = /^[\d:f]/,
-        reIsLocalhostRedirect = /\s+(?:broadcasthost|local|localhost|localhost\.localdomain)(?=\s|$)/,
+        reIsLocalhostRedirect = /\s+(?:0\.0\.0\.0|broadcasthost|localhost|local|ip6-\w+)\b/,
         reLocalIp = /^(?:0\.0\.0\.0|127\.0\.0\.1|::1|fe80::1%lo0)/,
-        line, lineRaw, c, pos,
-        lineIter = new this.LineIterator(rawText);
+        line, c, pos,
+        lineIter = new this.LineIterator(this.processDirectives(rawText));
 
     while ( lineIter.eot() === false ) {
-        line = lineRaw = lineIter.next().trim();
-
         // rhill 2014-04-18: The trim is important here, as without it there
         // could be a lingering `\r` which would cause problems in the
         // following parsing code.
-
+        line = lineIter.next().trim();
         if ( line.length === 0 ) { continue; }
 
         // Strip comments
@@ -801,9 +818,7 @@
 
         // Parse or skip cosmetic filters
         // All cosmetic filters are caught here
-        if ( cosmeticFilteringEngine.compile(line, cosmeticFilters) ) {
-            continue;
-        }
+        if ( staticExtFilteringEngine.compile(line, writer) ) { continue; }
 
         // Whatever else is next can be assumed to not be a cosmetic filter
 
@@ -835,12 +850,10 @@
 
         if ( line.length === 0 ) { continue; }
 
-        staticNetFilteringEngine.compile(line, networkFilters);
+        staticNetFilteringEngine.compile(line, writer);
     }
 
-    return networkFilters.toString() +
-           '\n/* end of network - start of cosmetic */\n' +
-           cosmeticFilters.toString();
+    return writer.toString();
 };
 
 /******************************************************************************/
@@ -851,30 +864,75 @@
 
 µBlock.applyCompiledFilters = function(rawText, firstparty) {
     if ( rawText === '' ) { return; }
-    var separator = '\n/* end of network - start of cosmetic */\n',
-        pos = rawText.indexOf(separator),
-        reader = new this.CompiledLineReader(rawText.slice(0, pos));
+    var reader = new this.CompiledLineReader(rawText);
     this.staticNetFilteringEngine.fromCompiledContent(reader);
-    this.cosmeticFilteringEngine.fromCompiledContent(
-        reader.reset(rawText.slice(pos + separator.length)),
-        this.userSettings.ignoreGenericCosmeticFilters,
-        !firstparty && !this.userSettings.parseAllABPHideFilters
-    );
+    this.staticExtFilteringEngine.fromCompiledContent(reader, {
+        skipGenericCosmetic: this.userSettings.ignoreGenericCosmeticFilters,
+        skipCosmetic: !firstparty && !this.userSettings.parseAllABPHideFilters
+    });
 };
 
 /******************************************************************************/
 
-µBlock.loadRedirectResources = function(callback) {
+// https://github.com/AdguardTeam/AdguardBrowserExtension/issues/917
+
+µBlock.processDirectives = function(content) {
+    var reIf = /^!#(if|endif)\b([^\n]*)/gm,
+        parts = [],
+        beg = 0, depth = 0, discard = false;
+    while ( beg < content.length ) {
+        var match = reIf.exec(content);
+        if ( match === null ) { break; }
+        if ( match[1] === 'if' ) {
+            var expr = match[2].trim();
+            var target = expr.startsWith('!');
+            if ( target ) { expr = expr.slice(1); }
+            var token = this.processDirectives.tokens.get(expr);
+            if (
+                depth === 0 &&
+                discard === false &&
+                token !== undefined &&
+                vAPI.webextFlavor.soup.has(token) === target
+            ) {
+                parts.push(content.slice(beg, match.index));
+                discard = true;
+            }
+            depth += 1;
+            continue;
+        }
+        depth -= 1;
+        if ( depth < 0 ) { break; }
+        if ( depth === 0 && discard ) {
+            beg = match.index + match[0].length + 1;
+            discard = false;
+        }
+    }
+    if ( depth === 0 && parts.length !== 0 ) {
+        parts.push(content.slice(beg));
+        content = parts.join('\n');
+    }
+    return content.trim();
+};
+
+µBlock.processDirectives.tokens = new Map([
+    [ 'ext_ublock', 'ublock' ],
+    [ 'env_chromium', 'chromium' ],
+    [ 'env_edge', 'edge' ],
+    [ 'env_firefox', 'firefox' ],
+    [ 'env_mobile', 'mobile' ],
+    [ 'env_safari', 'safari' ],
+    [ 'cap_html_filtering', 'html_filtering' ],
+    [ 'cap_user_stylesheet', 'user_stylesheet' ]
+]);
+
+/******************************************************************************/
+
+µBlock.loadRedirectResources = function(updatedContent) {
     var µb = this,
         content = '';
 
-    if ( typeof callback !== 'function' ) {
-        callback = this.noopFunc;
-    }
-
     var onDone = function() {
         µb.redirectEngine.resourcesFromString(content);
-        callback();
     };
 
     var onUserResourcesLoaded = function(details) {
@@ -894,7 +952,17 @@
         µb.assets.fetchText(µb.hiddenSettings.userResourcesLocation, onUserResourcesLoaded);
     };
 
-    this.assets.get('ublock-resources', onResourcesLoaded);
+    if ( typeof updatedContent === 'string' && updatedContent.length !== 0 ) {
+        return onResourcesLoaded({ content: updatedContent });
+    }
+
+    var onSelfieReady = function(success) {
+        if ( success !== true ) {
+            µb.assets.get('ublock-resources', onResourcesLoaded);
+        }
+    };
+
+    µb.redirectEngine.resourcesFromSelfie(onSelfieReady);
 };
 
 /******************************************************************************/
@@ -915,11 +983,18 @@
     };
 
     var onCompiledListLoaded = function(details) {
-        if ( details.content === '' ) {
+        var selfie;
+        try {
+            selfie = JSON.parse(details.content);
+        } catch (ex) {
+        }
+        if (
+            selfie === undefined ||
+            publicSuffixList.fromSelfie(selfie) === false
+        ) {
             µb.assets.get(assetKey, onRawListLoaded);
             return;
         }
-        publicSuffixList.fromSelfie(JSON.parse(details.content));
         callback();
     };
 
@@ -943,31 +1018,54 @@
 // some set time.
 
 µBlock.selfieManager = (function() {
-    var timer = null;
+    let µb = µBlock;
+    let timer = null;
 
-    var create = function() {
+    // As of 2018-05-31:
+    // JSON.stringify-ing ourselves results in a better baseline
+    // memory usage at selfie-load time. For some reasons.
+
+    let create = function() {
         timer = null;
-        var selfie = {
-            magic: this.systemSettings.selfieMagic,
-            publicSuffixList: publicSuffixList.toSelfie(),
-            availableFilterLists: this.availableFilterLists,
-            staticNetFilteringEngine: this.staticNetFilteringEngine.toSelfie(),
-            redirectEngine: this.redirectEngine.toSelfie(),
-            cosmeticFilteringEngine: this.cosmeticFilteringEngine.toSelfie()
+        let selfie = {
+            magic: µb.systemSettings.selfieMagic,
+            availableFilterLists: JSON.stringify(µb.availableFilterLists),
+            staticNetFilteringEngine: JSON.stringify(µb.staticNetFilteringEngine.toSelfie()),
+            redirectEngine: JSON.stringify(µb.redirectEngine.toSelfie()),
+            staticExtFilteringEngine: JSON.stringify(µb.staticExtFilteringEngine.toSelfie())
         };
         vAPI.cacheStorage.set({ selfie: selfie });
-    }.bind(µBlock);
+    };
 
-    var destroy = function() {
+    let load = function(callback) {
+        vAPI.cacheStorage.get('selfie', function(bin) {
+            if (
+                bin instanceof Object === false ||
+                bin.selfie instanceof Object === false ||
+                bin.selfie.magic !== µb.systemSettings.selfieMagic ||
+                bin.selfie.redirectEngine === undefined
+            ) {
+                return callback(false);
+            }
+            µb.availableFilterLists = JSON.parse(bin.selfie.availableFilterLists);
+            µb.staticNetFilteringEngine.fromSelfie(JSON.parse(bin.selfie.staticNetFilteringEngine));
+            µb.redirectEngine.fromSelfie(JSON.parse(bin.selfie.redirectEngine));
+            µb.staticExtFilteringEngine.fromSelfie(JSON.parse(bin.selfie.staticExtFilteringEngine));
+            callback(true);
+        });
+    };
+
+    let destroy = function() {
         if ( timer !== null ) {
             clearTimeout(timer);
             timer = null;
         }
         vAPI.cacheStorage.remove('selfie');
-        timer = vAPI.setTimeout(create, this.selfieAfter);
-    }.bind(µBlock);
+        timer = vAPI.setTimeout(create, µb.selfieAfter);
+    };
 
     return {
+        load: load,
         destroy: destroy
     };
 })();
@@ -1032,9 +1130,6 @@
         if ( Array.isArray(data.selectedFilterLists) ) {
             bin.selectedFilterLists = data.selectedFilterLists;
             binNotEmpty = true;
-        } else if ( typeof data.filterLists === 'object' ) {
-            bin.selectedFilterLists = µb.newListKeysFromOldData(data.filterLists);
-            binNotEmpty = true;
         }
 
         if ( typeof data.netWhitelist === 'string' ) {
@@ -1073,6 +1168,33 @@
 
 /******************************************************************************/
 
+// https://github.com/gorhill/uBlock/issues/2344
+//   Support mutliple locales per filter list.
+
+// https://github.com/gorhill/uBlock/issues/3210
+//   Support ability to auto-enable a filter list based on user agent.
+
+µBlock.listMatchesEnvironment = function(details) {
+    var re;
+    // Matches language?
+    if ( typeof details.lang === 'string' ) {
+        re = this.listMatchesEnvironment.reLang;
+        if ( re === undefined ) {
+            re = new RegExp('\\b' + self.navigator.language.slice(0, 2) + '\\b');
+            this.listMatchesEnvironment.reLang = re;
+        }
+        if ( re.test(details.lang) ) { return true; }
+    }
+    // Matches user agent?
+    if ( typeof details.ua === 'string' ) {
+        re = new RegExp('\\b' + this.escapeRegex(details.ua) + '\\b', 'i');
+        if ( re.test(self.navigator.userAgent) ) { return true; }
+    }
+    return false;
+};
+
+/******************************************************************************/
+
 µBlock.scheduleAssetUpdater = (function() {
     var timer, next = 0;
     return function(updateDelay) {
@@ -1112,7 +1234,7 @@
                 this.availableFilterLists.hasOwnProperty(details.assetKey) === false ||
                 this.selectedFilterLists.indexOf(details.assetKey) === -1
             ) {
-                return false;
+                return;
             }
         }
         // https://github.com/gorhill/uBlock/issues/2594
@@ -1121,10 +1243,10 @@
                 this.hiddenSettings.ignoreRedirectFilters === true &&
                 this.hiddenSettings.ignoreScriptInjectFilters === true
             ) {
-                return false;
+                return;
             }
         }
-        return;
+        return true;
     }
 
     // Compile the list while we have the raw version in memory
@@ -1150,8 +1272,9 @@
                 this.compilePublicSuffixList(details.content);
             }
         } else if ( details.assetKey === 'ublock-resources' ) {
+            this.redirectEngine.invalidateResourcesSelfie();
             if ( cached ) {
-                this.redirectEngine.resourcesFromString(details.content);
+                this.loadRedirectResources(details.content);
             }
         }
         vAPI.messaging.broadcast({
@@ -1200,7 +1323,7 @@
         if ( details.entry.content === 'filters' ) {
             if (
                 details.entry.off !== true ||
-                this.matchCurrentLanguage(details.entry.lang)
+                this.listMatchesEnvironment(details.entry)
             ) {
                 this.saveSelectedFilterLists([ details.assetKey ], true);
             }
