@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2014-2016 Raymond Hill
+    Copyright (C) 2014-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -128,7 +128,7 @@ housekeep itself.
 */
 
 µb.tabContextManager = (function() {
-    var tabContexts = Object.create(null);
+    var tabContexts = new Map();
 
     // https://github.com/chrisaljoudi/uBlock/issues/1001
     // This is to be used as last-resort fallback in case a tab is found to not
@@ -136,11 +136,17 @@ housekeep itself.
     var mostRecentRootDocURL = '';
     var mostRecentRootDocURLTimestamp = 0;
 
-    var popupCandidates = Object.create(null);
+    var popupCandidates = new Map();
 
     var PopupCandidate = function(targetTabId, openerTabId) {
         this.targetTabId = targetTabId;
-        this.openerTabId = openerTabId;
+        this.opener = {
+            tabId: openerTabId,
+            popunder: false,
+            trustedURL: openerTabId === µb.mouseEventRegister.tabId ?
+                µb.mouseEventRegister.url :
+                ''
+        };
         this.selfDestructionTimer = null;
         this.launchSelfDestruction();
     };
@@ -149,7 +155,7 @@ housekeep itself.
         if ( this.selfDestructionTimer !== null ) {
             clearTimeout(this.selfDestructionTimer);
         }
-        delete popupCandidates[this.targetTabId];
+        popupCandidates.delete(this.targetTabId);
     };
 
     PopupCandidate.prototype.launchSelfDestruction = function() {
@@ -160,24 +166,36 @@ housekeep itself.
     };
 
     var popupCandidateTest = function(targetTabId) {
-        var candidates = popupCandidates, entry;
-        for ( var tabId in candidates ) {
-            entry = candidates[tabId];
-            if ( targetTabId !== tabId && targetTabId !== entry.openerTabId ) {
+        for ( var entry of popupCandidates ) {
+            var tabId = entry[0];
+            var candidate = entry[1];
+            if (
+                targetTabId !== tabId &&
+                targetTabId !== candidate.opener.tabId
+            ) {
                 continue;
             }
-            if ( vAPI.tabs.onPopupUpdated(tabId, entry.openerTabId) === true ) {
-                entry.destroy();
+            // https://github.com/gorhill/uBlock/issues/3129
+            //   If the trigger is a change in the opener's URL, mark the entry
+            //   as candidate for popunder filtering.
+            if ( targetTabId === candidate.opener.tabId ) {
+                candidate.opener.popunder = true;
+            }
+            if ( vAPI.tabs.onPopupUpdated(tabId, candidate.opener) === true ) {
+                candidate.destroy();
             } else {
-                entry.launchSelfDestruction();
+                candidate.launchSelfDestruction();
             }
         }
     };
 
     vAPI.tabs.onPopupCreated = function(targetTabId, openerTabId) {
-        var popup = popupCandidates[targetTabId];
+        var popup = popupCandidates.get(targetTabId);
         if ( popup === undefined ) {
-            popupCandidates[targetTabId] = new PopupCandidate(targetTabId, openerTabId);
+            popupCandidates.set(
+                targetTabId,
+                new PopupCandidate(targetTabId, openerTabId)
+            );
         }
         popupCandidateTest(targetTabId);
     };
@@ -193,7 +211,7 @@ housekeep itself.
     };
 
     var TabContext = function(tabId) {
-        this.tabId = tabId.toString();
+        this.tabId = tabId;
         this.stack = [];
         this.rawURL =
         this.normalURL =
@@ -205,18 +223,16 @@ housekeep itself.
         this.netFiltering = true;
         this.netFilteringReadTime = 0;
 
-        tabContexts[tabId] = this;
+        tabContexts.set(tabId, this);
     };
 
     TabContext.prototype.destroy = function() {
-        if ( vAPI.isBehindTheSceneTabId(this.tabId) ) {
-            return;
-        }
+        if ( vAPI.isBehindTheSceneTabId(this.tabId) ) { return; }
         if ( this.gcTimer !== null ) {
             clearTimeout(this.gcTimer);
             this.gcTimer = null;
         }
-        delete tabContexts[this.tabId];
+        tabContexts.delete(this.tabId);
     };
 
     TabContext.prototype.onTab = function(tab) {
@@ -321,11 +337,16 @@ housekeep itself.
     // tab, there is no longer any ambiguity about which root URL is really
     // sitting in which tab.
     TabContext.prototype.commit = function(url) {
-        if ( vAPI.isBehindTheSceneTabId(this.tabId) ) {
-            return;
+        if ( vAPI.isBehindTheSceneTabId(this.tabId) ) { return; }
+        if ( this.stack.length !== 0 ) {
+            var top = this.stack[this.stack.length - 1];
+            if ( top.url === url && top.committed ) {
+                return false;
+            }
         }
         this.stack = [new StackEntry(url, true)];
         this.update();
+        return true;
     };
 
     TabContext.prototype.getNetFilteringSwitch = function() {
@@ -345,7 +366,7 @@ housekeep itself.
     // These are to be used for the API of the tab context manager.
 
     var push = function(tabId, url) {
-        var entry = tabContexts[tabId];
+        var entry = tabContexts.get(tabId);
         if ( entry === undefined ) {
             entry = new TabContext(tabId);
             entry.autodestroy();
@@ -358,13 +379,13 @@ housekeep itself.
 
     // Find a tab context for a specific tab.
     var lookup = function(tabId) {
-        return tabContexts[tabId] || null;
+        return tabContexts.get(tabId) || null;
     };
 
     // Find a tab context for a specific tab. If none is found, attempt to
     // fix this. When all fail, the behind-the-scene context is returned.
     var mustLookup = function(tabId) {
-        var entry = tabContexts[tabId];
+        var entry = tabContexts.get(tabId);
         if ( entry !== undefined ) {
             return entry;
         }
@@ -389,24 +410,23 @@ housekeep itself.
         // about to fall through the cracks.
         // Example: Chromium + case #12 at
         //          http://raymondhill.net/ublock/popup.html
-        return tabContexts[vAPI.noTabId];
+        return tabContexts.get(vAPI.noTabId);
     };
 
     // https://github.com/gorhill/uBlock/issues/1735
     //   Filter for popups if actually committing.
     var commit = function(tabId, url) {
-        var entry = tabContexts[tabId];
+        var entry = tabContexts.get(tabId);
         if ( entry === undefined ) {
             entry = push(tabId, url);
-        } else {
-            entry.commit(url);
+        } else if ( entry.commit(url) ) {
             popupCandidateTest(tabId);
         }
         return entry;
     };
 
     var exists = function(tabId) {
-        return tabContexts[tabId] !== undefined;
+        return tabContexts.get(tabId) !== undefined;
     };
 
     // Behind-the-scene tab context
@@ -417,6 +437,9 @@ housekeep itself.
         entry.normalURL = µb.normalizePageURL(entry.tabId);
         entry.rootHostname = µb.URI.hostnameFromURI(entry.normalURL);
         entry.rootDomain = µb.URI.domainFromHostname(entry.rootHostname);
+
+        entry = new TabContext(vAPI.anyTabId);
+        entry.stack.push(new StackEntry('', true));
     })();
 
     // Context object, typically to be used to feed filtering engines.
@@ -428,7 +451,7 @@ housekeep itself.
         var tabContext = lookup(tabId);
         this.rootHostname = tabContext.rootHostname;
         this.rootDomain = tabContext.rootDomain;
-        this.pageHostname = 
+        this.pageHostname =
         this.pageDomain =
         this.requestURL =
         this.requestHostname =
@@ -463,13 +486,18 @@ housekeep itself.
 // content has changed.
 
 vAPI.tabs.onNavigation = function(details) {
-    if ( details.frameId !== 0 ) {
-        return;
+    if ( details.frameId === 0 ) {
+        µb.tabContextManager.commit(details.tabId, details.url);
+        let pageStore = µb.bindTabToPageStats(details.tabId, 'tabCommitted');
+        if ( pageStore ) {
+            pageStore.journalAddRootFrame('committed', details.url);
+        }
     }
-    µb.tabContextManager.commit(details.tabId, details.url);
-    var pageStore = µb.bindTabToPageStats(details.tabId, 'tabCommitted');
-    if ( pageStore ) {
-        pageStore.journalAddRootFrame('committed', details.url);
+    if ( µb.canInjectScriptletsNow ) {
+        let pageStore = µb.pageStoreFromTabId(details.tabId);
+        if ( pageStore !== null && pageStore.getNetFilteringSwitch() ) {
+            µb.scriptletFilteringEngine.injectNow(details);
+        }
     }
 };
 
@@ -480,13 +508,8 @@ vAPI.tabs.onNavigation = function(details) {
 // the extension icon won't be properly refreshed.
 
 vAPI.tabs.onUpdated = function(tabId, changeInfo, tab) {
-    if ( !tab.url || tab.url === '' ) {
-        return;
-    }
-    if ( !changeInfo.url ) {
-        return;
-    }
-
+    if ( !tab.url || tab.url === '' ) { return; }
+    if ( !changeInfo.url ) { return; }
     µb.tabContextManager.commit(tabId, changeInfo.url);
     µb.bindTabToPageStats(tabId, 'tabUpdated');
 };
@@ -531,27 +554,26 @@ vAPI.tabs.onPopupUpdated = (function() {
         logData;
 
     // https://github.com/gorhill/uBlock/commit/1d448b85b2931412508aa01bf899e0b6f0033626#commitcomment-14944764
-    // See if two URLs are different, disregarding scheme -- because the scheme
-    // can be unilaterally changed by the browser.
+    //   See if two URLs are different, disregarding scheme -- because the
+    //   scheme can be unilaterally changed by the browser.
+    // https://github.com/gorhill/uBlock/issues/1378
+    //   Maybe no link element was clicked.
+    // https://github.com/gorhill/uBlock/issues/3287
+    //   Do not bail out if the target URL has no hostname.
     var areDifferentURLs = function(a, b) {
-        // https://github.com/gorhill/uBlock/issues/1378
-        // Maybe no link element was clicked.
-        if ( b === '' ) {
-            return true;
-        }
+        if ( b === '' ) { return true; }
+        if ( b.startsWith('about:') ) { return false; }
         var pos = a.indexOf('://');
-        if ( pos === -1 ) {
-            return false;
-        }
+        if ( pos === -1 ) { return false; }
         a = a.slice(pos);
         pos = b.indexOf('://');
-        if ( pos === -1 ) {
-            return false;
+        if ( pos !== -1 ) {
+            b = b.slice(pos);
         }
-        return b.slice(pos) !== a;
+        return b !== a;
     };
 
-    var popupMatch = function(openerURL, targetURL, clickedURL, popupType) {
+    var popupMatch = function(openerURL, targetURL, popupType) {
         var openerHostname = µb.URI.hostnameFromURI(openerURL),
             openerDomain = µb.URI.domainFromHostname(openerHostname),
             result;
@@ -583,23 +605,29 @@ vAPI.tabs.onPopupUpdated = (function() {
         //   URL.
         if ( openerHostname !== '' && targetURL !== 'about:blank' ) {
             // Check per-site switch first
-            if ( µb.hnSwitches.evaluateZ('no-popups', openerHostname) === true ) {
-                if (
-                    typeof clickedURL === 'string' &&
-                    areDifferentURLs(targetURL, clickedURL)
-                ) {
-                    logData = {
-                        source: 'switch',
-                        raw: 'no-popups: ' + µb.hnSwitches.z + ' true'
-                    };
-                    return 1;
-                }
+            // https://github.com/gorhill/uBlock/issues/3060
+            // - The no-popups switch must apply only to popups, not to
+            //   popunders.
+            if (
+                popupType === 'popup' &&
+                µb.hnSwitches.evaluateZ('no-popups', openerHostname)
+            ) {
+                logData = {
+                    raw: 'no-popups: ' + µb.hnSwitches.z + ' true',
+                    result: 1,
+                    source: 'switch'
+                };
+                return 1;
             }
 
             // https://github.com/gorhill/uBlock/issues/581
-            //   Take into account popup-specific rules in dynamic URL filtering, OR
-            //   generic allow rules.
-            result = µb.sessionURLFiltering.evaluateZ(openerHostname, targetURL, popupType);
+            //   Take into account popup-specific rules in dynamic URL
+            //   filtering, OR generic allow rules.
+            result = µb.sessionURLFiltering.evaluateZ(
+                openerHostname,
+                targetURL,
+                popupType
+            );
             if (
                 result === 1 && µb.sessionURLFiltering.type === popupType ||
                 result === 2
@@ -609,10 +637,14 @@ vAPI.tabs.onPopupUpdated = (function() {
             }
 
             // https://github.com/gorhill/uBlock/issues/581
-            //   Take into account `allow` rules in dynamic filtering: `block` rules
-            //   are ignored, as block rules are not meant to block specific types
-            //   like `popup` (just like with static filters).
-            result = µb.sessionFirewall.evaluateCellZY(openerHostname, context.requestHostname, popupType);
+            //   Take into account `allow` rules in dynamic filtering: `block`
+            //   rules are ignored, as block rules are not meant to block
+            //   specific types like `popup` (just like with static filters).
+            result = µb.sessionFirewall.evaluateCellZY(
+                openerHostname,
+                context.requestHostname,
+                popupType
+            );
             if ( result === 2 ) {
                 logData = µb.sessionFirewall.toLogData();
                 return 2;
@@ -648,7 +680,7 @@ vAPI.tabs.onPopupUpdated = (function() {
         if ( logData.token === µb.staticNetFilteringEngine.dotTokenHash ) {
             return result;
         }
-        var re = new RegExp(logData.regex),
+        var re = new RegExp(logData.regex, 'i'),
             matches = re.exec(popunderURL);
         if ( matches === null ) { return 0; }
         var beg = matches.index,
@@ -667,7 +699,7 @@ vAPI.tabs.onPopupUpdated = (function() {
     };
 
     var popunderMatch = function(openerURL, targetURL) {
-        var result = popupMatch(targetURL, openerURL, null, 'popunder');
+        var result = popupMatch(targetURL, openerURL, 'popunder');
         if ( result === 1 ) {
             return result;
         }
@@ -685,7 +717,7 @@ vAPI.tabs.onPopupUpdated = (function() {
         result = mapPopunderResult(
             popunderURL,
             popunderHostname,
-            popupMatch(targetURL, popunderURL, null, 'popup')
+            popupMatch(targetURL, popunderURL, 'popup')
         );
         if ( result !== 0 ) {
             return result;
@@ -699,12 +731,13 @@ vAPI.tabs.onPopupUpdated = (function() {
         return mapPopunderResult(
             popunderURL,
             popunderHostname,
-            popupMatch(targetURL, popunderURL, null, 'popup')
+            popupMatch(targetURL, popunderURL, 'popup')
         );
     };
 
-    return function(targetTabId, openerTabId) {
+    return function(targetTabId, openerDetails) {
         // Opener details.
+        var openerTabId = openerDetails.tabId;
         var tabContext = µb.tabContextManager.lookup(openerTabId);
         if ( tabContext === null ) { return; }
         var openerURL = tabContext.rawURL;
@@ -738,10 +771,15 @@ vAPI.tabs.onPopupUpdated = (function() {
 
         // Popup test.
         var popupType = 'popup',
-            result = popupMatch(openerURL, targetURL, µb.mouseURL, 'popup');
+            result = 0;
+        // https://github.com/gorhill/uBlock/issues/2919
+        // - If the target tab matches a clicked link, assume it's legit.
+        if ( areDifferentURLs(targetURL, openerDetails.trustedURL) ) {
+            result = popupMatch(openerURL, targetURL, 'popup');
+        }
 
         // Popunder test.
-        if ( result === 0 ) {
+        if ( result === 0 && openerDetails.popunder ) {
             result = popunderMatch(openerURL, targetURL);
             if ( result === 1 ) {
                 popupType = 'popunder';
@@ -778,7 +816,7 @@ vAPI.tabs.onPopupUpdated = (function() {
 
         // Blocked
         if ( µb.userSettings.showIconBadge ) {
-            µb.updateBadgeAsync(openerTabId);
+            µb.updateToolbarIcon(openerTabId, 0x02);
         }
 
         // It is a popup, block and remove the tab.
@@ -802,7 +840,7 @@ vAPI.tabs.registerListeners();
 // Create an entry for the tab if it doesn't exist.
 
 µb.bindTabToPageStats = function(tabId, context) {
-    this.updateBadgeAsync(tabId);
+    this.updateToolbarIcon(tabId, 0x03);
 
     // Do not create a page store for URLs which are of no interests
     if ( µb.tabContextManager.exists(tabId) === false ) {
@@ -811,13 +849,15 @@ vAPI.tabs.registerListeners();
     }
 
     // Reuse page store if one exists: this allows to guess if a tab is a popup
-    var pageStore = this.pageStores[tabId];
+    var pageStore = this.pageStores.get(tabId);
 
     // Tab is not bound
-    if ( !pageStore ) {
+    if ( pageStore === undefined ) {
         this.updateTitle(tabId);
+        pageStore = this.PageStore.factory(tabId, context);
+        this.pageStores.set(tabId, pageStore);
         this.pageStoresToken = Date.now();
-        return (this.pageStores[tabId] = this.PageStore.factory(tabId));
+        return pageStore;
     }
 
     // https://github.com/chrisaljoudi/uBlock/issues/516
@@ -847,10 +887,10 @@ vAPI.tabs.registerListeners();
 
 µb.unbindTabFromPageStats = function(tabId) {
     //console.debug('µBlock> unbindTabFromPageStats(%d)', tabId);
-    var pageStore = this.pageStores[tabId];
+    var pageStore = this.pageStores.get(tabId);
     if ( pageStore !== undefined ) {
         pageStore.dispose();
-        delete this.pageStores[tabId];
+        this.pageStores.delete(tabId);
         this.pageStoresToken = Date.now();
     }
 };
@@ -858,121 +898,128 @@ vAPI.tabs.registerListeners();
 /******************************************************************************/
 
 µb.pageStoreFromTabId = function(tabId) {
-    return this.pageStores[tabId] || null;
+    return this.pageStores.get(tabId) || null;
 };
 
 µb.mustPageStoreFromTabId = function(tabId) {
-    return this.pageStores[tabId] || this.pageStores[vAPI.noTabId];
+    return this.pageStores.get(tabId) || this.pageStores.get(vAPI.noTabId);
 };
 
 /******************************************************************************/
 
 // Permanent page store for behind-the-scene requests. Must never be removed.
 
-µb.pageStores[vAPI.noTabId] = µb.PageStore.factory(vAPI.noTabId);
-µb.pageStores[vAPI.noTabId].title = vAPI.i18n('logBehindTheScene');
+(function() {
+    var pageStore = µb.PageStore.factory(vAPI.noTabId);
+    µb.pageStores.set(pageStore.tabId, pageStore);
+    pageStore.title = vAPI.i18n('logBehindTheScene');
+
+    pageStore = µb.PageStore.factory(vAPI.anyTabId);
+    µb.pageStores.set(pageStore.tabId, pageStore);
+    pageStore.title = '[Any one of the known tabs]';
+})();
 
 /******************************************************************************/
 
 // Update visual of extension icon.
 
-µb.updateBadgeAsync = (function() {
-    var tabIdToTimer = Object.create(null);
+µb.updateToolbarIcon = (function() {
+    let tabIdToDetails = new Map();
 
-    var updateBadge = function(tabId) {
-        delete tabIdToTimer[tabId];
+    let updateBadge = function(tabId) {
+        let parts = tabIdToDetails.get(tabId);
+        tabIdToDetails.delete(tabId);
 
-        var state = false;
-        var badge = '';
+        let state = 0;
+        let badge = '';
 
-        var pageStore = this.pageStoreFromTabId(tabId);
+        let pageStore = this.pageStoreFromTabId(tabId);
         if ( pageStore !== null ) {
-            state = pageStore.getNetFilteringSwitch();
-            if ( state && this.userSettings.showIconBadge && pageStore.perLoadBlockedRequestCount ) {
+            state = pageStore.getNetFilteringSwitch() ? 1 : 0;
+            if (
+                state === 1 &&
+                this.userSettings.showIconBadge &&
+                pageStore.perLoadBlockedRequestCount
+            ) {
                 badge = this.formatCount(pageStore.perLoadBlockedRequestCount);
             }
         }
 
-        vAPI.setIcon(tabId, state ? 'on' : 'off', badge);
+        vAPI.setIcon(tabId, state, badge, parts);
     };
 
-    return function(tabId) {
-        if ( tabIdToTimer[tabId] ) {
-            return;
+    // parts: bit 0 = icon
+    //        bit 1 = badge
+
+    return function(tabId, newParts) {
+        if ( vAPI.isBehindTheSceneTabId(tabId) ) { return; }
+        if ( newParts === undefined ) { newParts = 0x03; }
+        let currentParts = tabIdToDetails.get(tabId);
+        if ( currentParts === newParts ) { return; }
+        if ( currentParts === undefined ) {
+            vAPI.setTimeout(updateBadge.bind(this, tabId), 701);
+        } else {
+            newParts |= currentParts;
         }
-        if ( vAPI.isBehindTheSceneTabId(tabId) ) {
-            return;
-        }
-        tabIdToTimer[tabId] = vAPI.setTimeout(updateBadge.bind(this, tabId), 666);
+        tabIdToDetails.set(tabId, newParts);
     };
 })();
 
 /******************************************************************************/
 
 µb.updateTitle = (function() {
-    var tabIdToTimer = Object.create(null);
-    var tabIdToTryCount = Object.create(null);
+    var tabIdToTimer = new Map();
     var delay = 499;
 
-    var tryNoMore = function(tabId) {
-        delete tabIdToTryCount[tabId];
-    };
-
-    var tryAgain = function(tabId) {
-        var count = tabIdToTryCount[tabId];
-        if ( count === undefined ) {
-            return false;
-        }
-        if ( count === 1 ) {
-            delete tabIdToTryCount[tabId];
-            return false;
-        }
-        tabIdToTryCount[tabId] = count - 1;
-        tabIdToTimer[tabId] = vAPI.setTimeout(updateTitle.bind(µb, tabId), delay);
+    var tryAgain = function(entry) {
+        if ( entry.count === 1 ) { return false; }
+        entry.count -= 1;
+        tabIdToTimer.set(
+            entry.tabId,
+            vAPI.setTimeout(updateTitle.bind(null, entry), delay)
+        );
         return true;
     };
 
-    var onTabReady = function(tabId, tab) {
-        if ( !tab ) {
-            return tryNoMore(tabId);
-        }
-        var pageStore = this.pageStoreFromTabId(tabId);
-        if ( pageStore === null ) {
-            return tryNoMore(tabId);
-        }
+    var onTabReady = function(entry, tab) {
+        if ( !tab ) { return; }
+        var µb = µBlock;
+        var pageStore = µb.pageStoreFromTabId(entry.tabId);
+        if ( pageStore === null ) { return; }
         // Firefox needs this: if you detach a tab, the new tab won't have
         // its rawURL set. Concretely, this causes the logger to report an
         // entry to itself in the logger's tab selector.
         // TODO: Investigate for a fix vAPI-side.
         pageStore.rawURL = tab.url;
-        this.pageStoresToken = Date.now();
-        if ( !tab.title && tryAgain(tabId) ) {
-            return;
-        }
+        µb.pageStoresToken = Date.now();
+        if ( !tab.title && tryAgain(entry) ) { return; }
         // https://github.com/gorhill/uMatrix/issues/225
         // Sometimes title changes while page is loading.
         var settled = tab.title && tab.title === pageStore.title;
         pageStore.title = tab.title || tab.url || '';
-        this.pageStoresToken = Date.now();
-        if ( settled || !tryAgain(tabId) ) {
-            tryNoMore(tabId);
+        if ( !settled ) {
+            tryAgain(entry);
         }
     };
 
-    var updateTitle = function(tabId) {
-        delete tabIdToTimer[tabId];
-        vAPI.tabs.get(tabId, onTabReady.bind(this, tabId));
+    var updateTitle = function(entry) {
+        tabIdToTimer.delete(entry.tabId);
+        vAPI.tabs.get(entry.tabId, onTabReady.bind(null, entry));
     };
 
     return function(tabId) {
-        if ( vAPI.isBehindTheSceneTabId(tabId) ) {
-            return;
+        if ( vAPI.isBehindTheSceneTabId(tabId) ) { return; }
+        var timer = tabIdToTimer.get(tabId);
+        if ( timer !== undefined ) {
+            clearTimeout(timer);
         }
-        if ( tabIdToTimer[tabId] ) {
-            clearTimeout(tabIdToTimer[tabId]);
-        }
-        tabIdToTimer[tabId] = vAPI.setTimeout(updateTitle.bind(this, tabId), delay);
-        tabIdToTryCount[tabId] = 5;
+        tabIdToTimer.set(
+            tabId,
+            vAPI.setTimeout(
+                updateTitle.bind(null, { tabId: tabId, count: 5 }),
+                delay
+            )
+        );
     };
 })();
 
@@ -987,11 +1034,10 @@ var pageStoreJanitorSampleSize = 10;
 
 var pageStoreJanitor = function() {
     var vapiTabs = vAPI.tabs;
-    var tabIds = Object.keys(µb.pageStores).sort();
+    var tabIds = Array.from(µb.pageStores.keys()).sort();
     var checkTab = function(tabId) {
         vapiTabs.get(tabId, function(tab) {
             if ( !tab ) {
-                //console.error('tab.js> pageStoreJanitor(): stale page store found:', µtabId);
                 µb.unbindTabFromPageStats(tabId);
             }
         });
@@ -999,14 +1045,10 @@ var pageStoreJanitor = function() {
     if ( pageStoreJanitorSampleAt >= tabIds.length ) {
         pageStoreJanitorSampleAt = 0;
     }
-    var tabId;
     var n = Math.min(pageStoreJanitorSampleAt + pageStoreJanitorSampleSize, tabIds.length);
     for ( var i = pageStoreJanitorSampleAt; i < n; i++ ) {
-        tabId = tabIds[i];
-        // Do not remove behind-the-scene page store
-        if ( vAPI.isBehindTheSceneTabId(tabId) ) {
-            continue;
-        }
+        var tabId = tabIds[i];
+        if ( vAPI.isBehindTheSceneTabId(tabId) ) { continue; }
         checkTab(tabId);
     }
     pageStoreJanitorSampleAt = n;
