@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2014-2017 Raymond Hill
+    Copyright (C) 2014-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -219,7 +219,7 @@ var matchBucket = function(url, hostname, bucket, start) {
         }
         // Plain hostname
         else if ( line.indexOf('/') === -1 ) {
-            if ( reInvalidHostname.test(line) ) {
+            if ( this.reWhitelistBadHostname.test(line) ) {
                 key = '#';
                 directive = '# ' + line;
             } else {
@@ -242,7 +242,7 @@ var matchBucket = function(url, hostname, bucket, start) {
         // label (or else it would be just impossible to make an efficient
         // dict.
         else {
-            matches = reHostnameExtractor.exec(line);
+            matches = this.reWhitelistHostnameExtractor.exec(line);
             if ( !matches || matches.length !== 2 ) {
                 key = '#';
                 directive = '# ' + line;
@@ -266,27 +266,9 @@ var matchBucket = function(url, hostname, bucket, start) {
     return whitelist;
 };
 
-µBlock.validateWhitelistString = function(s) {
-    var lineIter = new this.LineIterator(s), line;
-    while ( !lineIter.eot() ) {
-        line = lineIter.next().trim();
-        if ( line === '' ) { continue; }
-        if ( line.startsWith('#') ) { continue; } // Comment
-        if ( line.indexOf('/') === -1 ) { // Plain hostname
-            if ( reInvalidHostname.test(line) ) { return false; }
-            continue;
-        }
-        if ( line.length > 2 && line.startsWith('/') && line.endsWith('/') ) { // Regex-based
-            try { new RegExp(line.slice(1, -1)); } catch(ex) { return false; }
-            continue;
-        }
-        if ( reHostnameExtractor.test(line) === false ) { return false; } // URL
-    }
-    return true;
-};
-
-var reInvalidHostname = /[^a-z0-9.\-\[\]:]/,
-    reHostnameExtractor = /([a-z0-9\[][a-z0-9.\-]*[a-z0-9\]])(?::[\d*]+)?\/(?:[^\x00-\x20\/]|$)[^\x00-\x20]*$/;
+// https://github.com/gorhill/uBlock/issues/3717
+µBlock.reWhitelistBadHostname = /[^a-z0-9.\-_\[\]:]/;
+µBlock.reWhitelistHostnameExtractor = /([a-z0-9.\-_\[\]]+)(?::[\d*]+)?\/(?:[^\x00-\x20\/]|$)[^\x00-\x20]*$/;
 
 /******************************************************************************/
 
@@ -304,6 +286,7 @@ var reInvalidHostname = /[^a-z0-9.\-\[\]:]/,
         us.noCosmeticFiltering = this.hnSwitches.evaluate('no-cosmetic-filtering', '*') === 1;
         us.noLargeMedia = this.hnSwitches.evaluate('no-large-media', '*') === 1;
         us.noRemoteFonts = this.hnSwitches.evaluate('no-remote-fonts', '*') === 1;
+        us.noCSPReports = this.hnSwitches.evaluate('no-csp-reports', '*') === 1;
         return us;
     }
 
@@ -371,6 +354,11 @@ var reInvalidHostname = /[^a-z0-9.\-\[\]:]/,
             this.saveHostnameSwitches();
         }
         break;
+    case 'noCSPReports':
+        if ( this.hnSwitches.toggle('no-csp-reports', '*', value ? 1 : 0) ) {
+            this.saveHostnameSwitches();
+        }
+        break;
     case 'prefetchingDisabled':
         if ( this.privacySettingsSupported ) {
             vAPI.browserSettings.set({ 'prefetching': !value });
@@ -392,6 +380,21 @@ var reInvalidHostname = /[^a-z0-9.\-\[\]:]/,
 
 /******************************************************************************/
 
+// https://www.reddit.com/r/uBlockOrigin/comments/8524cf/my_custom_scriptlets_doesnt_work_what_am_i_doing/
+
+µBlock.changeHiddenSettings = function(hs) {
+    var mustReloadResources =
+        hs.userResourcesLocation !== this.hiddenSettings.userResourcesLocation;
+    this.hiddenSettings = hs;
+    this.saveHiddenSettings();
+    if ( mustReloadResources ) {
+        this.redirectEngine.invalidateResourcesSelfie();
+        this.loadRedirectResources();
+    }
+};
+
+/******************************************************************************/
+
 µBlock.elementPickerExec = function(tabId, targetElement, zap) {
     if ( vAPI.isBehindTheSceneTabId(tabId) ) {
         return;
@@ -493,11 +496,9 @@ var reInvalidHostname = /[^a-z0-9.\-\[\]:]/,
         );
         break;
     case 'no-large-media':
-        if ( details.state === false ) {
-            var pageStore = this.pageStoreFromTabId(details.tabId);
-            if ( pageStore !== null ) {
-                pageStore.temporarilyAllowLargeMediaElements();
-            }
+        var pageStore = this.pageStoreFromTabId(details.tabId);
+        if ( pageStore !== null ) {
+            pageStore.temporarilyAllowLargeMediaElements(!details.state);
         }
         break;
     }
@@ -505,31 +506,19 @@ var reInvalidHostname = /[^a-z0-9.\-\[\]:]/,
 
 /******************************************************************************/
 
-µBlock.logCosmeticFilters = (function() {
-    var tabIdToTimerMap = {};
-
-    var injectNow = function(tabId) {
-        delete tabIdToTimerMap[tabId];
-        µBlock.scriptlets.injectDeep(tabId, 'cosmetic-logger');
-    };
-
-    var injectAsync = function(tabId) {
-        if ( tabIdToTimerMap.hasOwnProperty(tabId) ) {
-            return;
-        }
-        tabIdToTimerMap[tabId] = vAPI.setTimeout(
-            injectNow.bind(null, tabId),
-            100
-        );
-    };
-
-    return injectAsync;
-})();
+µBlock.logCosmeticFilters = function(tabId, frameId) {
+    if ( this.logger.isEnabled() ) {
+        vAPI.tabs.injectScript(tabId, {
+            file: '/js/scriptlets/cosmetic-logger.js',
+            frameId: frameId
+        });
+    }
+};
 
 /******************************************************************************/
 
 µBlock.scriptlets = (function() {
-    var pendingEntries = Object.create(null);
+    var pendingEntries = new Map();
 
     var Entry = function(tabId, scriptlet, callback) {
         this.tabId = tabId;
@@ -541,8 +530,9 @@ var reInvalidHostname = /[^a-z0-9.\-\[\]:]/,
     Entry.prototype.service = function(response) {
         if ( this.timer !== null ) {
             clearTimeout(this.timer);
+            this.timer = null;
         }
-        delete pendingEntries[makeKey(this.tabId, this.scriptlet)];
+        pendingEntries.delete(makeKey(this.tabId, this.scriptlet));
         this.callback(response);
     };
 
@@ -552,10 +542,8 @@ var reInvalidHostname = /[^a-z0-9.\-\[\]:]/,
 
     var report = function(tabId, scriptlet, response) {
         var key = makeKey(tabId, scriptlet);
-        var entry = pendingEntries[key];
-        if ( entry === undefined ) {
-            return;
-        }
+        var entry = pendingEntries.get(key);
+        if ( entry === undefined ) { return; }
         entry.service(response);
     };
 
@@ -565,20 +553,25 @@ var reInvalidHostname = /[^a-z0-9.\-\[\]:]/,
                 callback();
                 return;
             }
-            var key = makeKey(tabId, scriptlet);
-            if ( pendingEntries[key] !== undefined ) {
-                callback();
+            var key = makeKey(tabId, scriptlet),
+                entry = pendingEntries.get(key);
+            if ( entry !== undefined ) {
+                if ( callback !== entry.callback ) {
+                    callback();
+                }
                 return;
             }
-            pendingEntries[key] = new Entry(tabId, scriptlet, callback);
+            pendingEntries.set(key, new Entry(tabId, scriptlet, callback));
         }
-        vAPI.tabs.injectScript(tabId, { file: 'js/scriptlets/' + scriptlet + '.js' });
+        vAPI.tabs.injectScript(tabId, {
+            file: '/js/scriptlets/' + scriptlet + '.js'
+        });
     };
 
     // TODO: think about a callback mechanism.
     var injectDeep = function(tabId, scriptlet) {
         vAPI.tabs.injectScript(tabId, {
-            file: 'js/scriptlets/' + scriptlet + '.js',
+            file: '/js/scriptlets/' + scriptlet + '.js',
             allFrames: true
         });
     };
