@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2014-2016 Raymond Hill
+    Copyright (C) 2014-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -37,6 +37,8 @@ Naming convention from https://en.wikipedia.org/wiki/URI_scheme#Examples
 
 /******************************************************************************/
 
+var punycode = self.punycode;
+
 // Favorite regex tool: http://regex101.com/
 
 // Ref: <http://tools.ietf.org/html/rfc3986#page-50>
@@ -49,9 +51,10 @@ var reRFC3986 = /^([^:\/?#]+:)?(\/\/[^\/?#]*)?([^?#]*)(\?[^#]*)?(#.*)?/;
 // Derived
 var reSchemeFromURI          = /^[^:\/?#]+:/;
 var reAuthorityFromURI       = /^(?:[^:\/?#]+:)?(\/\/[^\/?#]+)/;
-var reOriginFromURI          = /^(?:[^:\/?#]+:)?(?:\/\/[^\/?#]+)/;
+var reOriginFromURI          = /^(?:[^:\/?#]+:)\/\/(?:[^\/?#]+)?/;
 var reCommonHostnameFromURL  = /^https?:\/\/([0-9a-z_][0-9a-z._-]*[0-9a-z])\//;
 var rePathFromURI            = /^(?:[^:\/?#]+:)?(?:\/\/[^\/?#]*)?([^?#]*)/;
+var reMustNormalizeHostname  = /[^0-9a-z._-]/;
 
 // These are to parse authority field, not parsed by above official regex
 // IPv6 is seen as an exception: a non-compatible IPv6 is first tried, and
@@ -61,11 +64,11 @@ var rePathFromURI            = /^(?:[^:\/?#]+:)?(?:\/\/[^\/?#]*)?([^?#]*)/;
 // https://github.com/gorhill/httpswitchboard/issues/211
 // "While a hostname may not contain other characters, such as the
 // "underscore character (_), other DNS names may contain the underscore"
-var reHostPortFromAuthority  = /^(?:[^@]*@)?([0-9a-z._-]*)(:\d*)?$/i;
+var reHostPortFromAuthority  = /^(?:[^@]*@)?([^:]*)(:\d*)?$/;
 var reIPv6PortFromAuthority  = /^(?:[^@]*@)?(\[[0-9a-f:]*\])(:\d*)?$/i;
 
 var reHostFromNakedAuthority = /^[0-9a-z._-]+[0-9a-z]$/i;
-var reHostFromAuthority      = /^(?:[^@]*@)?([0-9a-z._-]+)(?::\d*)?$/i;
+var reHostFromAuthority      = /^(?:[^@]*@)?([^:]+)(?::\d*)?$/;
 var reIPv6FromAuthority      = /^(?:[^@]*@)?(\[[0-9a-f:]+\])(?::\d*)?$/i;
 
 // Coarse (but fast) tests
@@ -250,53 +253,54 @@ URI.authorityFromURI = function(uri) {
 
 // The most used function, so it better be fast.
 
+// https://github.com/gorhill/uBlock/issues/1559
+//   See http://en.wikipedia.org/wiki/FQDN
+// https://bugzilla.mozilla.org/show_bug.cgi?id=1360285
+//   Revisit punycode dependency when above issue is fixed in Firefox.
+
 URI.hostnameFromURI = function(uri) {
     var matches = reCommonHostnameFromURL.exec(uri);
-    if ( matches ) {
-        return matches[1];
-    }
+    if ( matches !== null ) { return matches[1]; }
     matches = reAuthorityFromURI.exec(uri);
-    if ( !matches ) {
-        return '';
-    }
+    if ( matches === null ) { return ''; }
     var authority = matches[1].slice(2);
     // Assume very simple authority (most common case for µBlock)
     if ( reHostFromNakedAuthority.test(authority) ) {
         return authority.toLowerCase();
     }
     matches = reHostFromAuthority.exec(authority);
-    if ( !matches ) {
+    if ( matches === null ) {
         matches = reIPv6FromAuthority.exec(authority);
-        if ( !matches ) {
-            return '';
-        }
+        if ( matches === null ) { return ''; }
     }
-    // http://en.wikipedia.org/wiki/FQDN
-    // Also:
-    // - https://github.com/gorhill/uBlock/issues/1559
     var hostname = matches[1];
     while ( hostname.endsWith('.') ) {
         hostname = hostname.slice(0, -1);
     }
-    return hostname.toLowerCase();
+    if ( reMustNormalizeHostname.test(hostname) ) {
+        hostname = punycode.toASCII(hostname.toLowerCase());
+    }
+    return hostname;
 };
 
 /******************************************************************************/
 
 URI.domainFromHostname = function(hostname) {
-    // Try to skip looking up the PSL database
-    var entry = domainCache[hostname];
+    let entry = domainCache.get(hostname);
     if ( entry !== undefined ) {
         entry.tstamp = Date.now();
         return entry.domain;
     }
-    // Meh.. will have to search it
     if ( reIPAddressNaive.test(hostname) === false ) {
         return domainCacheAdd(hostname, psl.getDomain(hostname));
     }
     return domainCacheAdd(hostname, hostname);
 };
 
+URI.domainFromHostnameNoCache = function(hostname) {
+    return reIPAddressNaive.test(hostname) ? hostname : psl.getDomain(hostname);
+};
+
 URI.domain = function() {
     return this.domainFromHostname(this.hostname);
 };
@@ -307,6 +311,13 @@ var psl = publicSuffixList;
 
 /******************************************************************************/
 
+URI.entityFromDomain = function(domain) {
+    var pos = domain.indexOf('.');
+    return pos !== -1 ? domain.slice(0, pos) + '.*' : '';
+};
+
+/******************************************************************************/
+
 URI.pathFromURI = function(uri) {
     var matches = rePathFromURI.exec(uri);
     return matches !== null ? matches[1] : '';
@@ -319,11 +330,11 @@ URI.pathFromURI = function(uri) {
 // specific set of hostnames within a narrow time span -- in other words, I
 // believe probability of cache hit are high in uBlock.
 
-var domainCache = Object.create(null);
-var domainCacheCount = 0;
-var domainCacheCountLowWaterMark = 35;
-var domainCacheCountHighWaterMark = 50;
-var domainCacheEntryJunkyardMax = domainCacheCountHighWaterMark - domainCacheCountLowWaterMark;
+var domainCache = new Map();
+var domainCacheCountLowWaterMark = 40;
+var domainCacheCountHighWaterMark = 60;
+var domainCacheEntryJunkyardMax =
+    domainCacheCountHighWaterMark - domainCacheCountLowWaterMark;
 
 var DomainCacheEntry = function(domain) {
     this.init(domain);
@@ -343,23 +354,20 @@ DomainCacheEntry.prototype.dispose = function() {
 };
 
 var domainCacheEntryFactory = function(domain) {
-    var entry = domainCacheEntryJunkyard.pop();
-    if ( entry ) {
-        return entry.init(domain);
-    }
-    return new DomainCacheEntry(domain);
+    return domainCacheEntryJunkyard.length !== 0 ?
+        domainCacheEntryJunkyard.pop().init(domain) :
+        new DomainCacheEntry(domain);
 };
 
 var domainCacheEntryJunkyard = [];
 
 var domainCacheAdd = function(hostname, domain) {
-    var entry = domainCache[hostname];
+    var entry = domainCache.get(hostname);
     if ( entry !== undefined ) {
         entry.tstamp = Date.now();
     } else {
-        domainCache[hostname] = domainCacheEntryFactory(domain);
-        domainCacheCount += 1;
-        if ( domainCacheCount === domainCacheCountHighWaterMark ) {
+        domainCache.set(hostname, domainCacheEntryFactory(domain));
+        if ( domainCache.size === domainCacheCountHighWaterMark ) {
             domainCachePrune();
         }
     }
@@ -367,29 +375,24 @@ var domainCacheAdd = function(hostname, domain) {
 };
 
 var domainCacheEntrySort = function(a, b) {
-    return domainCache[b].tstamp - domainCache[a].tstamp;
+    return domainCache.get(b).tstamp - domainCache.get(a).tstamp;
 };
 
 var domainCachePrune = function() {
-    var hostnames = Object.keys(domainCache)
-                          .sort(domainCacheEntrySort)
-                          .slice(domainCacheCountLowWaterMark);
+    var hostnames = Array.from(domainCache.keys())
+                         .sort(domainCacheEntrySort)
+                         .slice(domainCacheCountLowWaterMark);
     var i = hostnames.length;
-    domainCacheCount -= i;
-    var hostname;
     while ( i-- ) {
-        hostname = hostnames[i];
-        domainCache[hostname].dispose();
-        delete domainCache[hostname];
+        var hostname = hostnames[i];
+        domainCache.get(hostname).dispose();
+        domainCache.delete(hostname);
     }
 };
 
-var domainCacheReset = function() {
-    domainCache = Object.create(null);
-    domainCacheCount = 0;
-};
-
-psl.onChanged.addListener(domainCacheReset);
+window.addEventListener('publicSuffixList', function() {
+    domainCache.clear();
+});
 
 /******************************************************************************/
 
@@ -403,15 +406,19 @@ URI.domainFromURI = function(uri) {
 /******************************************************************************/
 
 URI.isNetworkURI = function(uri) {
-    return /^(?:ftps?|https?|wss?):\/\//.test(uri);
+    return reNetworkURI.test(uri);
 };
 
+var reNetworkURI = /^(?:ftps?|https?|wss?):\/\//;
+
 /******************************************************************************/
 
 URI.isNetworkScheme = function(scheme) {
-    return /^(?:ftps?|https?|wss?)$/.test(scheme);
+    return reNetworkScheme.test(scheme);
 };
 
+var reNetworkScheme = /^(?:ftps?|https?|wss?)$/;
+
 /******************************************************************************/
 
 // Normalize the way µBlock expects it
