@@ -1,7 +1,7 @@
 /*******************************************************************************
 
     uBlock Origin - a browser extension to block requests.
-    Copyright (C) 2014-2017 Raymond Hill
+    Copyright (C) 2014-2018 Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -225,7 +225,9 @@
 /******************************************************************************/
 
 µBlock.CompiledLineWriter = function() {
-    this.output = [];
+    this.blockId = undefined;
+    this.block = undefined;
+    this.blocks = new Map();
     this.stringifier = JSON.stringify;
 };
 
@@ -235,46 +237,81 @@
 
 µBlock.CompiledLineWriter.prototype = {
     push: function(args) {
-        this.output[this.output.length] = this.stringifier(args);
+        this.block[this.block.length] = this.stringifier(args);
+    },
+    select: function(blockId) {
+        if ( blockId === this.blockId ) { return; }
+        this.blockId = blockId;
+        this.block = this.blocks.get(blockId);
+        if ( this.block === undefined ) {
+            this.blocks.set(blockId, (this.block = []));
+        }
     },
     toString: function() {
-        return this.output.join('\n');
+        var result = [];
+        for ( var entry of this.blocks ) {
+            if ( entry[1].length === 0 ) { continue; }
+            result.push(
+                '#block-start-' + entry[0],
+                entry[1].join('\n'),
+                '#block-end-' + entry[0]
+            );
+        }
+        return result.join('\n');
     }
 };
 
-µBlock.CompiledLineReader = function(raw) {
-    this.reset(raw);
+/******************************************************************************/
+
+µBlock.CompiledLineReader = function(raw, blockId) {
+    this.block = '';
+    this.len = 0;
+    this.offset = 0;
+    this.line = '';
     this.parser = JSON.parse;
+    this.blocks = new Map();
+    var reBlockStart = /^#block-start-(\d+)\n/gm,
+        match = reBlockStart.exec(raw),
+        beg, end;
+    while ( match !== null ) {
+        beg = match.index + match[0].length;
+        end = raw.indexOf('#block-end-' + match[1], beg);
+        this.blocks.set(parseInt(match[1], 10), raw.slice(beg, end));
+        reBlockStart.lastIndex = end;
+        match = reBlockStart.exec(raw);
+    }
+    if ( blockId !== undefined ) {
+        this.select(blockId);
+    }
 };
 
 µBlock.CompiledLineReader.prototype = {
-    reset: function(raw) {
-        this.input = raw;
-        this.len = raw.length;
-        this.offset = 0;
-        this.s = '';
-        return this;
-    },
     next: function() {
         if ( this.offset === this.len ) {
-            this.s = '';
+            this.line = '';
             return false;
         }
-        var pos = this.input.indexOf('\n', this.offset);
+        var pos = this.block.indexOf('\n', this.offset);
         if ( pos !== -1 ) {
-            this.s = this.input.slice(this.offset, pos);
+            this.line = this.block.slice(this.offset, pos);
             this.offset = pos + 1;
         } else {
-            this.s = this.input.slice(this.offset);
+            this.line = this.block.slice(this.offset);
             this.offset = this.len;
         }
         return true;
     },
+    select: function(blockId) {
+        this.block = this.blocks.get(blockId) || '';
+        this.len = this.block.length;
+        this.offset = 0;
+        return this;
+    },
     fingerprint: function() {
-        return this.s;
+        return this.line;
     },
     args: function() {
-        return this.parser(this.s);
+        return this.parser(this.line);
     }
 };
 
@@ -285,32 +322,26 @@
 // longer needed. A timer will be used for self-garbage-collect.
 // Cleaning up 10s after last hit sounds reasonable.
 
-// https://github.com/gorhill/uBlock/issues/2656
-// Can't use chained calls if we want to support legacy Map().
-
 µBlock.stringDeduplicater = {
     strings: new Map(),
     timer: undefined,
     last: 0,
 
     lookup: function(s) {
-        var t = this.strings.get(s);
+        let t = this.strings.get(s);
         if ( t === undefined ) {
-            this.strings.set(s, s);
-            t = this.strings.get(s);
-            if ( this.timer === undefined ) { this.cleanupAsync(); }
+            t = this.strings.set(s, s).get(s);
+            if ( this.timer === undefined ) {
+                this.timer = vAPI.setTimeout(() => { this.cleanup(); }, 10000);
+            }
         }
         this.last = Date.now();
         return t;
     },
 
-    cleanupAsync: function() {
-        this.timer = vAPI.setTimeout(this.cleanup.bind(this), 10000);
-    },
-
     cleanup: function() {
         if ( (Date.now() - this.last) < 10000 ) {
-            this.timer = vAPI.setTimeout(this.cleanup.bind(this), 10000);
+            this.timer = vAPI.setTimeout(() => { this.cleanup(); }, 10000);
         } else {
             this.timer = undefined;
             this.strings.clear();
@@ -320,36 +351,6 @@
 
 /******************************************************************************/
 
-µBlock.mapToArray = typeof Array.from === 'function'
-    ? Array.from
-    : function(map) {
-        var out = [];
-        for ( var entry of map ) {
-            out.push(entry);
-        }
-        return out;
-    };
-
-µBlock.mapFromArray = function(arr) {
-    return new Map(arr);
-};
-
-µBlock.setToArray = typeof Array.from === 'function'
-    ? Array.from
-    : function(dict) {
-        var out = [];
-        for ( var value of dict ) {
-            out.push(value);
-        }
-        return out;
-    };
-
-µBlock.setFromArray = function(arr) {
-    return new Set(arr);
-};
-
-/******************************************************************************/
-
 µBlock.openNewTab = function(details) {
     if ( details.url.startsWith('logger-ui.html') ) {
         if ( details.shiftKey ) {
@@ -365,14 +366,98 @@
 
 /******************************************************************************/
 
-// https://github.com/gorhill/uBlock/issues/2344
+µBlock.MRUCache = function(size) {
+    this.size = size;
+    this.array = [];
+    this.map = new Map();
+    this.resetTime = Date.now();
+};
 
-µBlock.matchCurrentLanguage = function(s) {
-    if ( typeof s !== 'string' ) { return false; }
-    if ( this.matchCurrentLanguage.reLang === undefined ) {
-        this.matchCurrentLanguage.reLang = new RegExp('\\b' + self.navigator.language.slice(0, 2) + '\\b');
+µBlock.MRUCache.prototype = {
+    add: function(key, value) {
+        var found = this.map.has(key);
+        this.map.set(key, value);
+        if ( !found ) {
+            if ( this.array.length === this.size ) {
+                this.map.delete(this.array.pop());
+            }
+            this.array.unshift(key);
+        }
+    },
+    remove: function(key) {
+        if ( this.map.has(key) ) {
+            this.array.splice(this.array.indexOf(key), 1);
+        }
+    },
+    lookup: function(key) {
+        var value = this.map.get(key);
+        if ( value !== undefined && this.array[0] !== key ) {
+            var i = this.array.indexOf(key);
+            do {
+                this.array[i] = this.array[i-1];
+            } while ( --i );
+            this.array[0] = key;
+        }
+        return value;
+    },
+    reset: function() {
+        this.array = [];
+        this.map.clear();
+        this.resetTime = Date.now();
     }
-    return this.matchCurrentLanguage.reLang.test(s);
 };
 
 /******************************************************************************/
+
+// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Regular_Expressions
+
+µBlock.escapeRegex = function(s) {
+    return s.replace(/[.*+?^${}()|[\]\\]/g, '\\$&');
+};
+
+/******************************************************************************/
+
+µBlock.decomposeHostname = (function() {
+    // For performance purpose, as simple tests as possible
+    let reHostnameVeryCoarse = /[g-z_-]/;
+    let reIPv4VeryCoarse = /\.\d+$/;
+
+    let isIPAddress = function(hostname) {
+        if ( reHostnameVeryCoarse.test(hostname) ) {
+            return false;
+        }
+        if ( reIPv4VeryCoarse.test(hostname) ) {
+            return true;
+        }
+        return hostname.startsWith('[');
+    };
+
+    let toBroaderHostname = function(hostname) {
+        let pos = hostname.indexOf('.');
+        if ( pos !== -1 ) {
+            return hostname.slice(pos + 1);
+        }
+        return hostname !== '*' && hostname !== '' ? '*' : '';
+    };
+
+    let toBroaderIPAddress = function(ipaddress) {
+        return ipaddress !== '*' && ipaddress !== '' ? '*' : '';
+    };
+
+    return function decomposeHostname(hostname, decomposed) {
+        if ( decomposed.length === 0 || decomposed[0] !== hostname ) {
+            let broaden = isIPAddress(hostname) ?
+                toBroaderIPAddress :
+                toBroaderHostname;
+            decomposed[0] = hostname;
+            let i = 1;
+            for (;;) {
+                hostname = broaden(hostname);
+                if ( hostname === '' ) { break; }
+                decomposed[i++] = hostname;
+            }
+            decomposed.length = i;
+        }
+        return decomposed;
+    };
+})();
